#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "multipht.h"


#undef DEBUG
#undef DEBUG2


	/* 
	 * this structure holds information on a bunch of stars 
	 * which all lie within the matching radius of each other.
	 */

struct s_group {
	int group_id;
	struct s_starp *starp_list;
	struct s_group *prev, *next;
};


	/* definitions of static functions in this file */
static int comp_xc_sort (struct s_star *s1, struct s_star *s2);
static int append_to_starp_list (struct s_starp *head_of_list, 
                                 struct s_starp *item);
static int append_to_group_list (struct s_group *head_of_list, 
                                 struct s_group *item);
#ifdef DEBUG
static void print_group_list (struct s_group *head_of_list);
static void print_star_list (struct s_image *image);
static void print_one_group (struct s_group *group);
#endif
static struct s_group *new_group (struct s_star *star);
static struct s_starp *new_starp (struct s_star *star);
static int insert_group (struct s_group **head_of_list,
                         struct s_group *new_group);
static int groups_overlap (struct s_group *group1,
                           struct s_group *group2);
static int merge_groups (struct s_group *group1,
                         struct s_group *group2);
static int remove_group (struct s_group *die_group);
static int free_groups (void);
static int delete_stars_in_groups (struct s_image *image, struct s_group *head);
static int copy_star_data (struct s_image *image, int to_index, int from_index);
static int comp_integer_values (int *i1, int *i2);



	/* global variables within this source code file */
int num_group;
struct s_group *group_head;




	/*********************************************************************
	 * PROCEDURE: remove_close_neighbors
	 *
	 * DESCRIPTION: Sometimes, the list of objects detected in a single
	 *              image contains bogus entries: objects which are 
	 *              so close to each other that they cannot be real.
	 *              If we allow multiple objects within a matching radius
	 *              to be passed along to the later analysis procedures,
	 *              we may encounter errors and insoluble situations.
	 *              "Garbage in, Garbage out."
	 *
	 *              So, this routine will run through the set of objects
	 *              read in from a file corresponding to a single image.
	 *              It checks to see if there are any instances of 
	 *              multiple objects located within one matching radius
	 *              of each other.  For each such group, the routine
	 *              keeps only the _brightest_, and discards the rest.
	 *
	 * RETURN:
	 *     0        if all goes well
	 *     1        if an error occurs
	 */
	
int 
remove_close_neighbors
	(
	struct s_image *image    /* I/O: check stars belonging to this image */
	)
{
	int i, j;
	int start_index, end_index;
	double matching_radius, matching_radius_sq;
	struct s_star *this_star, *other_star;

	group_head = (struct s_group *) NULL;
	num_group = 0;
	

	matching_radius = RADIUS;
	matching_radius_sq = matching_radius*matching_radius;


#ifdef DEBUG
	printf(" entering remove_close_neighbors for image %5d \n", 
			image->index);
	printf("     image %5d starts with %5d stars \n", 
			image->index, image->num_star);
#endif

	/* 
	 * We will sort the stars from the image by their "xc" values;
	 * makes some searching faster.
	 */
#ifdef DEBUG2
	{
		int i;

		printf(" before sorting, xc values are ... \n");
		for (i = 0; i < image->num_star; i++) {
			printf("  star %5d  xc %7.1f \n", i, image->stars[i].xc);
		}
	}
#endif
	qsort((char *) image->stars, image->num_star, sizeof(struct s_star),
					(PFI) comp_xc_sort);
#ifdef DEBUG2
	{
		int i;

		printf(" after  sorting, xc values are ... \n");
		for (i = 0; i < image->num_star; i++) {
			printf("  star %5d  xc %7.1f \n", i, image->stars[i].xc);
		}
	}
#endif

	/* 
	 * Now, we make a first pass through the list of all stars.  We check
	 * each star for neighbors within the critical radius.  If we find
	 * any such neighbors, we create a new s_group to hold a list of 
	 * all the stars in the group.
	 */
	for (i = 0; i < image->num_star; i++) {

		struct s_group *this_group = NULL;


		this_star = &(image->stars[i]);
#ifdef DEBUG2
		printf(" seeking group around star %5d (%7.1f %7.1f) \n",
					i, this_star->xc, this_star->yc);
#endif

		/* 
		 * we are going to look for other stars which are within the matching
		 * radius of this one.  We can eliminate all stars which are 
		 * more than the matching radius away in the x-direction.
		 * So, we do a simple walking up and down the array to find
		 * the first and last index within the array that we need to check.
		 */
		start_index = i;
		while (start_index > 0) {
			other_star = &(image->stars[start_index]);
			if (fabs(this_star->xc - other_star->xc) > matching_radius) {
				break;
			}
			start_index--;
		}

		end_index = i;
		while (end_index < (image->num_star - 1)) {
			other_star = &(image->stars[end_index]);
			if (fabs(other_star->xc - this_star->xc) > matching_radius) {
				break;
			}
			end_index++;
		}


		/*
		 * okay, check every star between these two indices in the array 
		 * to see if it lies within the matching radius of the current
		 * star.  If there are any matches at all, create a group and
		 * put pointers to each star into the group.
		 */
		
		for (j = start_index; j <= end_index; j++) {

			double dx, dy, dist_sq;

			/* don't compare the star to itself! */
			if (j == i) {
				continue;
			}

			other_star = &(image->stars[j]);
			dx = this_star->xc - other_star->xc;
			dy = this_star->yc - other_star->yc;
			dist_sq = (dx*dx + dy*dy);
			if (dist_sq < matching_radius_sq) {
#ifdef DEBUG2
				printf("   star %5d  (%7.1f %7.1f)  match %5d (%7.1f %7.1f)  dist %7.1f \n",
						i, this_star->xc, this_star->yc, 
						j, other_star->xc, other_star->yc, sqrt(dist_sq));
#endif

				/* create group here if necessary */
				if (this_group == NULL) {

					this_group = new_group(this_star);
					/* place group at the end of the linked list of groups */
					if (insert_group(&group_head, this_group) != 0) {
							fprintf(stderr, "remove_close_neighbors: insert_group fails, star %d \n", i);
							exit(1);
					}
				}

				/* add "other_star" to the group */
				{
					struct s_starp *other_p;

					other_p = (struct s_starp *) malloc(sizeof(struct s_starp));
					if (other_p == NULL) {
						fprintf(stderr, "remove_close_neighbors: malloc fails for other starp, star %d \n", i);
						exit(1);
					}
					other_p->star = other_star;
					other_p->next = NULL;

					/* place this star at end of the linked list of stars in group */
					if (append_to_starp_list(this_group->starp_list, other_p) != 0) {
							fprintf(stderr, "remove_close_neighbors: append_to_starp_list fails, star %d \n", i);
							exit(1);
						}
				}

			}
			else {
#ifdef DEBUG2
				printf("   star %5d  (%7.1f %7.1f)  no    %5d (%7.1f %7.1f)  dist %7.1f \n",
						i, this_star->xc, this_star->yc, 
						j, other_star->xc, other_star->yc, sqrt(dist_sq));
#endif
			}

		}
		


	}


	/*
	 * We now look at the list of groups we've created.   There may be some
	 * groups which are redundant, so we try to merge the groups into
	 * a unique set here.
	 */
	{
		struct s_group *this_group, *other_group, *next_group;
		int did_merge;


#ifdef DEBUG
		print_group_list(group_head);
#endif

		do {
			did_merge = 0;
#ifdef DEBUG
			printf(" about to check for overlapping groups \n");
#endif

			/* walk through list of groups, merging as we go */
			this_group = group_head;
			while (this_group != NULL) {

				other_group = this_group->next;
				while (other_group != NULL) {

					next_group = other_group->next;

					if (groups_overlap(this_group, other_group) == 1) {
#ifdef DEBUG
						printf(" groups %5d %5d overlap \n", 
								this_group->group_id, other_group->group_id);
#endif
						merge_groups(this_group, other_group);
						remove_group(other_group);
#ifdef DEBUG
						printf(" after removing group, here is list \n");
						print_group_list(group_head);
#endif
								
						did_merge = 1;
					}	
				
					other_group = next_group;
				}

				this_group = this_group->next;
			}

		} while (did_merge == 1);

	}
	

	/*
	 * We now look at each group, and discard from the image's list
	 * all but the brightest star in each group.  In the case of ties,
	 * we pick the first tied item.
	 */
	if (delete_stars_in_groups(image, group_head) != 0) {
		fprintf(stderr, "remove_close_neighbors: delete_stars_in_groups fails\n");
		return(1);		
	}


	/* 
	 * All done.  The image should now contain a list of stars,
	 * each of which is isolated from any others as far as the
	 * critical distance.
	 *
	 * We can free up all the memory we've allocated for the
	 * groups, since we won't use them any more.  But do NOT
	 * free the actual s_star structures to which the groups point!
	 */


#ifdef DEBUG
	printf("     image %5d ends   with %5d stars \n", 
			image->index, image->num_star);
#endif

	if (free_groups() != 0) {
		fprintf(stderr, "remove_close_neighbors: free_groups fails? \n");
		exit(1);
	}

	return(0);
}




   /**************************************************************
    * PROCEDURE: comp_xc_sort
    * 
    * DESCRIPTION: Compare the "xc" values for the two stars.
    *
    * RETURNS:
    *       -1 if the first star has an xc smaller than the second
    *        0                             equal to 
    *       +1                             larger than
    */

static int
comp_xc_sort
   (
   struct s_star *s1,
   struct s_star *s2
   )
{
   if (s1->xc < s2->xc) {
		return(-1);
   }
   if (s1->xc > s2->xc) {
      return(1);
   }
   return(0);
}





   /**************************************************************
    * PROCEDURE: append_to_starp_list
    * 
    * DESCRIPTION: Given a pointer to a list of "struct s_starp"
	 *              items, and a pointer to one s_starp, place
	 *              the given s_starp at the end of the linked list.
	 *
	 *              If the user passes a NULL pointer as the head
	 *              of the list, this routine will print error
	 *              message and terminate.
    *
    * RETURNS:
	 *              0           if all goes well
	 *              1           if handed NULL list or item
    */

static int
append_to_starp_list
   (
   struct s_starp *head_of_list,    /* I: head of a list of items */
   struct s_starp *item             /* I: item to be append to list */
   )
{
	struct s_starp *p;

	/* check that given pointers not NULL; if so, print message and return */
	if (head_of_list == NULL) {
		fprintf(stderr, "append_to_starp_list: given NULL list?! \n");
		return(1);
	}
	if (item == NULL) {
		fprintf(stderr, "append_to_starp_list: given NULL item?! \n");
		return(1);
	}

	p = head_of_list;
	while (p->next != NULL) {
		p = p->next;
	}
	p->next = item;

	return(0);
}


   /**************************************************************
    * PROCEDURE: append_to_group_list
    * 
    * DESCRIPTION: Given a pointer to a list of "struct s_group"
	 *              items, and a pointer to one s_group, place
	 *              the given s_group at the end of the linked list.
	 *
	 *              If the user passes a NULL pointer as the head
	 *              of the list, this routine will print error
	 *              message and terminate.
    *
    * RETURNS:
	 *              0           if all goes well
	 *              1           if handed NULL list or item
    */

static int
append_to_group_list
   (
   struct s_group *head_of_list,    /* I: head of a list of items */
   struct s_group *item             /* I: item to be append to list */
   )
{
	struct s_group *p;

	/* check that given pointers not NULL; if so, print message and return */
	if (head_of_list == NULL) {
		fprintf(stderr, "append_to_group_list: given NULL list?! \n");
		return(1);
	}
	if (item == NULL) {
		fprintf(stderr, "append_to_group_list: given NULL item?! \n");
		return(1);
	}

	p = head_of_list;
	while (p->next != NULL) {
		p = p->next;
	}
	p->next = item;
	item->prev = p;

	return(0);
}



   /**************************************************************
    * PROCEDURE: print_group_list
    * 
    * DESCRIPTION: Given a pointer to a linked list of 
	 *              s_group structures, print information on
	 *              each one in the list in turn.
    *
    * RETURNS:
	 *              nothing
    */

#ifdef DEBUG

static void
print_group_list
   (
   struct s_group *head_of_list     /* I: top of list of group */
   )
{
	struct s_group *p;

	for (p = head_of_list; p != NULL; p = p->next) {
		print_one_group(p);
	}

}

#endif

   /**************************************************************
    * PROCEDURE: print_one_group
    * 
    * DESCRIPTION: Given a pointer to one s_group, print out
	 *              a list which gives the group ID, and then
	 *              each star within the group, one per line.
    *
    * RETURNS:
	 *              nothing
    */

#ifdef DEBUG

static void
print_one_group
   (
   struct s_group *group       /* I: print info on this group */
   )
{
	struct s_starp *starp_p;
	struct s_star *sp;

	printf(" group %6d  \n", group->group_id);

	starp_p = group->starp_list;
	while (starp_p != NULL) {
		sp = starp_p->star;
		printf("    star %6d  (%7.1f %7.1f) \n",
				sp->id, sp->xc, sp->yc);
		starp_p = starp_p->next;
	}

}

#endif


   /**************************************************************
    * PROCEDURE: print_star_list
    * 
    * DESCRIPTION: Given a pointer to one image,
	 *              print a list of stars belonging to the image,
	 *              one per line.
    *
    * RETURNS:
	 *              nothing
    */

#ifdef DEBUG

static void
print_star_list
   (
   struct s_image *image        /* I: print stars in this image */
   )
{
	int i;
	struct s_star *star;
	
	for (i = 0; i < image->num_star; i++) {
		star = &(image->stars[i]);
		printf("    star %5d  at %7.1f %7.1f  %6.3f \n",
				i, star->xc, star->yc, star->mag);
	}
}

#endif


   /**************************************************************
    * PROCEDURE: new_group
    * 
    * DESCRIPTION: Create a new group structure, and place
	 *              into it a starp structure which points
	 *              to the given star.
    *
    * RETURNS:
	 *              pointer to the new s_group structure
    */

static struct s_group *
new_group
   (
   struct s_star *star      /* I: have new group include this star */
   )
{
	struct s_group *this_group;
	struct s_starp *starp_p;

	this_group = (struct s_group *) malloc(sizeof(struct s_group));
	if (this_group == NULL) {
		fprintf(stderr, "new_group: malloc fails for group, star %5d (%7.1f %7.1f)\n",
						star->id, star->xc, star->yc);
		exit(1);
	}
		
	/* initialize the group */
	this_group->group_id = num_group++;
	starp_p = new_starp(star);
	this_group->starp_list = starp_p;
	this_group->next = NULL;
	this_group->prev = NULL;

	return(this_group);
}


   /**************************************************************
    * PROCEDURE: new_starp
    * 
    * DESCRIPTION: Create a new s_starp structure, and cause
	 *              it to point to the given star.
    *
    * RETURNS:
	 *              pointer to the new s_group structure
    */

static struct s_starp *
new_starp
   (
   struct s_star *star      /* I: have new group include this star */
   )
{
	struct s_starp *new;

	new = (struct s_starp *) malloc(sizeof(struct s_starp));
	if (new == NULL) {
		fprintf(stderr, "new_starp: malloc fails for starp, star %5d (%7.1f %7.1f)\n",
						star->id, star->xc, star->yc);
		exit(1);
	}
	
	new->star = star;
	new->next = NULL;

	return(new);
}

		


   /**************************************************************
    * PROCEDURE: insert_group
    * 
    * DESCRIPTION: Given the pointer to the pointer to the
	 *              head of a linked list of s_group structures,
	 *              and one new s_group, we want to insert
	 *              the new element into the list.
	 * 
	 *              If the head of the list is NULL, make the
	 *              head point to the new group.
	 *              Otherwise, walk to the end of the existing
	 *              list and insert the new group there.
    *
    * RETURNS:
	 *              0                 if all goes well
	 *              1                 if an error occurs
    */

static int
insert_group
   (
	struct s_group **head_of_list,  /* I/O: we want to add onto this list */
   struct s_group *new_group       /* I: put this at end of list */
   )
{

#ifdef DEBUG2
	printf(" insert_group: before we add \n");
	print_group_list(*head_of_list);
#endif

	if ((*head_of_list) == NULL) {
		*head_of_list = new_group;
	}
	else {
		if (append_to_group_list(*head_of_list, new_group) != 0) {
			fprintf(stderr, "insert_group: append_to_group_list fails \n");
			exit(1);
		}
	}

#ifdef DEBUG2
	printf(" insert_group: after adding  \n");
	print_group_list(*head_of_list);
#endif

	return(0);
}


   /**************************************************************
    * PROCEDURE: groups_overlap
    * 
    * DESCRIPTION: Given pointers to two s_group structures,
	 *              look to see if they share any stars.
	 *
	 *              Groups should be small enough, I hope,
	 *              that a brute force N^2 algorithm is okay.
    *
    * RETURNS:
	 *              0                 if groups do NOT share stars
	 *              1                 if groups do share stars
    */

static int
groups_overlap
   (
	struct s_group *group1,     /* I: do any stars in this group ... */
   struct s_group *group2      /* I:  ... also appear in this group? */
   )
{
	struct s_starp *starp_1, *starp_2;
	struct s_star *sp1, *sp2;

	for (starp_1 = group1->starp_list; starp_1 != NULL;
				starp_1 = starp_1->next) {
		
		sp1 = starp_1->star;

		for (starp_2 = group2->starp_list; starp_2 != NULL;
					starp_2 = starp_2->next) {

			sp2 = starp_2->star;
			
			if (sp1 == sp2) {
				return(1);
			}

		}

	}

	return(0);
}


   /**************************************************************
    * PROCEDURE: merge_groups
    * 
    * DESCRIPTION: Given pointers to two s_group structures,
	 *              which contain at least some stars in common,
	 *              this routine adds into the first group
	 *              all the stars from the second group
	 *              which do not yet appear in the first
	 *              group.
	 *
	 *              In other words, after calling this function,
	 *              the first group will contain all the 
	 *              stars of both groups.  We can then
	 *              later (in another function) delete
	 *              the second group.
	 *
	 *              Groups should be small enough, I hope,
	 *              that a brute force N^2 algorithm is okay.
    *
    * RETURNS:
	 *              0                 if groups do NOT share stars
	 *              1                 if groups do share stars
    */

static int
merge_groups
   (
	struct s_group *group1,     /* I: insert into this group ... */
   struct s_group *group2      /* I: ... stars from this group  */
   )
{
	int found;
	struct s_starp *starp_1, *starp_2, *copy_starp_2;
	struct s_star *sp1, *sp2;

#ifdef DEBUG2
	printf(" entering merge_groups \n");
	printf(" here comes group1 ... \n");
	print_one_group(group1);
	printf(" here comes group2 ... \n");
	print_one_group(group2);
#endif


	for (starp_2 = group2->starp_list; starp_2 != NULL;
				starp_2 = starp_2->next) {
		
		sp2 = starp_2->star;

		/* check to see if sp2 appears in group1 yet */
		found = 0;
		for (starp_1 = group1->starp_list; starp_1 != NULL;
					starp_1 = starp_1->next) {

			sp1 = starp_1->star;
			if (sp1 == sp2) {
				/* yes, so no need to add a copy */
				found = 1;
				break;
			}
		}
		if (found == 0) {
			/* 
			 * we need to add to group1 a struct s_starp 
			 * which points to sp2 
			 */
			copy_starp_2 = new_starp(sp2);
			append_to_starp_list(group1->starp_list, copy_starp_2);
		}

	}

#ifdef DEBUG2
	printf(" about to leave merge_groups \n");
	printf(" here is new group1 ... \n");
	print_one_group(group1);
	printf(" here is unchanged group2 ... \n");
	print_one_group(group2);
#endif

	return(0);
}


   /**************************************************************
    * PROCEDURE: remove_group
    * 
    * DESCRIPTION: Remove the given group.  What this means is
	 *              
	 *                a) remove it from the linked list of 
	 *                   all groups (and decrement the number
	 *                   of groups in that list)
	 *
	 *                b) de-allocate the memory for all
	 *                   the s_starp structures
	 *                   BUT LEAVE THE S_STAR STRUCTURES ALONE
	 *
	 *                c) de-allocate the memory for the
	 *                   s_group structure itself
	 *
    * RETURNS:
	 *              0                 if groups do NOT share stars
	 *              1                 if groups do share stars
    */

static int
remove_group
   (
	struct s_group *die_group   /* I: get rid of this group */
   )
{
	struct s_starp *starp_p, *starp_p_next;

	if (die_group == NULL) {
		return(0);
	}

	/* 
	 * first, remove from the big linked list.
	 * There are two special cases and one general case.
	 */
	if (group_head == die_group) {
		/* special case one: this is the head of list */
		group_head = die_group->next;
		if (die_group->next != NULL) {
			(die_group->next)->prev = NULL;
		}
	}
	else if (die_group->next == NULL) {
		/* special case two: this is tail of list */
		(die_group->prev)->next = NULL;
	}
	else {
		/* general case: in middle of list */
		(die_group->prev)->next = die_group->next;
		(die_group->next)->prev = die_group->prev;
	}
	num_group--;
	

	/* second, free the starp structures */
	starp_p = die_group->starp_list;
	while (starp_p != NULL) {
		starp_p_next = starp_p->next;
		free(starp_p);
		starp_p = starp_p_next;
	}

	/* finally, free the group structure itself */
	free(die_group);

	return(0);
}


   /**************************************************************
    * PROCEDURE: free_groups
    * 
    * DESCRIPTION: Deallocate space for all the group structures
	 *              and their s_starp structures.
	 *
    * RETURNS:
	 *              0                 if groups do NOT share stars
	 *              1                 if groups do share stars
    */

static int
free_groups
   (
	void
   )
{
	while (group_head != NULL) {
		remove_group(group_head);
	}

	return(0);
}



   /**************************************************************
    * PROCEDURE: delete_stars_in_groups
    * 
    * DESCRIPTION: Given an image and a list of all the groups
	 *              of neighboring stars in that image, we
	 *              want to get rid of all but the brightest
	 *              star in each group.
	 *
	 *              Walk through the list of groups.  For each one,
	 *              we identify the brightest star, then add
	 *              all the _other_ (fainter) stars to an array
	 *              if star IDs which should be removed.
	 *              When this array is complete, we sort it by
	 *              ID number -- which is the same as position
	 *              in the array of stars.  We can then make a
	 *              single pass through the array of items to be
	 *              deleted, shifting image->stars[] array elements
	 *              as we go.
	 *
	 *              We must re-set the image->num_star value, too.
	 *
    * RETURNS:
	 *              0                 if groups do NOT share stars
	 *              1                 if groups do share stars
    */

static int
delete_stars_in_groups
   (
	struct s_image *image,      /* I: image from which stars come */
	struct s_group *head        /* I: head of linked list of all groups */
   )
{
	int i, j, max_num_to_delete;
	int *stars_to_die, num_stars_to_die;
	int die_index;
	struct s_group *group;
	struct s_star *star, *brightest_star;
	struct s_starp *starp;

	/* 
	 * count the total number of stars in all groups combined.
	 * The number we will delete must be less than this ...
	 */
	max_num_to_delete = 0;
	group = head;
	while (group != NULL) {
		for (starp = group->starp_list; starp != NULL; starp = starp->next) {
			max_num_to_delete++;
		}
		group = group->next;
	}
		
	/* 
	 * allocate space for an array of integers.  This array will
	 * hold the ID numbers of each star which we will delete.
	 * The array will be a bit bigger than we need, but that's okay.
	 */
	if ((stars_to_die = (int *) malloc(max_num_to_delete*sizeof(int))) == NULL) {
		fprintf(stderr, "delete_stars_in_groups: malloc fails for %d ints \n",
				max_num_to_delete);
		exit(1);
	}

	/*
	 * Each struct s_star element has a field called "id", which
	 * will _later_ be used as the star's identifier in the
	 * overall ensemble.  At the moment, this field is not yet
	 * used, and has been set to -1.  We will briefly use
	 * this field now: set it to the position of the star in
	 * the image->stars[] array.  That will give each star
	 * a unique "id" number.
	 *
	 * At the end of this function, we'll reset the "id" field
	 * to its value of -1.
	 */
	for (i = 0; i < image->num_star; i++) {
		star = &(image->stars[i]);
		if (star->id != -1) {
			fprintf(stderr, "delete_stars_in_groups: star %d has id number %d, not -1 ?!\n",
					i, star->id);
			exit(1);
		}
		star->id = i;
	}


	num_stars_to_die = 0;
	group = head;
	while (group != NULL) {

		/* identify brightest star in this group */
		brightest_star = NULL;
		for (starp = group->starp_list; starp != NULL; starp = starp->next) {
			star = starp->star;
			if (star == NULL) {
				fprintf(stderr, "delete_stars_in_groups: group %d has NULL star?! \n",
						group->group_id);
				return(1);
			}
			if (brightest_star == NULL) {
				brightest_star = star;
			}
			else {
				if (star->mag < brightest_star->mag) {
					brightest_star = star;
				}
			}
		}

		/* a sanity check -- if no stars in group, just return */
		if (brightest_star == NULL) {
#ifdef DEBUG
			printf("delete_stars_in_groups: empty group %d ?! \n",
					group->group_id);
#endif
			return(0);
		}
			
		
		/* mark all stars except brightest as "to be deleted" */
		for (starp = group->starp_list; starp != NULL; starp = starp->next) {
			star = starp->star;
			if (star != brightest_star) {

				/* put this star's ID value into array "stars_to_die[]" */
				stars_to_die[num_stars_to_die] = star->id;
				num_stars_to_die++;

			}
		}

		group = group->next;
	}

#ifdef DEBUG2
	printf(" here comes array of stars to die for image %5d \n",
				image->index);
	for (i = 0; i < num_stars_to_die; i++) {
		printf(" %6d \n", stars_to_die[i]);
	}
#endif

	/* 
	 * sort the "stars_to_die[]" array by star ID number, which 
	 * is the same as the position of each star in the 
	 * image->stars[] array.
	 */
	qsort((char *) stars_to_die, num_stars_to_die, sizeof(int),
					(PFI) comp_integer_values);
	
#ifdef DEBUG2
	printf(" after sorting, again array of stars to die for image %5d \n",
				image->index);
	for (i = 0; i < num_stars_to_die; i++) {
		printf(" %6d \n", stars_to_die[i]);
	}
#endif

	/* 
	 * Now walk through the "stars_to_die[]" array, deleting
	 * each one from the image->stars[] array.
	 * Each time we delete a star, we must decrement the 
	 * image->num_star counter.
	 */
#ifdef DEBUG2
	printf(" image %5d has all these stars before we delete ...  \n",
				image->index);
	for (i = 0; i < image->num_star; i++) {
		printf(" %6d %7.1f %7.1f \n", image->stars[i].id,
				image->stars[i].xc, image->stars[i].yc);
	}
#endif

	die_index = 0;
	for (i = 0; i < image->num_star; i++) {
		if (image->stars[i].id == stars_to_die[die_index]) {

#ifdef DEBUG2
			printf(" going to delete array elem %5d == ID %5d \n",
						i, image->stars[i].id);
#endif

			/* 
			 * delete item i from star array by shifting all the
			 * following elements upwards in array 
			 */
			for (j = i; j < image->num_star - 1; j++) {
#ifdef DEBUG2
				printf(" about to shift star %5d <- %5d \n", j, j+1);
#endif
				if (copy_star_data(image, j, j+1) != 0) {
					fprintf(stderr, "delete_stars_in_groups: copy_star_data fails for image %d to %d from %d \n",
							image->index, j, j+1);
					exit(1);
				}
			}

			/* 
			 * decrease the counter i so that we check the star which
			 * has just moved into the place of the one we deleted 
			 */
			i--;

			/* decrement the number of stars left in the image */
			image->num_star--;

			if (++die_index >= num_stars_to_die) {
				/* all done -- no more to delete */

				break;
			}
		}
	}

#ifdef DEBUG
	printf(" image %5d has all these stars AFTER we delete ...  \n",
				image->index);
	for (i = 0; i < image->num_star; i++) {
		printf(" %6d %7.1f %7.1f \n", image->stars[i].id,
				image->stars[i].xc, image->stars[i].yc);
	}
#endif



	/*
	 * Reset the "id" fields of all surviving stars to -1.
	 */
	for (i = 0; i < image->num_star; i++) {
		star = &(image->stars[i]);
		star->id = -1;
	}



	return(0);
}


   /**************************************************************
    * PROCEDURE: copy_star_data
    * 
    * DESCRIPTION: Given an image and the indices of two
	 *              entries in the image's "stars[]" array,
	 *              copy all the information from the 
	 *              "from" element into the "to" element's position.
	 *
	 *              In other words, overwrite the information
	 *              in the "to" element with the information
	 *              in the "from" element.
	 *
    * RETURNS:
	 *              0                 if groups do NOT share stars
	 *              1                 if groups do share stars
    */

static int
copy_star_data
   (
	struct s_image *image,      /* I: image containing array of stars */
	int to_index,               /* I: overwrite data in this element */
	int from_index              /* I: take data from this element */
   )
{
	struct s_star *to_star, *from_star;

	/* sanity checks */
	if ((to_index < 0) || (to_index >= image->num_star)) {
		fprintf(stderr, "copy_star_data: image %d bad to_index %d \n",
				image->index, to_index);
		return(1);
	}
	if ((from_index < 0) || (from_index >= image->num_star)) {
		fprintf(stderr, "copy_star_data: image %d bad from_index %d \n",
				image->index, from_index);
		return(1);
	}
	if (from_index == to_index) {
		fprintf(stderr, "copy_star_data: to_index %d equals from_index %d !? \n",
				to_index, from_index);
		return(1);
	}

	to_star = &(image->stars[to_index]);
	from_star = &(image->stars[from_index]);

	to_star->image_id = from_star->image_id;
	to_star->id = from_star->id;
	to_star->hash = from_star->hash;
	to_star->xc = from_star->xc;
	to_star->yc = from_star->yc;
	to_star->jd = from_star->jd;
	to_star->mag = from_star->mag;
	to_star->err = from_star->err;
	to_star->w3 = from_star->w3;
	to_star->w4 = from_star->w4;
	to_star->saturated = from_star->saturated;


	return(0);
}

   /**************************************************************
    * PROCEDURE: comp_integer_values
    * 
    * DESCRIPTION: Compare the values of two integers.
    *
    * RETURNS:
    *       -1 if the first integer has a value smaller than the second
    *        0                                  equal to 
    *       +1                                  larger than
    */

static int
comp_integer_values
   (
   int *i1,          /* I: compare this integer's value ... */
   int *i2           /* I: ... to this one's */
   )
{
   if (*i1 < *i2) {
		return(-1);
   }
   if (*i1 > *i2) {
      return(1);
   }
   return(0);
}

