
	/********************************************************************
	 * Read a bunch of files, each of which has magnitude and position
	 * measurements for a set of stars in one image.  Re-arrange the
	 * data so that all the measurements for a given star are grouped
	 * together.
	 *
	 * This is the first stage in analysis.  The second stage reads
	 * the re-arranged data and finds the optimal shifts to be applied
	 * to each image to bring all measurements into agreement with each
	 * other.
	 *
	 * Based on the work of Kent Honeycutt, PASP 104, 435 (1992).
	 *
	 * MWR 9/8/2004
	 *
	 * Added "jd=" command-line option to specify column containing JD 
	 * (or any sort of time) values.
	 * MWR 9/15/2004
	 *
	 * Lots of changes:
	 *  - Removed obsolete command-line arg: "aper="
	 *  - Now abort if given a template which doesn't match an image name.
	 * MWR 10/19/2004
	 *
	 * Minor fixes:
	 *  - added command-line option: "quiet" 
	 * MWR 2/25/2005
	 *
	 * Added "min_match=" command-line option.
	 * MWR 6/13/2005
	 *
	 * Modified read_pht_line() so that it checks for "err" values less
	 *     than MIN_MAG_UNCERT; replaces any it finds with MIN_MAG_UNCERT.
	 *     This avoids cases where err=0 caused weight w4=1/(err*err) = inf
	 * MWR 5/27/2006
	 *
	 * Added "isolate" keyword and code to call new func 
	 *      "remove_close_neighbors()" if the user wishes.
	 * MWR 1/17/2007
	 */
	 

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "multipht.h"

#define SAMEPOS   1e-6		/* if two consecutive stars have both  */
                           /*   x and y positions */
                      		/*   that differ by less than this much, */
                     		/*   consider them duplicate entries for the */
                      		/*   same actual star */ 
                           /*   and ignore the second entry */


	/* definitions of static functions in this file */
static int do_listfile(char *file, int x_col, int y_col, 
		                               int mag_col, int err_col, int jd_col,
                                     int flags);
static int do_pht(char *file, int x_col, int y_col, int mag_col, int err_col,
		            int jd_col, int flags);
static int slurp_header(char *file, struct s_image *image);
static int slurp_pht_data(char *file, struct s_image *image,
		           int x_col, int y_col, int mag_col, int err_col, int jd_col);
static int read_pht_line(char *line, struct s_star *star,
		           int x_col, int y_col, int mag_col, int err_col, int jd_col);
static void print_star(struct s_star *s);
static void proc_pht(int flags, char *template, int min_match);
static void do_print(void);
static int num_in_file(char *file);
static int identical(struct s_image *im, int n);


char progname[] = "multipht";
static char usage[] = "multipht [debug] [print] \n                x= y= mag= err= [jd=] [auto] [list= ] [outfile=] \n                [template= ] [matchrad= ] [min_match= ] [quiet] \n                [star_cut=] [image_cut=] [isolate] file.pht ... ";

int
main
	(
	int argc, 
	char *argv[]
	)
{
	int i, flags, numargfile, *argfile;
	int x_col, y_col, mag_col, err_col, jd_col;
	int min_match;
	char file[LINELEN], outfile[LINELEN], template[LINELEN];
	double matchrad;

	RADIUS = DEF_RADIUS;
	min_match = MIN_MATCH;
	flags = 0;
	debug_flag = 0;
	quiet_flag = 0;
	star_cut = STAR_CUT;
	image_cut = IMAGE_CUT;
	num_images = 0;
	numargfile = 0;
	x_col = -1;
	y_col = -1;
	mag_col = -1;
	err_col = -1;
	jd_col = -1;
	strcpy(outfile, "");
	strcpy(template, "");
	saved_cutoff_value = 0;

	if (argc == 1) {
		fprintf(stderr, "usage: %s\n", usage);
		exit(0);
	}

	/* allocate enough pointers for all the images we'll read */
	if (argc >= MAX_IMAGE) {
		fprintf(stderr, "number of args >= MAX_IMAGE = %d\n", MAX_IMAGE);
		exit(-1);
	}
	if ((image_arr = (struct s_image *) calloc(MAX_IMAGE, sizeof(struct s_image))) == NULL) {
			fprintf(stderr, "can't calloc for image array \n");
			exit(-1);
	}
	if ((argfile = (int *) calloc(argc, sizeof(int))) == NULL) {
		fprintf(stderr, "can't calloc for argfile list?! \n");
		exit(-1);
	}
	
	/* read in data from all the files */
	for (i = 1; i < argc; i++) {
		if (strcmp(argv[i], "help") == 0) {
			fprintf(stderr, "usage: %s\n", usage);
			exit(0);
		}
		if (strcmp(argv[i], "debug") == 0) {
			debug_flag = 1;
			continue;
		}
		if (strcmp(argv[i], "quiet") == 0) {
			quiet_flag = 1;
			continue;
		}
		if (strcmp(argv[i], "print") == 0) {
			flags |= PRINT;
			continue;
		}
		if (strcmp(argv[i], "auto") == 0) {
			flags |= AUTO;
			continue;
		}
		if (strcmp(argv[i], "isolate") == 0) {
			flags |= ISOLATE;
			continue;
		}
		if (strncmp(argv[i], "x=", 2) == 0) {
			if (sscanf(argv[i] + 2, "%d", &x_col) != 1) {
				fprintf(stderr, "usage: %s\n", usage);
				exit(1);
			}
			continue;
		}
		if (strncmp(argv[i], "y=", 2) == 0) {
			if (sscanf(argv[i] + 2, "%d", &y_col) != 1) {
				fprintf(stderr, "usage: %s\n", usage);
				exit(1);
			}
			continue;
		}
		if (strncmp(argv[i], "mag=", 4) == 0) {
			if (sscanf(argv[i] + 4, "%d", &mag_col) != 1) {
				fprintf(stderr, "usage: %s\n", usage);
				exit(1);
			}
			continue;
		}
		if (strncmp(argv[i], "err=", 4) == 0) {
			if (sscanf(argv[i] + 4, "%d", &err_col) != 1) {
				fprintf(stderr, "usage: %s\n", usage);
				exit(1);
			}
			continue;
		}
		if (strncmp(argv[i], "jd=", 3) == 0) {
			if (sscanf(argv[i] + 3, "%d", &jd_col) != 1) {
				fprintf(stderr, "usage: %s\n", usage);
				exit(1);
			}
			continue;
		}
		if (strncmp(argv[i], "outfile=", 8) == 0) {
			if (sscanf(argv[i] + 8, "%s", outfile) != 1) {
				fprintf(stderr, "usage: %s\n", usage);
				exit(1);
			}
			continue;
		}
		if (strncmp(argv[i], "template=", 9) == 0) {
			if (sscanf(argv[i] + 9, "%s", template) != 1) {
				fprintf(stderr, "usage: %s\n", usage);
				exit(1);
			}
			continue;
		}
		if (strncmp(argv[i], "list=", 5) == 0) {
			continue;
		}
		if (strncmp(argv[i], "matchrad=", 9) == 0) {
			if (sscanf(argv[i] + 9, "%lf", &matchrad) != 1) {
				fprintf(stderr, "usage: %s\n", usage);
				exit(1);
			}
			RADIUS = matchrad;
			continue;
		}
		if (strncmp(argv[i], "star_cut=", 9) == 0) {
			if (sscanf(argv[i] + 9, "%lf", &star_cut) != 1) {
				fprintf(stderr, "usage: %s\n", usage);
				exit(1);
			}
			continue;
		}
		if (strncmp(argv[i], "image_cut=", 10) == 0) {
			if (sscanf(argv[i] + 10, "%lf", &image_cut) != 1) {
				fprintf(stderr, "usage: %s\n", usage);
				exit(1);
			}
			continue;
		}
		if (strncmp(argv[i], "min_match=", 10) == 0) {
			if (sscanf(argv[i] + 10, "%d", &min_match) != 1) {
				fprintf(stderr, "usage: %s\n", usage);
				exit(1);
			}
			continue;
		}

		/* assume that this argument is a .pht filename */
		argfile[numargfile++] = i;
	}

	/* make sure that we have all the required column values */
	if ((x_col < 0) || (y_col < 0) || (mag_col < 0) || (err_col < 0)) {
		fprintf(stderr, "missing values for required columns x, y, mag, err \n");
		exit(1);
	}
	
	/* 
	 * we now know the matching radius, so we can calc its square
	 *   (will speed up later calculations)
	 */
	RAD2 = RADIUS*RADIUS;

	/* take care of command-line .pht files */
	for (i = 0; i < numargfile; i++) {
		strcpy(file, argv[argfile[i]]);
		do_pht(file, x_col, y_col, mag_col, err_col, jd_col, flags);
	}
	/* and now take care of listfiles */
	for (i = 0; i < argc; i++) {
		if (strncmp(argv[i], "list=", 5) == 0) {
			if (sscanf(argv[i] + 5, "%s", file) != 1) {
				fprintf(stderr, "usage: %s\n", usage);
				exit(1);
			}
			do_listfile(file, x_col, y_col, mag_col, err_col, jd_col, flags);
			continue;
		}
	}


	/* now that we've read in all the data, time to process it */
	proc_pht(flags, template, min_match);

	/* and, finally, write the output to the named file (or stdout, by
	   default) */
	write_output(outfile, flags);

	exit(0);
}

	/**********************************************************************
	 * PROCEDURE: do_listfile
	 *
	 * DESCRIPTION: Process each of the files listed in the given file,
	 *              one at a time.
	 *
	 * RETURNS:
	 *     0          if all goes well
	 *     1          if an error occurs
	 */

static int
do_listfile
	(
	char *file,           /* I: name of file with a list of data file names */
	int x_col,
	int y_col,
	int mag_col,
	int err_col,
	int jd_col,
	int flags
	)
{
	int count;
	char name[LINELEN], line[LINELEN];
	FILE *fp;

	if ((fp = fopen(file, "r")) == NULL) {
		fprintf(stderr, "do_listfile: can't open file %s\n", file);
		exit(-1);
	}
	/* first, just count the number of images we'll need */
	count = 0;
	while (fgets(line, LINELEN, fp) != NULL) {
		count++;
	}
	if (count >= MAX_IMAGE) {
		fprintf(stderr, "do_listfile: %d images is more than MAX = %d\n",
					count, MAX_IMAGE);
		exit(-1);
	}

	rewind(fp);
	while (fgets(line, LINELEN, fp) != NULL) {
		if (sscanf(line, "%s", name) != 1) {
			continue;
		}
		do_pht(name, x_col, y_col, mag_col, err_col, jd_col, flags);
	}
	fclose(fp);

	return(0);
}


	/******************************************************************
	 * PROCEDURE: do_pht
	 *
	 * DESCRIPTION: Walk through a file containing measurements of 
	 *              stars in a single image.  Read the data for each
	 *              star and place it into arrays in memory.
	 *              Increment the number of images scanned if 
	 *              we finish without a problem.
	 *
	 *              If the "
	 *
	 * RETURNS:
	 *     0        if all goes well
	 *     1        if an error occurs
	 */

static int
do_pht
	(
	char *file,             /* I: name of file with input magnitudes */
	int x_col,              /* I: read x positions from this column */
	int y_col,              /* I: read y positions from this column */
	int mag_col,            /* I: read magnitudes from this column */
	int err_col,            /* I: read mag uncertainties from this column */
	int jd_col,             /* I: read a time value from this column */
	                        /*      if -1, use image index number for time */
	int flags               /* I: if ISOLATE is set, we call */
	                        /*      remove_close_neighbors() after reading */
	)
{
	struct s_image *im;
	int ret;

	im = &(image_arr[num_images]);
	im->index = num_images;		/* this gets reset later */
	im->num_match = 0;
	im->include = 1;

	if ((ret = slurp_header(file, im)) == -1) {
		fprintf(stderr, "%s.do_pht: slurp_header returns -1 for file %s\n",
					progname, file);
		return(1);
	}

	if (slurp_pht_data(file, im, x_col, y_col, mag_col, err_col, jd_col) != 0) {
		fprintf(stderr, "%s.do_pht: slurp_pht_data returns non-zero for file %s\n",
					progname, file);
		return(1);
	}

	/* 
	 * Now that we have read in all the information from this file,
	 * check to see if the "jd_col" variable is still at its default
	 * value of -1 (indicating that there is no real time associated
	 * with each image).  If the jd_col value is NOT -1, then
	 * we can get the JD value by reading the "jd" field of any
	 * star in the image.  We'll pick the zero'th one.
	 */
	if (jd_col > -1) {
		im->jd = im->stars[0].jd;
	}

	/*
	 * If the ISOLATE bit of "flags" is set, the user wants to 
	 * discard (spurious) detections which lie within the matching
	 * radius of each other.  The "remove_close_neighbors()"
	 * function gets rid of all but the brightest stars in 
	 * overlapping groups.
	 */
	if (flags & ISOLATE) {
		if (remove_close_neighbors(im) != 0) {
			fprintf(stderr, "%s: remove_close_neighbors fails for image %d\n",
						progname, num_images);
			exit(1);
		}
	}


	num_images++;
	if (num_images >= MAX_IMAGE) {
		fprintf(stderr, "%s: %d images exceeds MAX_IMAGE %d \n",
					progname, num_images, MAX_IMAGE);
		exit(1);
	}

	return(0);
}


	/*********************************************************************
	 * PROCEDURE: slurp_header
	 *
	 * DESCRIPTION: Given the name of a file which contains measurements
	 *              from a single image, figure out some minimal properties
	 *              shared by all data in the image.  Place this information
	 *              in arrays for later use.
	 *
	 * RETURN:
	 *     0        if all goes well
	 *     1        if an error occurs
	 */


static int
slurp_header
	(
	char *file, 
	struct s_image *image
	)
{

	/* image name is same as file name */
	strncpy(image->name, file, STRLEN);
	image->name[STRLEN-1] = '\0';

	/* we don't use exposure time yet, so always set to 1.0 */
	image->exptime = 1.0;

	/* 
	 * for the moment, set "jd" to the image index number.
	 * Will later come up with a way to read a single JD value 
	 * to be shared for all star in the image from .... something;
	 * perhaps a special header line in file.
	 */
	image->jd = image->index;

	if (debug_flag) {
		printf("image %4d: %10s %6.1lf %14.7lf \n",
			image->index, image->name, image->exptime, image->jd);
	}
	return(0);
}
	


	/*********************************************************************
	 * PROCEDURE: slurp_pht_data
	 *
	 * DESCRIPTION: Given the name of a file which contains measurements
	 *              from a single image, read the data for each star,
	 *              one by one, and place into arrays for future reference.
	 *              in arrays for later use.
	 *
	 * RETURN:
	 *     0        if all goes well
	 *     1        if an error occurs
	 */

		
static int 
slurp_pht_data
	(
	char *file, 
	struct s_image *im,
	int x_col,
	int y_col,
	int mag_col,
	int err_col,
	int jd_col
	)
{
	int i, n, found;
	char line[LINELEN];
	FILE *fp;

	if ((im->num_star = num_in_file(file)) < 1) {
		im->stars = (struct s_star *) NULL;
		return(0);
	}
	else if ((im->stars = (struct s_star *) calloc(im->num_star, 
						sizeof(struct s_star))) == NULL) {
		fprintf(stderr, "%s.slurp_pht_data: can't calloc for stars in file %s\n",
					progname, file);
		exit(-1);
	}

	/* initialize values for each star in the file */
	for (i = 0; i < im->num_star; i++) {
		im->stars[i].image = im;
		im->stars[i].image_id = -1;
		im->stars[i].hash = (struct s_hash *) NULL;
		im->stars[i].id = -1;
		im->stars[i].jd = -1.0;
		im->stars[i].xc = -1.0;
		im->stars[i].yc = -1.0;
		im->stars[i].w3 = 1;
		im->stars[i].w4 = 0.0;
		im->stars[i].saturated = 0;
	}

	if ((fp = fopen(file, "r")) == NULL) {
		fprintf(stderr, "%s.slurp_pht_data: can't open file %s\n",
					progname, file);
		return(-1);
	}
	/* first, skip past all comment lines in the file */
	found = 0;
	while (fgets(line, LINELEN, fp) != NULL) {
		if (line[0] != COMMENT_CHAR) {
			found = 1;
			break;
		}
	}
	if (found == 0) {
		fclose(fp);
		fprintf(stderr, "%s.slurp_pht_data: file %s has no data?! \n",
					progname, file);
		return(-1);
	}

	/* now that we've reached the data, read each line in */
	n = 0;
	do { 
		if (read_pht_line(line, &(im->stars[n]), x_col, y_col,
						mag_col, err_col, jd_col) != 0) {
			fprintf(stderr, "%s.slurp_pht_data: bad data line %d\n", 
						progname, n);
		}
		
		/* 
		 * if this is identical to previous entry in file, assume that it
		 * is a duplicate entry -- and ignore it 
		 */
		if (identical(im, n)) {
			im->num_star--;
		}
		else {
			n++;
		}

		if (debug_flag) {
			print_star(&(im->stars[n-1]));
		}
	} while (fgets(line, LINELEN, fp) != NULL);

	fclose(fp);

	return(0);
}
		


	/*********************************************************************
	 * PROCEDURE: read_pht_line
	 *
	 * DESCRIPTION: Read in one line's worth of photometric data --
	 *              one star in one image. 
	 *
	 *              If there is a column to hold the Julian Date, or
	 *              any sort of time, then read that into the "jd"
	 *              field of the star structure.  If no such column
	 *              was provided on command-line, then leave the
	 *              JD field at its initial value of -1.
	 *
	 * RETURN:
	 *     0        if all goes well
	 *     1        if an error occurs
	 */
	
static int 
read_pht_line
	(
	char *line, 
   struct s_star *s,
	int x_col,
	int y_col,
	int mag_col,
	int err_col,
	int jd_col
	)
{
	int num_words;
	char col[MAX_COL][LINELEN];

	/* yeah, yeah, ugly */
	shAssert(19 < MAX_COL);

	/* first, we read in LOTS and LOTS of columns worth of data */
	num_words = sscanf(line, "%s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s",
		&(col[0][0]), &(col[1][0]), &(col[2][0]), &(col[3][0]), 
		&(col[4][0]), &(col[5][0]), &(col[6][0]), &(col[7][0]), 
		&(col[8][0]), &(col[9][0]), &(col[10][0]), &(col[11][0]), 
		&(col[12][0]), &(col[13][0]), &(col[14][0]), &(col[15][0]), 
		&(col[16][0]), &(col[17][0]), &(col[18][0]), &(col[19][0]));

	/* did we succeed in reading the required columns? */
	if ((x_col >= num_words) || (y_col >= num_words) || 
	    (mag_col >= num_words) || (err_col >= num_words)) {
		/* oops, not enough columns in input */
		fprintf(stderr, "%s.read_pht_line: bad line ..%s.. \n",
				progname, line);
		return(1);
	}
	
	/* now we read each required piece of information */
	if (sscanf(&(col[x_col][0]), "%lf", &(s->xc)) != 1) {
		fprintf(stderr, "%s.read_pht_line: bad x col val ..%s.. \n",
				progname, col[x_col]);
		return(1);
	}
	if (sscanf(&(col[y_col][0]), "%lf", &(s->yc)) != 1) {
		fprintf(stderr, "%s.read_pht_line: bad y col val ..%s.. \n",
				progname, col[y_col]);
		return(1);
	}
	if (sscanf(&(col[mag_col][0]), "%lf", &(s->mag)) != 1) {
		fprintf(stderr, "%s.read_pht_line: bad mag col val ..%s.. \n",
				progname, col[mag_col]);
		return(1);
	}
	if (sscanf(&(col[err_col][0]), "%lf", &(s->err)) != 1) {
		fprintf(stderr, "%s.read_pht_line: bad err col val ..%s.. \n",
				progname, col[err_col]);
		return(1);
	}

	/* check to see if we should try to get a time value from one column */
	if (jd_col >= 0) {
		if (jd_col >= num_words) {
			/* oops, not enough columns in input */
			fprintf(stderr, "%s.read_pht_line: not enough cols for JD col ..%d.. \n",
						progname, jd_col);
			return(1);
		}
		if (sscanf(&(col[jd_col][0]), "%lf", &(s->jd)) != 1) {
			fprintf(stderr, "%s.read_pht_line: bad JD col val ..%s.. \n",
				progname, col[jd_col]);
			return(1);
		}
	}

		
	/* if the uncertainty value is negative, assume star is saturated */
	/*    and re-set uncertainty to be positive */
	if (s->err < 0) {
		s->err = 0.0 - s->err;
		s->saturated = 1;
	}

	/* 
	 * if the size of the uncertainty is too small, replace it with
	 * the minimum permitted uncertainty.  Is especially important
	 * not to permit uncertainty = 0.0
	 */
	if (s->err < MIN_MAG_UNCERT) {
		s->err = MIN_MAG_UNCERT;
	}

	return(0);
}


	/**********************************************************************
	 * PROCEDURE: print_star
	 *
	 * DESCRIPTION: Print out info in the given star structure.
	 *
	 * RETURNS:
	 *    nothing
	 */

static void
print_star
	(
	struct s_star *s
	)
{
	struct s_image *im;

	if ((im = s->image) != NULL) {
		printf("%-10s %6.2lf %14.7lf ", im->name, im->exptime, im->jd);
	}
	printf("%10.5lf %10.5lf  %1d  ", s->xc, s->yc, s->saturated);
	printf("%7.4lf %7.4lf  ", s->mag, s->err);
	printf("\n");

}

	/**********************************************************************
	 * PROCEDURE: proc_pht
	 *
	 * DESCRIPTION: Process the data from a single image.  Use the
	 *              "template" image as a guide for star IDs.
	 *              We will shift star coords to match their corresponding
	 *              entries in the template image.
	 *
	 * RETURNS:
	 *    nothing
	 */

static void 
proc_pht
	(
	int flags, 
	char *template,
	int min_match          /* I: require an image to have this many matches */
	)
{

	if (flags & PRINT) {
		do_print();
	}

	if (do_match(template, min_match) != 0) {
		fprintf(stderr, "proc_pht: do_match returns with error.  Aborting\n");
		exit(1);
	}
	assign_ids();

	do_weights(flags);

}

	/**********************************************************************
	 * PROCEDURE: do_print
	 *
	 * DESCRIPTION: Print out information on all image and all stars
	 *              in all the images.  Used for testing.
	 *
	 * RETURNS:
	 *    nothing
	 */

static void
do_print
	(
	void
	)
{
	int i, j;

	for (i = 0; i < num_images; i++) {
		printf("           image %3d: %s\n", i, image_arr[i].name);
		for (j = 0; j < image_arr[i].num_star; j++) {
			print_star(&(image_arr[i].stars[j]));
		}
	}
}

	
		
	/**********************************************************************
	 * PROCEDURE: num_in_file
	 *
	 * DESCRIPTION: Count the number of stars in the file corresponding
	 *              to a single image.  We assume one star per line.
	 *
	 * RETURNS:
	 *    the number of stars
	 */

static int
num_in_file
	(
	char *file            /* data file with data on all stars in one image */
	)
{
	int i, found;
	char line[LINELEN];
	FILE *fp;

	if ((fp = fopen(file, "r")) == NULL) {
		fprintf(stderr, "%s.num_in_file: can't open file %s\n",
					progname, file);
		return(1);
	}

	/* first, skip past all comment lines in the file */
	found = 0;
	while (fgets(line, LINELEN, fp) != NULL) {
		if (line[0] != COMMENT_CHAR) {
			found = 1;
			break;
		}
	}
	if (found == 0) {
		fclose(fp);
		fprintf(stderr, "%s.num_in_file: file %s has no data?! \n",
					progname, file);
		return(1);
	}
	/* we have already found one data line, so start counting from 1 */
	i = 1;
	while (fgets(line, LINELEN, fp) != NULL) {
		i++;
	}
	fclose(fp);

	return(i);
}

	/**********************************************************************
	 * PROCEDURE: identical
	 *
	 * DESCRIPTION: Look to see if N'th star in the given image 
	 *              has exactly the same position as the previous star.
	 *              If so, we assume that the second entry is an erroneous
	 *              entry for the same star.  
	 *
	 *              Sometimes images have lots of double entries,
	 *              so this routine can be used to get rid of them.
	 *
	 * RETURNS:
	 *      0       if stars are NOT identical in position
	 *      1       if stars are identical in position
	 */



static int
identical
	(
	struct s_image *im, 
	int n
	)
{
	double xc, yc;

	if (n < 1) {
		return(0);
	}
	xc = fabs(im->stars[n].xc - im->stars[n - 1].xc);
	yc = fabs(im->stars[n].yc - im->stars[n - 1].yc);
	if ((xc < SAMEPOS) && (yc < SAMEPOS)) {
		return(1);
	}
	return(0);
}


