
	/* assign unique IDs to all stars, based on their aligned positions */

	/***********************************************************************
	 * Changes:
	 *
	 *  - fixed off-by-one error in "get_star_index".
	 *      MWR  12/14/2002
	 *
	 *  - big change to "assign_ids()" to avoid making errors when 
	 *      assigning measurements from one image to the master star list.
	 *      Original version caused later solvepht analysis to hang
	 *      under some circumstances.
	 * 
	 *      The new version enforces the rule that each detection in an
	 *      image can be assigned only to a single master star entry,
	 *      and each master star entry can be assigned only to a single
	 *      detection in each image.
	 *      MWR   1/26/2007
	 *
	 */


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <search.h>
#include "multipht.h"

#undef DEBUG
#undef DEBUG2

#define HIST_BINS    20			/* number of bins in histogram */

static struct s_hash *hashtab;		/* hash table array */
static int nhash = 0;				/* number of hash table entries */

static void sort_stars(struct s_image *im);
static int comp_by_xc(struct s_star *s1, struct s_star *s2);
struct s_hash * lookup(struct s_star *star);
static struct s_hash * looklist(struct s_star *star, int index);
static void hash_insert(struct s_star *star);
static void hash_add(struct s_star *star, struct s_hash *p);
static void add_id(struct s_hash *p);
static void init_hash(void);
static int is_a_match(struct s_star *s1, struct s_hash *p);
static int get_star_index(struct s_star *starp);
static int two_stars_match(struct s_star *s1, struct s_star *s2,
                           double *distance_squared);
static int find_closest_match(struct s_star *master_star, int image_index);
static int find_star_by_xc(int image_index, double xc);
static struct s_star * find_closest_matching_master(struct s_star *this_detection, 
                                         int image_index);
static int find_master_star_by_xc(double xc);
#ifdef DEBUG
static void print_hash(void);
static void print_by_id(void);
static int verify_table(void);
#endif

	/**************************************************************
	 * PROCEDURE: good_images
	 *
	 * DESCRIPTION: Assign unique IDs to each star in each image, 
	 *              based upon its position in the template coordinate system. 
	 *              Note that the stars should already be almost - but not
	 *              quite - sorted by their rows, 
	 *              so we have to sort them first.  
	 *
	 *              Modified this code to prevent multiple stars
	 *              in one image being assigned to the same entry in the
	 *              master star list.  We make two passes to assign IDs:
	 * 
	 *                  step 1:  walk through the master star list
	 *                              look for the _closest_ matching star in
	 *                                    the current image 
	 *                              if found, 
	 *                                    verify that no other item in the 
	 *                                        master star list is closer
	 *                                    if true, 
	 *                                        append the star to the list of 
	 *                                            instances in the
	 *                                            master star list
	 *                                        and mark star as "taken"
	 *
	 *                  step 2:  now walk through the list of stars in
	 *                           the current image
	 *                              if a star has been "taken", skip it
	 *                              otherwise, insert a new entry into the
	 *                                    master star list for this star
	 *
	 *              This procedure should guarantee that only one measurement
	 *              from any image can be assigned to a particular master
	 *              star ID entry, _and_ that no measurement is assigned
	 *              to more than a single master star ID entry.
	 *
	 *              As each star is examined, the 'id' field of 
	 *              its structure is set the master ID number for all 
	 *              stars in all images.  Moreover, the hash table entry 
	 *              for that ID has the new star added to its linked 
	 *              list of stars that share the ID, and has its 
	 *              'num' field incremented.  thus, we can go 
	 *              through all stars BY ID efficiently after having 
	 *              built up the table, and we can quickly find out how
	 *              many stars share the same ID 
	 *              (i.e. in how many images was each star found).
	 *
	 * RETURNS:
	 *     the total number of unique stars found 
	 */

int
assign_ids
	(
	void
	)
{
	int i, j;

	init_hash();

	for (i = 0; i < num_images; i++) {

		/* I'll refer to image "i" as "the current image" in code below */

		sort_stars(&(image_arr[i]));


		/* 
		 * Step 1:  Walk through the master star list (in the hash table).  
		 *          Look for the
		 *          _closest_ matching item in the current image.
		 *          If found, assign that item to its match in the
		 *          master star list.
		 */
		{
			int hash_i;
			int star_index;

			struct s_hash *hash_p;
			struct s_starp *starp_p;
			struct s_star *star_p, *matching_star_p;

#ifdef DEBUG
			printf(" about to walk through master star list \n");
			printf("    to look for matches in image %d \n", i);
#endif

			for (hash_i = 0; hash_i < nhash; hash_i++) {
				hash_p = &(hashtab[hash_i]);

				while (hash_p != NULL) {

#ifdef DEBUG
					printf(" hashtab[%5d] has ID %6d  %7.1f %7.1f  num %5d \n",
							hash_i, hash_p->id, hash_p->xc, hash_p->yc, hash_p->num);
#endif


					if ((starp_p = hash_p->starp) != NULL) {

						/* 
					 	 * This is the master star entry.  We are going to look
					 	 * for the _closest matching_ star in the current image.
					 	 */
						starp_p = hash_p->starp;
						star_p = starp_p->star;
#ifdef DEBUG
						printf(" master star ID %6d at %7.1f %7.1f  has %5d obs ... ",
									star_p->id, hash_p->xc, hash_p->yc, hash_p->num);
#endif

						star_index = find_closest_match(star_p, i);
						matching_star_p = &(image_arr[i].stars[star_index]);

						if (star_index != -1) {

							struct s_star *closest_master_star;

							/* 
							 * We found a match!  But is this the best match?
							 * That is, does another master star lie even closer
							 * to this detection?  If so, we should not assign
							 * the detection to the current master star, but wait
							 * to assign it to the other master star later.
							 */

							closest_master_star = 
									find_closest_matching_master(matching_star_p, i);

							if (closest_master_star == star_p) {

								/* Yes, is best.  Append to the master star's entry */
#ifdef DEBUG
								printf(" best match %5d \n", star_index);
#endif
								hash_add(matching_star_p, hash_p);

							}
							else {
								/* 
								 * nope, another master list entry is a better
								 * match to this measurement.  We'll assign it to
								 * that other entry later.
								 */
#ifdef DEBUG
								printf(" a better match to master ID %6d \n",
										closest_master_star->id);
#endif

							}
						}
						else {
							/* no star in current image matches.  Oh, well. */
#ifdef DEBUG
							printf(" no match \n");
#endif
						}
					
					}

					/* 
					 * If there are additional master star IDs hanging off 
					 * this element of the hash table, look for the closest
					 * matching star in the current image for each of them, too.
				 	 */
					hash_p = hash_p->next;

				}

			}
		}

		/* 
		 * Step 2:  Walk through the stars detected in this image.
		 *          Any star not marked as belonging to an ID in the
		 *          master star list must be novel; we'll insert a new
		 *          entry into the master star list for it.
		 *
		 *          It's easy to see which stars aren't yet marked
		 *          as belonging to a master star: the "id" value
		 *          is -1.
		 */

		for (j = 0; j < image_arr[i].num_star; j++) {

			struct s_star *image_star_p;

			image_star_p = &(image_arr[i].stars[j]);

			/* 
			 * has this detection already been assigned to a master star
			 * entry in the hash table? 
			 */
			if (image_star_p->id != -1) {
				/* yes, it has.  Skip it */
#ifdef DEBUG
				printf("assigned to existing hash   image %6d star %5d \n", i, j);
				verify_table();
#endif

			}
			else {
				/* 
				 * no, it has not.  We need to create a new entry in the
				 * master star list (hash table) for this item.
				 */
				hash_insert(image_star_p);
#ifdef DEBUG
				printf("adding a new hash entry for image %6d star %5d \n", i, j);
				verify_table();
#endif
			}

		}



	}

	if (debug_flag) {
#ifdef DEBUG
		print_by_id();
		print_hash();
		print_star_hist();
		print_image_hist();
#endif
	}

	return(num_ids);
}

	/**************************************************************
	 * PROCEDURE: sort_stars
	 *
	 * DESCRIPTION: Sort all the stars in the given image by 
	 *              their 'xc' value.
	 *
	 * RETURNS:
	 *     nothing
	 */


static void
sort_stars
	(
   struct s_image *im
	)
{
	qsort((char *) &(im->stars[0]), im->num_star, sizeof(struct s_star),
				(PFI) comp_by_xc);
}


	/**************************************************************
	 * PROCEDURE: comp_by_xc
	 *
	 * DESCRIPTION: Compare the 'xc' field of two star structures.
	 *
	 * RETURNS:
	 *     -1 if the first star has an 'xc' coord less than second
	 *      0                                     equal to 
	 *     +1                                     greater than 
	 */

static int
comp_by_xc
	(
	struct s_star *s1, 
	struct s_star *s2
	)
{
	if (s1->xc < s2->xc) {
		return(-1);
	}
	else if (s1->xc > s2->xc) {
		return(1);;
	}
	else {
		return(0);
	}
}
	

	/**************************************************************
	 * PROCEDURE: lookup
	 *
	 * DESCRIPTION: Lookup a star in the hash table, 
	 *              using its 'xc' coord as a key.
	 *
	 * RETURNS:
	 *      a pointer to the hash table entry, if found
	 *      NULL,                              if not found. 
	 */


struct s_hash *
lookup
	(
	struct s_star *star
	)
{
	int i, index;
	struct s_hash *found;

	index = get_star_index(star);

	/* look first in the linked list of stars with the same 'xc' value */
	if ((found = looklist(star, index)) != (struct s_hash *) NULL) {
		return(found);
	}
	
	/* now look in the linked lists of stars with progressively smaller xc
	   values, down to 'xc - 2*RAD'. */
	for (i = index - 1; (i >= 0) && (i >= index - (int)(2*RADIUS)); i--) {
		if ((found = looklist(star, i)) != (struct s_hash *) NULL) {
			return(found);
		}
	}

	/* now look in the linked lists of stars with progressively larger xc
	   values, up to 'xc + RAD'. */
	for (i = index + 1; (i < nhash) && (i <= index + (int)(2*RADIUS)); i++) {
		if ((found = looklist(star, i)) != (struct s_hash *) NULL) {
			return(found);
		}
	}

	return((struct s_hash *) NULL);
}


	/**************************************************************
	 * PROCEDURE: looklist
	 *
	 * DESCRIPTION: Look for a star in the given linked list of 
	 *              hash entries.  
	 *
	 * RETURNS:
	 *     a pointer to the has entry, if the star is found
	 *     NULL                        if the star is not found
	 */

static struct s_hash *
looklist
	(
	struct s_star *star,
	int index
	)
{
	struct s_hash *p;

	if ((index < 0) || (index >= nhash)) {
		return((struct s_hash *) NULL);
	}
	for (p = &(hashtab[index]); (p != NULL) && (p->id != -1); p = p->next) {
		if (is_a_match(star, p)) {
			return(p);
		}
	}

	return((struct s_hash *) NULL);
}



	/************************************************************************8
	 * PROCEDURE: verify_table
	 *
	 * DESCRIPTION: walk through the hash table, and all its entries,
	 *              checking the pointer values.
	 *
	 *              Used only in debugging.
	 *
	 * RETURNS:
	 *   0          if all goes well
	 *   1          if a problem is found
	 */

#ifdef DEBUG
static int
verify_table
	(
	void
	)
{
	int i, j;
	struct s_hash *hash_p;
	struct s_starp *starp_p;

	for (i = 0; i < nhash; i++) {
		hash_p = &(hashtab[i]);
		if (hash_p != NULL) {
			/* check all the stars with this ID */
			starp_p = hash_p->starp;
			for (j = 0; j < hash_p->num; j++) {
				if (starp_p == NULL) {
					printf("verify_table: hashtab[%d] star %d has null pointer\n",
							      i, j);
					return(1);
				}
				starp_p = starp_p->next;
			}
			if ((hash_p->num > 0) && (starp_p != NULL)) {
				printf("verify_table: hashtab[%d] starlist missing NULL starp->next\n", 
								i);
			}
			/* now check all other hash entries with identical has value */
			while (hash_p->next != NULL) {
				hash_p = hash_p->next;
				/* check _this_ entry's stars */
				starp_p = hash_p->starp;
				for (j = 0; j < hash_p->num; j++) {
					if (starp_p == NULL) {
						printf("verify_table: hashtab[%d] XX star %d has null pointer\n",
								      i, j);
						return(1);
					}
					starp_p = starp_p->next;
				}
			}
		}
	}

	return(0);
}
#endif


	/**************************************************************
	 * PROCEDURE: hash_insert
	 *
	 * DESCRIPTION: Insert a star into the hash table
	 *
	 * RETURNS:
	 *     nothing
	 */

static void
hash_insert
	(
	struct s_star *star
	)
{
	int index, flag;
	struct s_hash *p, *q;

	index = get_star_index(star);
	if (hashtab[index].id == -1) {
		flag = 1;
	}
	else {
		flag = 0;
	}

	if (flag == 1) {
		p = &(hashtab[index]);
	}
	else {
		if ((p = (struct s_hash *) malloc(sizeof(struct s_hash))) == NULL) {
			fprintf(stderr, "hash_insert: can't malloc for new entry, num_ids=%d\n",
					num_ids);
			exit(-1);
		}
		p->num = 0;
	}
	p->xc = star->xc;
	p->yc = star->yc;
	p->id = num_ids++;
	star->id = p->id;
	star->hash = p;
	p->num++;
	if ((p->starp = (struct s_starp *) malloc(sizeof(struct s_starp))) == NULL) {
		fprintf(stderr, "hash_insert: can't malloc for starp, num_ids=%d\n", 
							num_ids);
		exit(-1);
	}
	p->starp->star = star;
	p->starp->next = (struct s_starp *) NULL;
	p->next = (struct s_hash *) NULL;
	p->include = 1;

	if (flag == 0) {
		/* walk down the linked list of hash structures to the end, then
		   put the new structure at the end of the list */	
		for (q = &(hashtab[index]); q->next != NULL; q = q->next) {
			;
		}
		q->next = p;
	}

	add_id(p);

#ifdef DEBUG
	if (p->starp->next == NULL) {
		printf("hash_insert: after inserting index %6d, starp->next is NULL\n",
				index);
	}
	else {
		printf("hash_insert: after inserting index %6d, starp->next bad??\n",
				index);
	}
#endif

}
	

	/**************************************************************
	 * PROCEDURE: hash_add
	 *
	 * DESCRIPTION: Add a star to the hash entry that matches it.  
	 *              Modify the hash entry's 'xc' and 'yc' values so 
	 *              that they are the average of all matching star values. 
	 *
	 * RETURNS:
	 *     nothing
	 */

static void
hash_add
	(
	struct s_star *star,
	struct s_hash *p
	)
{
	double dn;
	struct s_starp *sp, *q;

#ifdef DEBUG
	printf("hash_add: about to add star for hash %d \n", p->id);
#endif

	if ((sp = (struct s_starp *) malloc(sizeof(struct s_starp))) == NULL) {
		fprintf(stderr, 
                  "hash_add: can't malloc for starp, num_ids=%d\n", num_ids);
		exit(-1);
	}
	sp->star = star;
	sp->next = (struct s_starp *) NULL;
	p->num++;
	star->id = p->id;
	star->hash = p;
	for (q = p->starp; (q != NULL) && (q->next != NULL); q = q->next) {
		;
	}
	q->next = sp;
	
	/* now modify the hash entry's xc and yc values */
	dn = p->num;
	p->xc = ((p->xc*dn) + star->xc)/(dn + 1.0);
	p->yc = ((p->yc*dn) + star->yc)/(dn + 1.0);
}	



	/**************************************************************
	 * PROCEDURE: add_id
	 *
	 * DESCRIPTION: Add a new entry to the linked list of hash entries 
	 *              which are in order by their ID numbers.  
	 *              The head of this list is 'star_list', and by walking 
	 *              down it, you can walk through all hash entries 
	 *              (and, hence, stars) by their ID number 
	 *
	 * RETURNS:
	 *     nothing
	 */

static void
add_id
	(
	struct s_hash *p
	)
{
	static int first = 0;
	static struct s_hashp *last;
	struct s_hashp *hp;

	if ((hp = (struct s_hashp *) malloc(sizeof(struct s_hashp))) == NULL) {
		fprintf(stderr, "add_id: can't malloc for hashp, num_ids=%d\n", num_ids);
		exit(-1);
	}
	hp->hash = p;
	hp->next = (struct s_hashp *) NULL;
	if (first == 0) {
		first = 1;
		star_list = hp;
		last = star_list;
	}
	else {	
		last->next = hp;
		last = last->next;
	}
}

	/***********************************************************************
	 * PROCEDURE: init_hash
	 *
	 * DESCRIPTION: Initialize the hash table we use for assigning 
	 *              IDs to individual measurements.  
	 *
	 *              It's probably better to read the maximum size
	 *              of (x,y) coord system from command line than to
	 *              handle as #define, but change later.
	 *
	 * RETURNS:
	 *   nothing
	 */

static void
init_hash
	(
	void
	)
{
	int i;

	nhash = MAX_COORD;

	/* allocate an array large enough to fit 'rowmax' rows */
	if ((hashtab = (struct s_hash *) calloc(nhash, sizeof(struct s_hash))) == NULL) {
		fprintf(stderr, "init_hash: can't calloc for %d hash entries \n", 
							nhash);
		exit(-1);
	}

	/* initialize the values in the table */
	for (i = 0; i < nhash; i++) {
		hashtab[i].xc = -1.0;
		hashtab[i].yc = -1.0;
		hashtab[i].id = -1;
		hashtab[i].num = 0;
		hashtab[i].starp = (struct s_starp *) NULL;
		hashtab[i].next = (struct s_hash *) NULL;
		hashtab[i].include = 0;
	}
}

	/**************************************************************
	 * PROCEDURE: is_a_match
	 *
	 * DESCRIPTION: Compare the distance between the star 
	 *              and the hash-entry-centroid to the critical distance 
	 *              for a match; if the distance is less, 
	 *              we have a match.
	 *
	 * RETURNS:
	 *     0        if this star doesn't match the given hash entry
	 *     1        if this is a match
	 */


static int
is_a_match
	(
	struct s_star *s1,
	struct s_hash *p
	)
{
	double dr, dc;

	dr = (s1->xc - p->xc);
	dc = (s1->yc - p->yc);
	if ((dr*dr + dc*dc) <= RAD2) {
		return(1);
	}
	else {
		return(0);
	}
} 

	/**************************************************************
	 * PROCEDURE: print_hash
	 *
	 * DESCRIPTION: Print out the entire hash table - 
	 *              for debugging purposes only
	 *
	 * RETURNS:
	 *     nothing
	 */


#ifdef DEBUG
static void
print_hash
	(
	void
	)
{
	int i, n, nn;
	struct s_hash *p;
	struct s_starp *s;

	for (i = 0; i < nhash; i++) {
		if (hashtab[i].id == -1) {
			printf("hashtab[%4d]   ", i);
			printf("empty \n");
		}
		else {
			n = 0;
			printf("hashtab[%4d]   ", i);
			p = &(hashtab[i]);
			while (p != NULL) {
				if (n == 0) {
					printf("entry %3d: ID=%5d  num=%3d  xc=%6.2lf yc=%6.2lf inc=%d\n", 
							n, p->id, p->num, p->xc, p->yc, p->include);
				}
				else {
					printf("%15s entry %3d: ID=%5d  num=%3d  xc=%6.2lf yc=%6.2lf inc=%d\n", 
							" ", n, p->id, p->num, p->xc, p->yc, p->include);
				}
				nn = 0;
				s = p->starp;
				while (s != NULL) {
					printf("%20s starp=%d  xc=%6.2lf yc=%6.2lf reqid=%s\n",
								" ", nn, s->star->xc, s->star->yc, 
								s->star->image->name);
					nn++;
					s = s->next;
				}
				n++;
				p = p->next;
			}
		}
	}
}


	/**************************************************************
	 * PROCEDURE: print_by_id
	 *
	 * DESCRIPTION: Print out a nice list of stars by their ID numbers 
	 *
	 * RETURNS:
	 *     nothing
	 */

static void
print_by_id
	(
	void
	)
{
	int nn;
	struct s_hashp *hp;
	struct s_hash *p;
	struct s_starp *s;

	for (hp = star_list; hp != NULL; hp = hp->next) {
		p = hp->hash;
		printf("ID=%5d  num=%3d  xc=%6.2lf yc=%6.2lf inc=%d\n", 
						p->id, p->num, p->xc, p->yc, p->include);
		nn = 0;
		s = p->starp;
		while (s != NULL) {
			printf("%9s starp=%d  xc=%6.2lf yc=%6.2lf w3=%d w4=%-8.1lf  reqid=%s\n", 
						" ", nn, s->star->xc, s->star->yc, s->star->w3,
						s->star->w4, s->star->image->name);
			printf(" \n");
			nn++;
			s = s->next;
		}
	}

}

#endif


	/**************************************************************
	 * PROCEDURE: print_star_hist
	 *
	 * DESCRIPTION: Print out a histogram of the number of 
	 *              images in which each star appears.  The user can
	 *              use this to decide which stars ought to be included
	 *              in the solution.
	 *
	 * RETURNS:
	 *     nothing
	 */

void
print_star_hist()
{
	int i, div, hist[HIST_BINS], max;
	int index;
	struct s_hashp *hp;
	struct s_hash *p;

	for (i = 0; i < HIST_BINS; i++)
		hist[i] = 0;

	max = 0;
	for (hp = star_list; hp != NULL; hp = hp->next) {
		p = hp->hash;
		if (p->num > max) {
			max = p->num;
		}
	}
	div = ((max - 1)/HIST_BINS) + 1;
	for (hp = star_list; hp != NULL; hp = hp->next) {
		p = hp->hash;
		index = (p->num) / div;
		if (index >= HIST_BINS) {
			index = HIST_BINS - 1;
		}
		hist[index]++;
	}
	for (i = 0; i < HIST_BINS; i++) {
		printf("%5d  %3d\n", (int) ((i+0.5)*div), hist[i]);
	}

}


	/**************************************************************
	 * PROCEDURE: print_image_hist
	 *
	 * DESCRIPTION: Print out a histogram of the number of images 
	 *              with at least N stars, vs. N 
	 *
	 * RETURNS:
	 *     nothing
	 */

void
print_image_hist()
{
	int i, div, hist[HIST_BINS], max;
	int index;

	for (i = 0; i < HIST_BINS; i++)
		hist[i] = 0;

	max = 0;
	for (i = 0; i < num_images; i++) {
		if (image_arr[i].num_star > max) {
			max = image_arr[i].num_star;
		}
	}
	div = ((max - 1)/HIST_BINS) + 1;
	for (i = 0; i < num_images; i++) {
		index = (image_arr[i].num_star / div);
		if (index >= HIST_BINS) {
			index = HIST_BINS - 1;
		}
		hist[index]++;
	}
	for (i = 0; i < HIST_BINS; i++) {
		printf("%5d  %3d\n", (int) ((i+0.5)*div), hist[i]);
	}

}


	/**********************************************************************
	 * PROCEDURE: get_star_index
	 *
	 * DESCRIPTION: given a STAR structure, use its "xc" value to
	 *              determine an "index" into the hash table.
	 *              We simply set
	 *                     index = xc
	 *              except that we make sure the index is within
	 *              legal ranges (i.e. >=0 and < nhash).
	 *
	 * RETURNS:
	 *     a hash index
	 *     0                   if we're handed NULL pointer
	 */

static int 
get_star_index
	(
	struct s_star *starp          /* I: pointer to STAR structure */
	)
{
	int index;

	if (starp == NULL) {
		fprintf(stderr, "get_star_index: handed NULL starp?  Return 0\n");
		return(0);
	}
	index = floor(starp->xc);
	if (index < 0) {
		index = 0;
	} 
	if (index >= nhash) {
		index = nhash - 1;
	}

	return(index);

}




	/**********************************************************************
	 * PROCEDURE: two_stars_match
	 *
	 * DESCRIPTION: Given two s_star structures, check to see if the
	 *              are within the critical matching radius.  
	 *              That critical value must already have been 
	 *              computed and set in the global variable RAD2
	 *              (which actually contains the _square_ of the
	 *              critical radius).
	 *
	 *              We check once to ensure that RAD2 has been set
	 *              to some positive value.
	 *
	 *              Place the square of the distance between them 
	 *              into the final argument, in case the calling
	 *              function is looking for the closest match ...
	 *
	 * RETURNS:
	 *              1        if the stars are within critical radius
	 *              0        if not
	 */

static int 
two_stars_match
	(
	struct s_star *s1,
	struct s_star *s2,
	double *distance_squared    /* O: distance between stars */
	)
{
	static int first = 0;
	double dx, dy, rad2;

	if (first == 0) {
		first = 1;
		if (RAD2 <= 0) {
			fprintf(stderr, "is_a_match: RAD2 is %lf, should be positive! \n",
						RAD2);
			exit(1);
		}
	}

	dx = (s1->xc - s2->xc);
	dy = (s1->yc - s2->yc);
	rad2 = dx*dx + dy*dy;
	*distance_squared = rad2;

	if (rad2 <= RAD2) {
		return(1);
	}
	else {
		return(0);
	}
}



	/**********************************************************************
	 * PROCEDURE: find_closest_match
	 *
	 * DESCRIPTION: We are given one star taken from the master hash table,
	 *              and asked to find the single closest matching star
	 *              in a given image (if one exists at all).
	 *
	 *              If there are multiple entries in the current image
	 *              which fall within the matching radius of the
	 *              given star, pick only the closest one.
	 *
	 * RETURNS:
	 *              index of closest star in current image     if found
	 *              -1                                         if no match
	 */

static int 
find_closest_match
	(
	struct s_star *master_star,
	int image_index
	)
{
	int star_i;
	int start_index, end_index;
	int closest_match_index;
	double min_distance_sq;
	struct s_star *this_star;

	/* 
	 * Simple but slow method would be to walk through the entire
	 * set of all stars in the current image, comparing each one
	 * to the master_star.  Faster but more complex method would
	 * involving using the fact that stars in the current image
	 * have been sorted by "xc" value, so we can quickly narrow
	 * down the range of stars to check.
	 */

	/* 
	 * Let's try fast but tricky method.  We know that any matching
	 * star must have an "xc" coordinate within the range
	 *
	 *    master_star->xc - RADIUS     to   master_star->xc + RADIUS
	 *
	 * We want to find two 
	 * places in the sorted list of stars in the current image:
	 * the star with an "xc" value less than (master_star->xc-RADIUS),
	 * and the star with "xc" greater than (master_star->xc+RADIUS).
	 *
	 * The "find_star_by_xc()" function looks for the star closest
	 * to the given position; to ensure that we are _less than_ 
	 * the value in the first case, and _greater than_ the value
	 * in the second case, we simply move outwards in index value
	 * (unless we are already at the boundaries of the array).
	 */
	start_index = find_star_by_xc(image_index, master_star->xc - RADIUS);
	if (start_index > 0) {
		start_index--;
	}
	end_index = find_star_by_xc(image_index, master_star->xc + RADIUS);
	if (end_index < (image_arr[image_index].num_star - 1) ) {
		end_index++;
	}

	/* 
	 * okay, all we have to do is check for matches within the 
	 * subset of all stars in the current image with index i between
	 *
	 *      start_index  <=  i  <=  end_index
	 * 
	 * If there are hundreds of stars in each image, this notion
	 * of restricting the search to a subset could speed things up.
	 */
	min_distance_sq = 1e99;
	closest_match_index = -1;
	for (star_i = start_index; star_i <= end_index; star_i++) {
		double distance_sq;

		this_star = &(image_arr[image_index].stars[star_i]);
		if (two_stars_match(master_star, this_star, &distance_sq) == 1) {
			if (distance_sq < min_distance_sq) {
				min_distance_sq = distance_sq;
				closest_match_index = star_i;
#ifdef DEBUG2
				printf("  star %5d yields distsq %7.1f  best \n",
						star_i, distance_sq);
#endif
			}
		}

	}


#ifdef DEBUG
	printf("   min_distance_sq = %9.4e   best is index %6d \n",
			min_distance_sq, closest_match_index);
#endif
	return(closest_match_index);
}



	/**********************************************************************
	 * PROCEDURE: find_star_by_xc
	 *
	 * DESCRIPTION: Given an image index, and an "xc" position,
	 *              find the star in that image which is closest to the
	 *              "xc" position.  
	 *
	 *              The stars within the image must have been already
	 *              sorted by their "xc" values.  We use a binary
	 *              search to find the closest star quickly.
	 *
	 * RETURNS:
	 *              index of closest star in current image
	 */

static int 
find_star_by_xc
	(
	int image_index,           /* I: index of image in image_arr[] */
	double xc                  /* I: we seek star closest to this position */
	)
{
	int top, bottom, mid;

	if (image_index < 0) {
		fprintf(stderr, "find_star_by_xc: given negative image index %d ?!\n",
					image_index);
		exit(1);
	}

#ifdef DEBUG2
	printf("\n");
	printf("  find_star_by_xc: looking for xc %7.1f \n", xc);
#endif
	
	top = 0;
	bottom = image_arr[image_index].num_star - 1;

	while (bottom - top >= 2) {
		mid = (top + bottom)/2;
#ifdef DEBUG2
		printf(" array[%5d] %7.1f  array[%5d] %7.1f  array[%5d] %7.1f \n",
				top, image_arr[image_index].stars[top].xc,
				mid, image_arr[image_index].stars[mid].xc,
				bottom, image_arr[image_index].stars[bottom].xc);
#endif
		if (image_arr[image_index].stars[mid].xc < xc) {
			top = mid;
		}
		else {
			bottom = mid;
		}
	}
#ifdef DEBUG2
		printf(" array[%5d] %7.1f                        array[%5d] %7.1f \n",
				top, image_arr[image_index].stars[top].xc,
				bottom, image_arr[image_index].stars[bottom].xc);
#endif

	/* 
	 * when we reach this point, the item we seek is either "top"
	 * or "bottom" in the array 
	 */
	if (image_arr[image_index].stars[top].xc <= xc) {
#ifdef DEBUG2
		printf(" returning bottom array[%5d] %7.1f \n",
					bottom, image_arr[image_index].stars[bottom].xc);
#endif
		return(bottom);
	}
	else {
#ifdef DEBUG2
		printf(" returning top    array[%5d] %7.1f \n",
					top, image_arr[image_index].stars[top].xc);
#endif
		return(top);
	}
	
}


	/**********************************************************************
	 * PROCEDURE: find_closest_matching_master
	 *
	 * DESCRIPTION: We are given one detection in the current image,
	 *              and asked to find the closest star in the 
	 *              master star list (hash table).
	 *
	 * RETURNS:
	 *              pointer to closest master star     if one exists
	 *              NULL                               if no such star
	 */

static struct s_star * 
find_closest_matching_master
	(
	struct s_star *this_detection,
	int image_index
	)
{
	int hash_i;
	int start_index, end_index;
	double min_distance_sq;
	struct s_star *closest_master_star = NULL;


	/* 
	 * Stars in the master list are sorted (roughly) by their
	 * "xc" coordinates.  We can quickly find a small subset
	 * of the entire master list which must include any matching
	 * master stars; then we can walk through that subset.
	 *
	 * We know that any matching
	 * star must have an "xc" coordinate within the range
	 *
	 *    this_detection->xc - RADIUS    to   this_detection->xc + RADIUS
	 *
	 * We want to find two 
	 * places in the master list of stars: 
	 * the star with an "xc" value less than (this_detection->xc-RADIUS),
	 * and the star with "xc" greater than (this_detection->xc+RADIUS).
	 *
	 * The "find_master_star_by_xc()" function looks for the entry
	 * in the hash table which has roughly the same "xc" value.
	 * To ensure that we are _less than_ 
	 * the value in the first case, and _greater than_ the value
	 * in the second case, we simply move outwards in index value
	 * (unless we are already at the boundaries of the array).
	 */
	start_index = find_master_star_by_xc(this_detection->xc - RADIUS);
	if (start_index > 0) {
		start_index--;
	}
	end_index = find_master_star_by_xc(this_detection->xc + RADIUS);
	if (end_index < nhash - 1) {
		end_index++;
	}

	/* 
	 * okay, all we have to do is check for matches within the 
	 * subset of all master stars with index i in hashtab between
	 *
	 *      start_index  <=  i  <=  end_index
	 * 
	 * If there are hundreds of stars in each image, this notion
	 * of restricting the search to a subset could speed things up.
	 */
	min_distance_sq = 1e99;
	for (hash_i = start_index; hash_i <= end_index; hash_i++) {

		double distance_sq;
		struct s_star *master_star;
		struct s_hash *hash_p;

		/* 
		 * Each entry in hashtab could contain a linked list,
		 * and we have to walk down though all such linked lists
		 * if we encounter them, checking all the master stars
		 * on the lists.
		 */
		hash_p = &(hashtab[hash_i]);

		while (hash_p != NULL) {

			if (hash_p->starp != NULL) {
				/* check the master star of this hash struct */
				master_star = hash_p->starp->star;
				if (two_stars_match(master_star, this_detection, &distance_sq) == 1) {
					if (distance_sq < min_distance_sq) {
						min_distance_sq = distance_sq;
						if (distance_sq < RAD2) {
							closest_master_star = master_star;
#ifdef DEBUG2
							printf("  hash %6d ID  %6d yields distsq %7.1f  best \n",
								hash_i, master_star->id, distance_sq);
#endif
						}
					}
				}
			}

			/* walk down linked list of hash structs */
			hash_p = hash_p->next;
		}

	}

#ifdef DEBUG
	if (closest_master_star == NULL) {
		printf("   min_distance_sq = %9.4e   no match \n",
			min_distance_sq);
	}
	else {
		printf("   min_distance_sq = %9.4e   closest master ID %6d \n",
			min_distance_sq, closest_master_star->id);
	}
#endif

	return(closest_master_star);
}


	/**********************************************************************
	 * PROCEDURE: find_master_star_by_xc
	 *
	 * DESCRIPTION: Given the "xc" position of a star,
	 *              find the entry in the master star list 
	 *              (hash table) which has an "xc" value closest
	 *              to the given position.
	 *
	 *              The hash table itself is arranged by "xc" 
	 *              in ascending order, so this is easy.
	 *              Now, each entry may contain a linked list of 
	 *              master stars with similar "xc" values, but 
	 *              we don't bother checking them.  We only find 
	 *              the index in the hash table which
	 *              gets us into the neighborhood of the given "xc".
	 *
	 * RETURNS:
	 *              index into hash table of master star with similar xc
	 */

static int 
find_master_star_by_xc
	(
	double xc              /* I: we seek master star closest to this position */
	)
{
	int index;


	/* simple method, assumes hashtab[] index is simply "xc" coord */

	index = floor(xc + 0.5);
	if (index < 0) {
		index = 0;
	}
	if (index >= nhash) {
		index = nhash - 1;
	}
	return(index);

}

