

	/*********************************************************************
	 * These routines set up a series of simultaneous equations which 
	 * correspond to the photometric solution of a large number
	 * of observations.  We set up a big matrix equation of the form
	 *
	 *              A  *  x  =   b
	 *
	 * where 'A' is a matrix, 'b' is a vector with the observed values of
	 * instr. mags and weights, and 'x' is the unknown vector of
	 * magnitude adjustments that minimizes the sums of the squares
	 * of the residuals from the mean. 
	 *
	 * Then, we call some Numerical Recipes code is used to solve it.
	 *
	 * MWR 10/18/2003 
	 *
	 *
	 * Added FORCE_STAR_MAG option, which may help solution under
	 *   some circumstances.
	 * Added a shift to the image adjustment values so that the
	 *   input (raw) mags and output (true) mags do obey the relationship
	 *          true_mag   +   image_adj   =   raw_mag
	 *
	 * MWR 10/28/2003
	 *
	 * Modified code to include terms for zero-point offsets of each
	 * filterchip.  That means that we now solve for magnitude of each
	 * star, adjustment of each image, and offset of each chip.
	 *
	 * MWR 3/23/2004
	 *
	 * Fixed huge error in "zero_vector()" which actually prevented
	 * any elements from being set to zero (!).
	 * MWR 3/25/2004
	 *
	 * Commented out the "calc_model_mag()" function, since it  
	 * isn't used any more in this simplified version.
	 * MWR 11/8/2004
	 *
	 * Replaced all the Numerical Recipes routines with equivalents
	 * from the GNU Scientific Library.  This means vector and matrix
	 * allocation and de-allocation, plus functions to do the SVD
	 * analysis of the matrix and then apply the result to find the
	 * solution vector.
	 * The big pain was changing all the old 1-index code to adhere
	 * to the GSL 0-index convention for vectors and matrices.
	 * MWR 12/13/2004
	 *
	 *
	 * Fixed bug in "find_sigm()".  Code which used neighboring stars
	 *     to estimate uncertainty did not check to make sure that
	 *     the neighboring star's uncertainty was valid, so some values
	 *     of NODATA = 99.0 were being included in calculations.
	 * MWR 4/8/2006
	 *
	 * More bugs in "find_sigm()".  Not really a bug: in sparse fields,
	 *     there may be no other stars in the field within MAG_MATCH 
	 *     mag of the variable star -- which means no estimation of 
	 *     uncertainty based on similar stars is possible.  Added some
	 *     DEBUG statements to find the problem.  The solution is simply
	 *     to increase the MAG_MATCH value from default of 0.30 mag
	 *     to some larger value in sparse fields.  
	 * MWR 4/12/2006
	 *
	 * Added "mag_match" arg to find_sigm(), which can be set via
	 *     optional command-line arguement.  Replaces the 
	 *     fixed MAG_MATCH value.
	 * MWR 5/9/2006
	 *   
	 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "solvepht.h"
#include "gsl/gsl_vector.h"
#include "gsl/gsl_matrix.h"
#include "gsl/gsl_linalg.h"

#define TOLERANCE    1.0e-7
#define TINY         (2.0*TOLERANCE) 
#define MAG_MATCH    0.60		/* used by find_sigm() */

#undef DEBUG


  /* 
	* if this is defined, then we set matrix elements corresponding 
	*   to the first star in the list to be zero.   That means that
	*   the first star's mag in the output should be zero, too.
	*   After solving the matrix equation, we need to go back and 
	*   set the output vector x element for this first star to be
	*   zero.  And _then_ we can run the normalize() function,
	*   which will shift the output magnitude values so that the
	*   BRIGHTEST star has average mag zero, rather than the first
	*   star in the list.
	*
	* Setting this option may cause the matrix solution to behave
	*   better.  In some cases, however, the code doesn't need us
	*   to take this step; it returns a reasonable solution on its
	*   own.
	*/
#define FORCE_STAR_MAG 



static void init_matrix(int ee, int ss);
static void calc_em(int m, int n);
static void calc_mags(int m, int n);
static void calc_sigmas(int m, int n, double mag_match);
static void zero_vector(gsl_vector *p, int m, double tolerance);
static void print_vector(gsl_vector *p, int m);
static void print_matrix(gsl_matrix *p, int m, int n);
static void fill_bvec(gsl_vector *b, int m, int n);
static void read_data(int ee, int ss);
static void fill_matrix(gsl_matrix *p, int ee, int ss);
static void copy_matrix(gsl_matrix *p, gsl_matrix *q, int m, int n);
static void normalize(int m, int n, double value);
static void find_means(int m, int n);
static double find_sigm(int n, int e, int s, double mag_match);
static void print_curves(int m, int n);
static void print_sigmas(int m, int n);
static double calc_model_correction(int num_images, int num_stars, 
		                        int image_index, int star_index);
#if 0
static double calc_model_mag(int num_images, int num_stars, 
		                        int image_index, int star_index);
#endif


	/* these are inputs */
static gsl_matrix *mag;          /* ee by ss matrix of instr. mags */
static gsl_matrix *weight;			/* ee by ss matrix of weights */

	/* these are computational values */
static gsl_matrix *a;				/* ee+ss+cc by ee+ss+cc matrix; A above */
static gsl_matrix *u;				/* ee+ss+cc by ee+ss+cc: part 1 of the SVD output */
static gsl_vector *w;				/* ee+ss+cc vector: part 2 of SVD output */
static gsl_matrix *v;				/* ee+ss+cc by ee+ss+cc: part 3 of SVD output */
static gsl_vector *b;				/* ee+ss+cc vector: 'b' above */
static gsl_vector *x;				/* ee+ss+cc vector: 'x' above, the 'solution' */
	/* and these are output results */
static gsl_vector *em;				/* ee vector: mean correction for image e */
static gsl_vector *mag0;				/* ss vector: mean mag of star s */
static gsl_vector *sigm0;			/* ss vector: sigma of mean value of star s */
static gsl_matrix *Mes;				/* ee by ss matrix: solved mag of star s */
                   				/*    on image e */
static gsl_matrix *sigMes;			/* ee by ss matrix: total sigma of */
                  				/*    solution's value of star s on image e */
static gsl_vector *sigem;			/* ee vector: sigma of exp corr. of image e */
static gsl_vector *sigemn;			/* ee vector: sigem/# stars in image e */



	/*********************************************************************
	 * PROCEDURE: solve_matrix
	 *
	 * DESCRIPTION: set up all the equations and solve them.
	 *              If the debug flag is set, print out some results.
	 *
	 * RETURNS:
	 *    nothing
	 */

void
solve_matrix
	(
	double zero_point,      /* I: set brightest star to this mag on output */
	double mag_match        /* I: used to find stars "similar" enough to */
	                        /*       a variable to estimate uncert in its */
	                        /*       individual mag measurements */
	)
{
	int i;
	int ee, ss;
	double tolerance;
	gsl_vector *work;

	tolerance = TOLERANCE;
	ee = num_images;
	ss = num_stars;
	init_matrix(ee, ss);

	read_data(ee, ss);
	fill_matrix(a, ee, ss);
	if (debug_flag) {
		printf("\n This is the matrix \n");
		print_matrix(a, ee + ss, ee + ss);
	}
	copy_matrix(a, u, ee + ss, ee + ss);

	fill_bvec(b, ee, ss);
	if (debug_flag) {
		printf(" This is the B vector \n");
		print_vector(b, ee + ss);
	}

#if 0
	svdcmp(u, ee + ss, ee + ss, w, v);
#endif
	if ((work = gsl_vector_calloc(ee + ss)) == NULL) {
		fprintf(stderr, "init_matrix: can't allocate for vector work\n");
		exit(-1);
	}
	if (gsl_linalg_SV_decomp(u, v, w, work) != 0) {
		fprintf(stderr, "solve_matrix: gsl_linalg_SV_decomp returns with error\n");
		exit(1);
	}

	if (debug_flag) {
		printf("this is the svd matrix U \n");
		print_matrix(u, ee + ss, ee + ss);
		printf("this is the vector of w(j) - which could need zero'ing \n");
		print_vector(w, ee + ss);
	}

	zero_vector(w, ee + ss, tolerance);
	if (debug_flag) {
		printf("this is the vector w(j) after zeroing tiny elements\n");
		print_vector(w, ee + ss);
	}

#if 0
	svbksb(u, w, v, ee + ss, ee + ss, b, x);
#endif
   if (gsl_linalg_SV_solve(u, v, w, b, x) != 0) {
		fprintf(stderr, "solve_matrix: gsl_linalg_SV_solve returns with error\n");
		exit(1);
	}

	if (debug_flag) {
		printf("\nand this is the vector of x - the solution\n");
		print_vector(x, ee + ss);
	}

#ifdef FORCE_STAR_MAG
	/* 
	 * the first star's mag in the solution vector x has been unconstrained
	 *   if we chose the FORCE_STAR_MAG option.  We need to set it equal
	 *   to zero explicitly before we go any further.
	 */
	gsl_vector_set(x, ee, 0.0);
#endif

	calc_em(ee, ss);
	calc_mags(ee, ss);
	find_means(ee, ss);
	normalize(ee, ss, zero_point);
	calc_sigmas(ee, ss, mag_match);

	if (debug_flag) {
		print_curves(ee, ss);
		print_sigmas(ee, ss);
	}

	/* 
	 * set the "star->mag0" and "star->sigma" fields to the 
	 *   mean magnitude and scatter-around-mean-magnitude
	 *   in the solution.
	 */
	for (i = 0; i < num_stars; i++) {
		if (star_arr[i].include != 0)  {
			star_arr[i].mag0 = gsl_vector_get(mag0, i);
			star_arr[i].sigma = gsl_vector_get(sigm0, i);
		}
	}


}


	/************************************************************************
	 * PROCEDURE: init_matrix
	 *
	 * DESCRIPTION: allocate space for all the 1-indexed arrays and matrices
	 *              that will be needed to set up and solve the series
	 *              of linear equations.
	 *
	 * RETURNS:
	 *    nothing
	 */


static void
init_matrix
	(
	int ee,              /* I: number of images in solution */
	int ss               /* I: number of stars in solution */
	)
{


	if ((mag = gsl_matrix_calloc(ee, ss)) == NULL) {
		fprintf(stderr, "init_matrix: can't allocate for matrix mag\n");
		exit(-1);
	}
	if ((weight = gsl_matrix_calloc(ee, ss)) == NULL) {
		fprintf(stderr, "init_matrix: can't allocate for matrix weight\n");
		exit(-1);
	}
	if ((a = gsl_matrix_calloc(ee + ss, ee + ss)) == NULL) {
		fprintf(stderr, "init_matrix: can't allocate for matrix a\n");
		exit(-1);
	}
	if ((b = gsl_vector_calloc(ee + ss)) == NULL) {
		fprintf(stderr, "init_matrix: can't allocate for vector b \n");
		exit(-1);
	}
	if ((u = gsl_matrix_calloc(ee + ss, ee + ss)) == NULL) {
		fprintf(stderr, "init_matrix: can't allocate for matrix u\n");
		exit(-1);
	}
	if ((v = gsl_matrix_calloc(ee + ss, ee + ss)) == NULL) {
		fprintf(stderr, "init_matrix: can't allocate for matrix v\n");
		exit(-1);
	}
	if ((w = gsl_vector_calloc(ee + ss)) == NULL) {
		fprintf(stderr, "init_matrix: can't allocate for vector w \n");
		exit(-1);
	}
	if ((x = gsl_vector_calloc(ee + ss)) == NULL) {
		fprintf(stderr, "init_matrix: can't allocate for vector x \n");
		exit(-1);
	}
	if ((em = gsl_vector_calloc(ee)) == NULL) {
		fprintf(stderr, "init_matrix: can't allocate for vector em \n");
		exit(-1);
	}
	if ((mag0 = gsl_vector_calloc(ss)) == NULL) {
		fprintf(stderr, "init_matrix: can't allocate for vector mag0 \n");
		exit(-1);
	}
	if ((sigm0 = gsl_vector_calloc(ss)) == NULL) {
		fprintf(stderr, "init_matrix: can't allocate for vector sigm0 \n");
		exit(-1);
	}
	if ((Mes = gsl_matrix_calloc(ee, ss)) == NULL) {
		fprintf(stderr, "init_matrix: can't allocate for matrix Mes \n");
		exit(-1);
	}
	if ((sigMes = gsl_matrix_calloc(ee, ss)) == NULL) {
		fprintf(stderr, "init_matrix: can't allocate for matrix sigMes \n");
		exit(-1);
	}
	if ((sigem = gsl_vector_calloc(ee)) == NULL) {
		fprintf(stderr, "init_matrix: can't allocate for vector sigem \n");
		exit(-1);
	}
	if ((sigemn = gsl_vector_calloc(ee)) == NULL) {
		fprintf(stderr, "init_matrix: can't allocate for vector sigemn \n");
		exit(-1);
	}
}



	/**********************************************************************
	 * PROCEDURE: calc_sigmas
	 *
	 * DESCRIPTION: After we have solved the equations, we can go back
	 *              and calculate values of the RMS for each exposure 
	 *              (as a whole), for mean magnitude of each star, 
	 *              and for each observation of each star (individually).
	 *
	 *              The tricky part is to handle values of NODATA 
	 *              correctly so that they don't throw averages off. 
	 *
	 * RETURNS:
	 *    nothing
	 */

static void
calc_sigmas
	(
	int m,              /* I: number of images in solution */
	int n,              /* I: number of stars in solution */
	double mag_match    /* I: pass to find_sigm() function */
	)
{
	int i, j, num;
	double temp, temp2, top, bot;
	struct s_measurep *mp;
	struct s_measure *measure;

	/* set everything to 0.0, first of all */
	for (i = 0; i < m; i++) {
		gsl_vector_set(sigem, i, 0.0);
		gsl_vector_set(sigemn, i, 0.0);
		for (j = 0; j < n; j++) {
			gsl_matrix_set(sigMes, i, j, 0.0);
		}
	}
	for (i = 0; i < n; i++) {
		gsl_vector_set(sigm0, i, 0.0);
	}

	/* first find sigma(em(e)) */
	if (n >= 2) {
		for (i = 0; i < m; i++) {
			num = 0;
/****
			for (sp = image_arr[i].starp; sp != NULL; sp = sp->next)
				if (sp->star->include != 0)
					num++;
***/
			top = 0.0;
			bot = 0.0;
			for (j = 0; j < n; j++) {
				if ((gsl_matrix_get(Mes, i, j) == NODATA) || 
					 (gsl_vector_get(mag0, j) == NODATA)) {
					continue;
				}
				num++;
				temp = (gsl_matrix_get(Mes, i, j) - gsl_vector_get(mag0, j));
				top += temp*temp*gsl_matrix_get(weight, i, j);
				bot += gsl_matrix_get(weight, i, j);
			}
			if (bot > 0.0) {
				temp = sqrt((n*top)/((n - 1)*bot));
				gsl_vector_set(sigem, i, temp);
				if (num > 0) {
					temp = gsl_vector_get(sigem, i)/(double) sqrt((double) num);
					gsl_vector_set(sigemn, i, temp);
				}
			}
			else {
				gsl_vector_set(sigem, i, NODATA);	
				gsl_vector_set(sigemn, i, NODATA);	
			}
		}
	}

	/* now, find sigma(m0(s)) */
	if (m >= 2) {
		for (i = 0; i < n; i++) {
			top = 0.0;
			bot = 0.0;
			if (star_arr[i].variable == 0) {
				for (j = 0; j < m; j++) {
					if ((gsl_matrix_get(Mes, j, i) == NODATA) || 
					               (gsl_vector_get(mag0, i) == NODATA) ||
										(image_arr[j].include == 0)) {
						continue;
					}
					temp = gsl_matrix_get(Mes, j, i) - gsl_vector_get(mag0, i);
					top += temp*temp*gsl_matrix_get(weight, j, i);
					bot += gsl_matrix_get(weight, j, i);
				}
			}
			else {
				for (mp = star_arr[i].mp; mp != NULL; mp = mp->next) {
					measure = mp->measure;
					shAssert(measure != NULL);
					j = measure->image_id;
					if ((gsl_matrix_get(Mes, j, i) == NODATA) || 
					               (gsl_vector_get(mag0, i) == NODATA) ||
										(image_arr[j].include == 0)) {
						continue;
					}
					temp = gsl_matrix_get(Mes, j, i) - gsl_vector_get(mag0, i);
					top += temp*temp*measure->w4;
					bot += measure->w4;
				}
			}
			if (bot > 0.0) {
				temp = sqrt((m*top)/((m - 1)*bot));
				gsl_vector_set(sigm0, i, temp);
			}
			else {
				gsl_vector_set(sigm0, i, NODATA);
			}
		}
	}

	/* now we can finally calculate sigma(M(e,s)) */
	for (i = 0; i < m; i++) {
		for (j = 0; j < n; j++) {
			temp = find_sigm(n, i, j, mag_match);
			if (temp == NODATA) {
				gsl_matrix_set(sigMes, i, j, NODATA);
			}
			else if ((temp2 = gsl_vector_get(sigemn, i)) != NODATA) {
				temp = sqrt(temp*temp + temp2*temp2);
				gsl_matrix_set(sigMes, i, j, temp);
			}
			else {
				gsl_matrix_set(sigMes, i, j, temp);
			}
		}
	}
}


	/************************************************************************
	 * PROCEDURE: print_sigmas
	 *
	 * DESCRIPTION: print out some derived quantities.  For debugging
	 *              purposes only.
	 *
	 * RETURNS:
	 *    nothing
	 */

static void
print_sigmas
	(
	int m,            /* I: number of images */
	int n             /* I: number of stars */
	)
{
	int i;

	/* first find sigma(em(e)) */
	printf("   Sigma of Exposures: image #, sigem, sigemn\n");
	for (i = 0; i < m; i++) {
		printf(" %4d  %5.3f  %5.3f\n", image_arr[i].index, 
							 gsl_vector_get(sigem, i), 
							 gsl_vector_get(sigemn, i)); 
	}
	printf("\n");

	/* now, find sigma(m0(s)) */
	printf("  Sigma of stellar mean mags: star #, m0, sig(m0) \n");
	for (i = 0; i < n; i++) {
		printf("  %6d %6.3f %5.3f  \n", star_arr[i].id, 
				  			gsl_vector_get(mag0, i), 
							gsl_vector_get(sigm0, i)); 
	}
	printf("\n");
}



	/***********************************************************************
	 * PROCEDURE: calc_em
	 *
	 * DESCRIPTION: Calculate the 'em' matrix - the overal shift of 
	 *              each image's gnitudes to bring them close to the mean 
	 *              values. 
	 *
	 * RETURNS:
	 *    nothing
	 */

static void
calc_em
	(
	int m,          /* I: number of images */
	int n           /* I: number of stars */
	)
{
	int i;
	
	for (i = 0; i < m; i++) { 
		if (image_arr[i].include) {
			gsl_vector_set(em, i, gsl_vector_get(x, i));
		}
		else {
			gsl_vector_set(em, i, NODATA);
		}
	}
}


	/***********************************************************************
	 * PROCEDURE: calc_mags
	 *
	 * DESCRIPTION: Adjust the instrumental magnitudes for each star, 
	 *              adding the results of the matrix inversion to the
	 *              raw instrumental magnitude.  The result is the
	 *              differential magnitude of the star relative to
	 *              all the others in all images.
	 *
	 * RETURNS:
	 *    nothing
	 */

static void
calc_mags
	(
	int m,          /* I: number of images */
	int n           /* I: number of stars */
	)
{
	int i, j;	
	double model_correction;
	double temp;
	
	for (i = 0; i < m; i++) {
		for (j = 0; j < n; j++) {
			if ((gsl_matrix_get(mag, i, j) != NODATA) && 
			    (image_arr[i].include != 0)) {
				model_correction = calc_model_correction(m, n, i, j);
#ifdef DEBUG
				printf ("  calc_mags: %6.3lf %6.3lf   %6.3lf \n",
						gsl_matrix_get(mag, i, j), 
						model_correction, 
						gsl_matrix_get(mag, i, j) - model_correction);
#endif
				if (model_correction != NODATA) {
					temp = gsl_matrix_get(mag, i, j) - model_correction;
					gsl_matrix_set(Mes, i, j, temp);
				}
				else {
					gsl_matrix_set(Mes, i, j, NODATA);
				}
			}
			else {
				gsl_matrix_set(Mes, i, j, NODATA);
			}
		}
	}
}





	/************************************************************************
	 * PROCEDURE: print_curves
	 *
	 * DESCRIPTION: Print out light curves for each star 
	 *              (used only in debugging)
	 *
	 * RETURNS:
	 *    nothing
	 */

static void
print_curves
	(
	int m,             /* I: number of images */
	int n              /* I: number of stars */
	)
{
	int i, j;	
	
	printf("\n");
	for (i = 0; i < n; i++) {
		printf("star   %2d: ", i);
		for (j = 0; j < m; j++) {
			printf("%6.3f %6.3f  ", gsl_matrix_get(Mes, j, i), 
								 gsl_matrix_get(sigMes, j, i));
		}
		printf("\n");
	}
}




	/************************************************************************
	 * PROCEDURE: zero_vector
	 *
	 * DESCRIPTION: Set to zero all elements of vector p which are 
	 *              less than 'tolerance' times the maximum element of p
	 *
	 * RETURNS:
	 *    nothing
	 */

static void
zero_vector
	(
	gsl_vector *p,        /* I/O: a vector whose elements we may set to zero */
	int m,                /* I: number of elements in vector */
	double tolerance      /* I: if elem i < this times max elem, set to 0  */
	)
{
	int i;
	double max, min;

	max = 0.0;
	for (i = 0; i < m; i++) {
		if (gsl_vector_get(p, i) > max) {
			max = gsl_vector_get(p, i);
		}
	}
	min = tolerance*max;
	for (i = 0; i < m; i++) {
		if (gsl_vector_get(p, i) < min) {
			gsl_vector_set(p, i, 0.0);
		}
	}
}	



	/*******************************************************************
	 * PROCEDURE: print_vector
	 *
	 * DESCRIPTION: Print a vector (which goes 1-m) in a nice format
	 *
	 * RETURNS:
	 *    nothing
	 */

static void
print_vector
	(
	gsl_vector *p,      /* I: the vector we'll print */
   int m               /* I: number of elements in the vector */
	)
{
	int i;

	for (i = 0; i < m; i++) {
		printf("%9.3e ", gsl_vector_get(p, i));
	}
	putchar('\n');
}


	/**********************************************************************
	 * PROCEDURE: print_matrix
	 *
	 * DESCRIPTION: Print a matrix (which goes 1-m, 1-n) in a nice format 
	 *
	 * RETURNS:
	 *   nothing
	 */

static void
print_matrix
	(
	gsl_matrix *p,         /* I: matrix whose elements we print */
	int m,                 /* I: number of rows in matrix */
	int n                  /* I: number of columns in vector */
	)
{
	int i, j;

	for (i = 0; i < m; i++) {
		for (j = 0; j < n; j++) {
			printf("%9.3e ", gsl_matrix_get(p, i, j));
		}
		putchar('\n');
	}
}



	/***********************************************************************
	 * PROCEDURE: fill_bvec
	 *
	 * DESCRIPTION: Fill up the vector B, which is computed from 
	 *              observed magnitudes and weight.  This sits on the
	 *              right-hand side of the matrix equation.
	 *
	 *              There are two pieces to the B vector.  
	 *                 First, a set of ee values, 1 per image
	 *                 Next,  a set of ss values, 1 per star
	 *
	 * RETURNS:
	 *    nothing
	 */

static void 
fill_bvec
	(
	gsl_vector *b,         /* O: fill elements of this vector */
	int m,                 /* I: number of images */
	int n                  /* I: number of stars */
	)
{
	int i, j;
	double x;
	double temp1, temp2;

	/* 
	 * add up weights in each image; 
	 *   this includes an implicit sum over all filterchips
	 */
	for (i = 0; i < m; i++) {
		x = 0;
		for (j = 0; j < n; j++) {
			temp1 = gsl_matrix_get(weight, i, j);
			temp2 = gsl_matrix_get(mag, i, j);
			x += temp1*temp2;
		}
		gsl_vector_set(b, i, x);
	}

	/* 
	 * add up weights for each star; 
	 */
	for (j = 0; j < n; j++) {
		x = 0;
		for (i = 0; i < m; i++) {
			temp1 = gsl_matrix_get(weight, i, j);
			temp2 = gsl_matrix_get(mag, i, j);
			x += temp1*temp2;
		}
		gsl_vector_set(b, j + m, x);
	}

}


	/************************************************************************
	 * PROCEDURE: read_data
	 *
	 * DESCRIPTION: take the magnitude and uncertainty data, and figure
	 *              out how to turn it into weights for the
	 *              linear equations.
	 *
	 * RETURNS:
	 *    nothing
	 */

static void
read_data
	(
	int m,              /* I: number of images */
	int n              /* I: number of stars */
	)
{
	int i, j;
	int num_measures;
	struct s_star *star;
	struct s_measurep *mp;
	struct s_measure *measure;

	for (i = 0; i < m; i++) {
		for (j = 0; j < n; j++) {
			gsl_matrix_set(mag, i, j, 0.0);
			gsl_matrix_set(weight, i, j, 0.0);
		}
	}

	for (j = 0; j < n; j++) {
		star = &(star_arr[j]);	
		shAssert(star != NULL);
		mp = star->mp;
		num_measures = 0;
		while (mp != NULL) {
			measure = mp->measure;
			shAssert(measure != NULL);
			i = measure->image_id;
			gsl_matrix_set(mag, i, j, (double) measure->obs_mag);
			if ((star->include != 0) && (star->variable == 0) &&
			    (measure->include != 0) &&
				 (measure->image->include != 0)) {
				gsl_matrix_set(weight, i, j, (double) measure->w4);
			}
#ifdef DEBUG
			printf("  j %6d  i %6d  mag %6.3f  weight %6.3f \n", j, i, 
					gsl_matrix_get(mag, i, j), gsl_matrix_get(weight, i, j));
#endif

			num_measures++;
			mp = mp->next;
		}
		if (num_measures != star->num_measures) {
			fprintf(stderr, "read_data: num_measures %d but star_num %d ?!",
					num_measures, star->num_measures);
			exit(0);
		}
	}

}



	/***********************************************************************
	 * PROCEDURE: fill_matrix
	 *
	 * DESCRIPTION: Fill up the big matrix with properly weighted quantities.
	 *
	 *              The matrix can be broken into four "pieces"
	 *
	 *                   A      B      
	 *                   
	 *                   C      D 
	 *
	 *                   
	 *              where piece "A"  sum over all stars for each image
	 *                          "B"  weights of each image, star
	 *                          "C"  same as "B"
	 *                          "D"  sum over all images for each star
	 *
	 * RETURNS:
	 *    nothing
	 */

static void
fill_matrix
	(
	gsl_matrix *p,        /* O: fill the elements of this matrix */
	int m,                /* I: number of images */
	int n                 /* I: number of stars */
	)
{
	int i, j;
	double x;

	for (i = 0; i < m + n; i++) {
		for (j = 0; j < m + n; j++) { 	
			gsl_matrix_set(p, i, j, 0.0);
		}
	}

	/* fill up the real matrix in four steps, one per piece */

	/*   piece "A": sum weights of all stars in each image */
	for (i = 0; i < m; i++) {
		x = 0;
		for (j = 0; j < n; j++) {
			x += gsl_matrix_get(weight, i, j);
		}
		gsl_matrix_set(p, i, i, x);
	}

	/*   pieces "B" and "C": weight for each individual star, image */
	for (i = 0; i < m; i++) {
		for (j = 0; j < n; j++) {
			gsl_matrix_set(p, i, j + m, gsl_matrix_get(weight, i, j));
			gsl_matrix_set(p, j + m, i, gsl_matrix_get(weight, i, j));
		}
	}

	/*   piece "D": sum weights for all instances of each star */
	for (i = 0; i < n; i++) {
		x = 0;
		for (j = 0; j < m; j++) {
			x += gsl_matrix_get(weight, j, i);
		}
		gsl_matrix_set(p, i + m, i + m, x);
	}


#ifdef FORCE_STAR_MAG
	/* now we force the first star to have instrumental mag 0 */
	for (i = 0; i < m + n; i++) {
		gsl_matrix_set(p, i, m, 0.0);
	}
#endif

}


	/************************************************************************
	 * PROCEDURE: copy_matrix
	 *
	 * DESCRIPTION: copy the values of one matrix to another one
	 *
	 * RETURNS:
	 *    nothing
	 */

static void 
copy_matrix
	(
	gsl_matrix *p,         /* I: copy FROM this matrix ... */
	gsl_matrix *q,         /* O: INTO this matrix */
	int m,              /* I: number of rows in matrices */
	int n               /* I: number of columns in matrices */
	)
{
	int i, j;

	for (i = 0; i < m; i++) {
		for (j = 0; j < n; j++) {
			gsl_matrix_set(q, i, j, gsl_matrix_get(p, i, j));
		}
	}
}


	/***********************************************************************
	 * PROCEDURE: normalize
	 *
	 * DESCRIPTION: add a constant to all magnitudes 
	 *              (and to the x[] vector as well) so that the brightest 
	 *              star has the given mean magnitude value.
	 *              And also SUBTRACT this same constant to the image
	 *              adjustment factors so that we restore the 
	 *              equation connecting raw and corrected magnitudes
	 *              via the image adjustments.
	 *
	 * RETURNS:
	 *    nothing
	 *
	 */

static void
normalize
	(
	int m, 
	int n, 
	double value
	)
{
	int i, j;
	double min, diff, temp;

	min = 1.0e6;
	for (i = 0; i < n; i++) {
		if ((gsl_vector_get(mag0, i) < min) && 
		    (gsl_vector_get(mag0, i) != NODATA)) {
			min = gsl_vector_get(mag0, i);
		}
	}
	diff = (double) (value - min);

	for (i = 0; i < n; i++) {
		if (gsl_vector_get(mag0, i) != NODATA) {
			temp = gsl_vector_get(mag0, i) + diff;
			gsl_vector_set(mag0, i, temp);
		}
		for (j = 0; j < m; j++) {
			if (gsl_matrix_get(Mes, j, i) != NODATA) {
				temp = gsl_matrix_get(Mes, j, i) + diff;
				gsl_matrix_set(Mes, j, i, temp);
			}
		}
	}

	/* 
	 * now, shift the image adjustment magnitudes by this same amount
	 * but in the opposite direction, to preserve the connection
	 * between input and output magnitudes
	 */
	for (i = 0; i < m; i++) { 
		if (gsl_vector_get(em, i) != NODATA) {
			temp = gsl_vector_get(em, i) - diff;
			gsl_vector_set(em, i, temp);
		}
	}
			
}



	/***********************************************************************
	 * PROCEDURE: find_means
	 *
	 * DESCRIPTION: find the mean magnitude of each star, either
	 *
	 *                 a. by simply looking in the 'x' vector, for 
	 *                    stars with non-zero weights
	 *                    (we solved for the mean mags of these stars)
	 *
	 *               or 
	 *
	 *                 b. by calculating it from the weighted average of
	 *                    (corrected) mag on each night 
	 *                    (these are "bad" or "variable" stars)
	 *
	 *               If there is no information for a star, 
	 *               use the NODATA value as its mean mag 
	 *
	 *               Once we know the value for a star, set its entry
	 *               in the "mag0" array to that value.
	 *
	 * RETURNS:
	 *    nothing
	 */

static void
find_means(m, n)
int m, n;
{
	int i, index;
	double top, bot;
	struct s_measurep *mp;
	struct s_measure *measure;

	for (i = 0; i < n; i++) {
		if ((star_arr[i].include != 0) && (star_arr[i].variable == 0)) {
			gsl_vector_set(mag0, i, gsl_vector_get(x, i + m));
		}
		else if (star_arr[i].variable != 0) {
			top = 0.0;
			bot = 0.0;
			mp = star_arr[i].mp;
			while (mp != NULL) {
				measure = mp->measure;
				shAssert(measure != NULL);

				index = measure->image_id;
				if (gsl_matrix_get(Mes, index, i) == NODATA) {
					mp = mp->next;
					continue;
				}
				top += gsl_matrix_get(Mes, index, i)*measure->w4;
				bot += measure->w4;
				mp = mp->next;
			}

			if (bot > 0.0) {
				gsl_vector_set(mag0, i, top/bot);
			}
			else {
				gsl_vector_set(mag0, i, NODATA);
			}

		}
	}
}


	/********************************************************************
	 * PROCEDURE: find_sigm
	 *
	 * DESCRIPTION: Return an estimate of the uncertainty sigma(m(e,s)) 
	 *              for star s in image e, which is only PART of the 
	 *              total error estimate sigma(M(e,s)).  
	 *              Find this value in two ways:
	 * 
	 *                1. try to find the average value of sigma(m0) for
	 *                          m0 ~= m(e,s)
	 *
	 *                2. if impossible, then use a value calculated
	 *                   from the weight of the value for m(e,s)
	 *                   since it is really based on input estimate
	 *                   of uncertainty, like so:
 	 *                      sig(m(e,s)) = 1.0/sqrt(weight(e,s))
	 *
	 *              If neither is possible, then the uncertainty
	 *              will be NODATA
	 *
	 * RETURNS:
	 *    the value of uncertainty in mean magnitude
	 */

static double
find_sigm
	(
	int n,              /* I: number of stars overall in solution */
	int e,              /* I: we are interested in a star in this image ... */
	int s,              /* I: ... and it's star 's' in this image */
	double mag_match    /* I: how similar in mag must star be to act */
	                    /*      as stand-in to variable? */

	)
{
	int i, num;
	double dist, sum, wt, ret;
   double temp, other_star_sigm0;

#ifdef DEBUG
	printf("   find_sigm: star %3d in image %3d \n", s, e);
#endif

	/* first, try to find a mean m0 whose magnitude is close to 
	   m(e,s).  use a weighted average of the nearby sigm0. */
	sum = 0.0;
	wt = 0.0;
	num = 0;
	for (i = 0; i < n; i++) {
		/* 
		 * don't include the star itself  
		 */
		if (i == s) {
			continue;
		}
		temp = gsl_vector_get(mag0, i) - gsl_matrix_get(Mes, e, s);
#ifdef DEBUG
		printf("       star %3d has mag0 %6.2f  target %3d has Mes %6.2f \n",
						i, gsl_vector_get(mag0, i), s, gsl_matrix_get(Mes, e, s));
#endif
		/* 
		 * fixed bug here by adding extra check on sigm0 validity.
		 *   MWR 4/8/2006
		 */
		if ((dist = fabs(temp)) < mag_match) {
#ifdef DEBUG
			printf("       dist %6.2f < %6.2f qualifies ", dist, mag_match);
#endif
			other_star_sigm0 = gsl_vector_get(sigm0, i);
#ifdef DEBUG
			printf(" other_sigm0 %6.2f ", other_star_sigm0);
#endif
			if (other_star_sigm0 != NODATA) {
				if (dist < TINY) {
					dist = TINY;
				}
				sum += gsl_vector_get(sigm0, i)/dist;
#ifdef DEBUG
				printf(" add to sum now %6.2f\n", sum);
#endif
				wt += 1.0/dist;
				num++;
			}
#ifdef DEBUG
			else {
				printf(" no good, ignore \n");
			}
#endif
		}
	}
	if (num > 0) {
		ret = sum/wt;
		return(ret);
	}

	/* second best: use the weight of instrumental mag(e,s) as an
	   estimator */
	if ((temp = gsl_matrix_get(weight, e, s)) > 0.0) {
		ret = 1.0/sqrt(temp);
		return(ret);
	}

	/* nothing left to do but return NODATA */
	ret = NODATA;
	return(ret);
}



	/***********************************************************************
	 * PROCEDURE: write_images
	 *
	 * DESCRIPTION: Write output about images ONLY to the given file 
	 *              (or stdout, if the given string is "").
	 *
	 * RETURNS:
	 *    0         if all goes well
	 *    1         if an error occurs
	 *
	 */

int
write_images
	(
	char *file
	)
{
	int i;
	FILE *fp;
	struct s_image *im;

	if (strcmp(file, "") == 0) {
		fp = stdout;
	}
	else if ((fp = fopen(file, "w")) == NULL) {
		fprintf(stderr, "write_output: can't open file %s\n", file);
		return(1);
	}
	if ((fp = fopen(file, "w")) == NULL) {
		fprintf(stderr, "write_output: can't open file %s\n", file);
		return(1);
	}

	for (i = 0; i < num_images; i++) {
		if (image_arr[i].include == 0) {
			continue;
		}
		im = &(image_arr[i]);
		fprintf(fp, "%-4d %30s %6.1f  %14.7f ",
			im->index, im->name, im->exptime, im->jd);
		fprintf(fp, "%5.3f %5.3f %5.3f\n", 
							gsl_vector_get(em, i), 
							gsl_vector_get(sigem, i), 
							gsl_vector_get(sigemn, i));
	}
	if (fp != stdout) {
		fclose(fp);
	}

	return(0);
}


	/***********************************************************************
	 * PROCEDURE: write_stars
	 *
	 * DESCRIPTION: Write information about stars ONLY, one line per
	 *              star per image, to the given file 
	 *              (or stdout if file == "").  
    *
	 *              We add the given "delta_mag" to all magnitude values
	 *              before printing them out.  It will be zero
	 *              unless the user has given the "mrange" commend-line
	 *              option.
	 *
	 * RETURNS:
	 *    0         if all goes well
	 *    1         if an error occurs
	 *
	 */

int
write_stars
	(
	char *file,
	double delta_mag         /* I: add this to all output mag values */
	)
{
	int i, index;
	FILE *fp;
	struct s_star *star;
	struct s_image *im;
	struct s_measurep *mp;
	struct s_measure *measure;

	if (strcmp(file, "") == 0) {
		fp = stdout;
	}
	else if ((fp = fopen(file, "w")) == NULL) {
		fprintf(stderr, "write_output: can't open file %s\n", file);
		return(0);
	}
	if ((fp = fopen(file, "w")) == NULL) {
		fprintf(stderr, "write_output: can't open file %s\n", file);
		return(0);
	}

	for (i = 0; i < num_stars; i++) {
		if (star_arr[i].include == 0) {
			continue;
		}
		star = &(star_arr[i]);
		mp = star->mp;
		while (mp != NULL) {
			measure = mp->measure;
			shAssert(mp != NULL);

			im = measure->image;
			shAssert(im != NULL);
			if (im->include != 0) {
				index = im->index;
				fprintf(fp, "%-6d %9.5f %9.5f %4d %11.5f ",
						star->id, star->ra, star->dec, index, im->jd);
				fprintf(fp, " %6.3f %6.3f %5.3f ", 
						gsl_vector_get(mag0, i) + delta_mag,
						gsl_matrix_get(Mes, index, i) + delta_mag, 
						gsl_matrix_get(sigMes, index, i));
				fprintf(fp, " %2d %7.2f \n",
						star->variable, star->nsigma);
			}
			mp = mp->next;
		}
	}
	if (fp != stdout) {
		fclose(fp);
	}

	return(0);
}





	/************************************************************************
	 * PROCEDURE: write_sigma
	 *
	 * DESCRIPTION: Write information about sigma0 vs. mag0 
	 *              (standard deviation from mean mag, versus mean mag)
	 *              to the given file (or stdout if the given file == "") 
	 *
	 *              We add the given "delta_mag" to all magnitude values
	 *              before printing them out.  It will be zero
	 *              unless the user has given the "mrange" commend-line
	 *              option.
	 *
	 * RETURNS:
	 *    0         if all goes well
	 *    1         if an error occurs
	 *
	 */

int
write_sigma
	(
	char *file,
	double delta_mag       /* I: add this to all output magnitudes */
	)
{
	int i;
	FILE *fp;

	if (strcmp(file, "") == 0) {
		fp = stdout;
	}
	else if ((fp = fopen(file, "w")) == NULL) {
		fprintf(stderr, "write_output: can't open file %s\n", file);
		return(1);
	}
	if ((fp = fopen(file, "w")) == NULL) {
		fprintf(stderr, "write_output: can't open file %s\n", file);
		return(1);
	}

	for (i = 0; i < num_stars; i++) {
		if (star_arr[i].include != 0)  {
			fprintf(fp, "%6d %6.3f %6.3f %2d %7.2f %12.5f %12.5f\n", 
						star_arr[i].id, 
						star_arr[i].mag0 + delta_mag,
						star_arr[i].sigma,
						star_arr[i].variable,
						star_arr[i].nsigma,
						star_arr[i].ra,
						star_arr[i].dec);
		}
	}

	if (fp != stdout) {
		fclose(fp);
	}

	return(0);
}




#if 0
	/************************************************************************
	 * PROCEDURE: calc_model_mag
	 *
	 * DESCRIPTION: We have finished finding the photometric solution,
	 *              so we have the image adjustments, mean star magnitudes,
	 *              and chip offsets.  Here, we figure out the magnitude
	 *              a star _ought_ to have, based on its image, star,
	 *              chip, etc.  We go back to the x[] vector values
	 *              for all these calculations.
	 *
	 *              The value returned can be compared to the actual
	 *              measured magnitude; in a perfect world, they
	 *              would be the same.  In the real world, we expect
	 *              small residuals which (if we have done a good 
	 *              job in the model) will be randomly scattered
	 *              around zero.
	 *
	 * RETURNS:
	 *     magnitude of star from solution
	 *     NODATA         if this isn't a valid measurement
	 */

static double
calc_model_mag
	(
	int num_images,        /* I: total number of images in solution */
	int num_stars,         /* I: total number of stars in solution */
	int image_index,       /* I: index of image, 1 ... num_images */
	int star_index         /* I: index of star, 1 .. num_stars */
	)
{
	double star_mag0, mag_correction;
	double model_mag;

	/* we don't (yet) consider whether this is a valid measurement */

	star_mag0 = x[num_images + star_index];
	mag_correction = calc_model_correction(num_images, num_stars,
			              image_index, star_index);

	if (mag_correction == NODATA) {
		model_mag = NODATA;
	}
	else {
		model_mag = star_mag0 + mag_correction;
	}

	return(model_mag);
}

#endif



	/************************************************************************
	 * PROCEDURE: calc_model_correction
	 
	 * DESCRIPTION: We have finished finding the photometric solution,
	 *              so we have the image adjustments, mean star magnitudes,
	 *              and chip offsets.  Here, we figure out the total
	 *              size of all the adjustments, due to image zeropoint,
	 *              chip offset, filter mismatch, etc.
	 *
	 *              The result should be used in this sense:
	 *
	 *                  obs_mag  =   true_mag  +  corrections
	 *
	 *
	 * RETURNS:
	 *     corrections to star's mag
	 *     NODATA       if we can't calculate it
	 */

static double
calc_model_correction
	(
	int num_images,        /* I: total number of images in solution */
	int num_stars,         /* I: total number of stars in solution */
	int image_index,       /* I: index of image, 1 ... num_images */
	int star_index         /* I: index of star, 1 .. num_stars */
	)
{
	double image_em;
	double mag_correction;

	/* we don't (yet) consider whether this is a valid measurement */
	image_em = gsl_vector_get(x, image_index);

	mag_correction = image_em;

	return(mag_correction);
}




