

	/*************************************************************************
	 * Take as input a file which contains the instrumental photometry from
	 * many SNAP images, re-organized so that all data for a given star
	 * is collected together.  Set up a set of linear equations which
	 * represent the photometric model of the entire system, and then 
	 * solve those equations.  Print out final reduced differential
	 * magnitudes for each star, as well as diagnostic information about
	 * each image.
	 *
	 * MWR 10/18/2003
	 *
	 * Added command-line option "primary=" to pick out primary fiducial
	 * filter index.
	 *
	 * MWR 10/31/2003
	 *
	 * Added command-line option "fcfile=" to place .fc output into a
	 * desired file instead of "solvesnap.fc".
	 *  
	 * MWR 8/6/2004
	 *
	 * Reverted to the format of the old "solvepht", which expects
	 * simply a list of stellar measurements from different exposures
	 * of a single chip.  Took out all the SNAP-specific stuff to
	 * make a simple, general tool.
	 * MWR 9/19/2004
	 *
	 * Fixed the "markoff()" procedure so that it correctly
	 * compares the ID value given by user on command-line
	 * to the ID values of stars (not their index positions).
	 * Also make sure that we set variable field to 1 for
	 * varstar= stars (in some old versions, we set to 0, making
	 * variables _not_ variable!).
	 * MWR 10/26/2004
	 *
	 * Modified "fit_median()" function so that it handles distributions
	 * of stars which fall into just one or two bins of magnitude 
	 * reasonably.  Old version would halt program execution.
	 * MWR 6/13/2005
	 *
	 * Fixed a bug in the define_scatter() function: I wasn't allocating
	 * enough space for the arrays of binned quantities.  This caused
	 * problems only in sparse fields.
	 * Also, added two command-line options:
	 *         scatterbin=   optional, sets width in mag of bins used to
	 *                            estimate uncertainty as func of mag
	 *         magmatch=     optional, sets width in mag used to find
	 *                            "similar" stars which can be used to
	 *                            estimate uncertainty per measurement
	 *                            in a variable object
	 * MWR 5/9/2006
	 *
	 * Added new command-line options:
	 *         irange=      allows one to reset the zero-point of output 
	 *                            magnitude scale based on stars within
	 *                            a range of INPUT magnitudes.
	 *         varmin=      set input mag value below which stars are marked
	 *                            as possibly variable (due to saturation)
	 * MWR 5/29/2006
	 *
	 * Modified heavily the manner in which the program tries to 
	 * determine the typical scatter as a function of mag and identify
	 * variable objects.  The new version
	 *    a) replaces a parabolic fit for "sigma-vs-mag" with a 
	 *          simple piecewise linear interpolation method
	 *    b) removes the use of the "range_width" parameter
	 *
	 * MWR 8/7/2015
	 *
	 * Added new command-line options:
	 *         varmax=      set input mag value ABOVE which stars are
	 *                            marked as possibly variable
	 *                            (so faint stars don't mess up solution)
	 *         vardet=      any object detected <= this many times
	 *                            is marked as possibly variable
	 * 
	 * MWR 4/26/2017
	 *
	 *
	 */

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <gsl/gsl_multifit.h>
#include "solvepht.h"

#undef DEBUG
#undef DEBUG5

#define IMFILE     "solvepht.img"
#define SIGFILE    "solvepht.sig"
#define OUTFILE    "solvepht.out"

static char *progname = "solvepht";

typedef struct s_scatter {
		double mag;
		double sigma;
} S_SCATTER;


static int markoff(int nbadim, int *badim, int nbadstar, int *badstar, 
                    int nvarstar, int *varstar);
static int define_scatter(double scatter_nsigma, double min_scatter,
                          double scatter_bin_size);
static int calc_stats(int num_in_bin, S_SCATTER *scat_arr,
                          double *median, double *width);
static int comp_scatter(struct s_scatter *s1, struct s_scatter *s2);
static int fit_median(int num_bins, int *num_stars_array, double *mag_array,
                          double *median_array,  
                          double *a, double *b, double *c);
static int fit_range(int num_bins, int *num_stars_array, double *mag_array, 
                          double *range_array, double *a_coeff);
static int mark_variable(double nsigma, double min_scatter, int num_bins,
                          double *bin_mag_array, double *median_array);
static int calc_mag_shift(double mrange_low, double mrange_high,
                          double *delta_mag);
static int calc_imag_shift(double irange_low, double irange_high,
                           double *delta_mag);


char usage[] = "usage: solvepht [debug] [infile=] [mconst=] [outfile=] [imfile=] [sigfile=] \n                [badim=] [badstar=] [varstar=] [min=] [max=] [varmin=] [varmax=] \n                [nsigma=] [niter=] [mrange[=low,high]] [irange[=low,high]] \n                [scatterbin=] [magmatch=] [minscatter=] [vardet=]]  \n"; 

int
main
	(
	int argc, 
	char *argv[]
	)
{
	char infile[100], outfile[100], imfile[100], sigfile[100];
	int i, nn, nbadim, nbadstar, nvarstar, *badim, *badstar, *varstar;
	int iter, num_iter;
	int mrange_flag, irange_flag;
	int vardet;
	double mconst, min, max, varmin, varmax;
	double scatter_nsigma;
	double mrange_low, mrange_high, delta_mag;
	double irange_low, irange_high;
	double scatter_bin_size;
	double mag_match;
	double min_scatter;

	/* set some global variables to initial known values */
	image_arr = NULL;
	star_arr = NULL;
	measure_list = NULL;

	debug_flag = 0;
	nbadim = 0;
	nbadstar = 0;
	nvarstar = 0;
	mrange_flag = 0;
	irange_flag = 0;
	min = -1.0e6;
	max = 1.0e6;
	varmin = -1.0e6;
	varmax =  1.0e6;
	vardet = 0;
	if ((badim = (int *) calloc(argc, sizeof(int))) == NULL) {
		fprintf(stderr, "%s: can't malloc for %d ints \n", progname, argc);
		exit(1);
	}
	if ((badstar = (int *) calloc(argc, sizeof(int))) == NULL) {
		fprintf(stderr, "%s: can't malloc for %d ints \n", progname, argc);
		exit(1);
	}
	if ((varstar = (int *) calloc(argc, sizeof(int))) == NULL) {
		fprintf(stderr, "%s: can't malloc for %d ints \n", progname, argc);
		exit(1);
	}
	mconst = MAG_CONST;
	scatter_nsigma = SCATTER_NSIGMA;
	num_iter = NUM_ITERATIONS;
	mrange_low = MRANGE_LOW;
	mrange_high = MRANGE_HIGH;
	irange_low = IRANGE_LOW;
	irange_high = IRANGE_HIGH;
	scatter_bin_size = SCATTER_BIN_SIZE;
	mag_match = MAG_MATCH;
	min_scatter = MIN_POSSIBLE_SCATTER;
	

	strcpy(infile, "");
	strcpy(outfile, OUTFILE);
	strcpy(imfile, IMFILE);
	strcpy(sigfile, SIGFILE);

	for (i = 1; i < argc; i++) {
		if (strcmp(argv[i], "help") == 0) {
			fprintf(stderr, "%s\n", usage);
			exit(0);
		}
		if (strcmp(argv[i], "debug") == 0) {
			debug_flag = 1;
			continue;
		}
		if (strncmp(argv[i], "infile=", 7) == 0) {
			if (sscanf(argv[i] + 7, "%s", infile) != 1) {
				fprintf(stderr, "%s: bad value for infile ..%s..\n", 
						              progname, argv[i]);
				exit(1);
			}
			continue;
		}
		if (strncmp(argv[i], "outfile=", 8) == 0) {
			if (sscanf(argv[i] + 8, "%s", outfile) != 1) {
				fprintf(stderr, "%s: bad value for outfile ..%s..\n", 
						              progname, argv[i]);
				exit(1);
			}
			continue;
		}
		if (strncmp(argv[i], "imfile=", 7) == 0) {
			if (sscanf(argv[i] + 7, "%s", imfile) != 1) {
				fprintf(stderr, "%s: bad value for imfile ..%s..\n", 
						              progname, argv[i]);
				exit(1);
			}
			continue;
		}
		if (strncmp(argv[i], "sigfile=", 8) == 0) {
			if (sscanf(argv[i] + 8, "%s", sigfile) != 1) {
				fprintf(stderr, "%s: bad value for sigfile ..%s..\n", 
						               progname, argv[i]);
				exit(1);
			}
			continue;
		}
		if (strncmp(argv[i], "mconst=", 7) == 0) {
			if (sscanf(argv[i] + 7, "%lf", &mconst) != 1) {
				fprintf(stderr, "%s: bad value for mconst ..%s..\n", 
						               progname, argv[i]);
				exit(1);
			}
			continue;
		}
		if (strncmp(argv[i], "badim=", 6) == 0) {
			if (sscanf(argv[i] + 6, "%d", &nn) != 1) {
				fprintf(stderr, "%s: bad value for badim ..%s..\n", 
						                progname, argv[i]);
				exit(1);
			}
			badim[nbadim++] = nn;
			continue;
		}
		if (strncmp(argv[i], "badstar=", 8) == 0) {
			if (sscanf(argv[i] + 8, "%d", &nn) != 1) {
				fprintf(stderr, "%s: bad value for badstar ..%s..\n", 
						                progname, argv[i]);
				exit(1);
			}
			badstar[nbadstar++] = nn;
			continue;
		}
		if (strncmp(argv[i], "varstar=", 8) == 0) {
			if (sscanf(argv[i] + 8, "%d", &nn) != 1) {
				fprintf(stderr, "%s: bad value for varstar ..%s..\n", 
						                progname, argv[i]);
				exit(1);
			}
			varstar[nvarstar++] = nn;
			continue;
		}
		if (strncmp(argv[i], "min=", 4) == 0) {
			if (sscanf(argv[i] + 4, "%lf", &min) != 1) {
				fprintf(stderr, "%s: bad value for min ..%s..\n", 
						                progname, argv[i]);
				exit(1);
			}
			continue;
		}
		if (strncmp(argv[i], "max=", 4) == 0) {
			if (sscanf(argv[i] + 4, "%lf", &max) != 1) {
				fprintf(stderr, "%s: bad value for max ..%s..\n", 
						                progname, argv[i]);
				exit(1);
			}
			continue;
		}
		if (strncmp(argv[i], "varmin=", 7) == 0) {
			if (sscanf(argv[i] + 7, "%lf", &varmin) != 1) {
				fprintf(stderr, "%s: bad value for varmin ..%s..\n", 
						                progname, argv[i]);
				exit(1);
			}
			continue;
		}
		if (strncmp(argv[i], "varmax=", 7) == 0) {
			if (sscanf(argv[i] + 7, "%lf", &varmax) != 1) {
				fprintf(stderr, "%s: bad value for varmax ..%s..\n", 
						                progname, argv[i]);
				exit(1);
			}
			continue;
		}
		if (strncmp(argv[i], "vardet=", 7) == 0) {
			if (sscanf(argv[i] + 7, "%d", &vardet) != 1) {
				fprintf(stderr, "%s: bad value for vardet ..%s..\n", 
						                progname, argv[i]);
				exit(1);
			}
			continue;
		}
		if (strncmp(argv[i], "nsigma=", 7) == 0) {
			if (sscanf(argv[i] + 7, "%lf", &scatter_nsigma) != 1) {
				fprintf(stderr, "%s: bad value for nsigma ..%s..\n", 
						                progname, argv[i]);
				exit(1);
			}
			continue;
		}
		if (strncmp(argv[i], "niter=", 6) == 0) {
			if (sscanf(argv[i] + 6, "%d", &num_iter) != 1) {
				fprintf(stderr, "%s: bad value for nsigma ..%s..\n", 
						                progname, argv[i]);
				exit(1);
			}
			continue;
		}
		if (strncmp(argv[i], "mrange", 6) == 0) {
			mrange_flag = 1;
			if (strncmp(argv[i], "mrange=", 7) == 0) {
				if (sscanf(argv[i] + 7, "%lf,%lf", &mrange_low, &mrange_high) != 2) {
					fprintf(stderr, "%s: bad value(s) for mrange limits ..%s..\n", 
						                progname, argv[i]);
					exit(1);
				}
				/* sanity check */
				if (mrange_low > mrange_high) {
					fprintf(stderr, "%s: mrange_low %lf cannot be > mrange_high %lf \n",
								progname, mrange_low, mrange_high);
					exit(1);
				}	
			}
			continue;
		}
		if (strncmp(argv[i], "irange", 6) == 0) {
			irange_flag = 1;
			if (strncmp(argv[i], "irange=", 7) == 0) {
				if (sscanf(argv[i] + 7, "%lf,%lf", &irange_low, &irange_high) != 2) {
					fprintf(stderr, "%s: bad value(s) for irange limits ..%s..\n", 
						                progname, argv[i]);
					exit(1);
				}
				/* sanity check */
				if (irange_low > irange_high) {
					fprintf(stderr, "%s: irange_low %lf cannot be > irange_high %lf \n",
								progname, irange_low, irange_high);
					exit(1);
				}	
			}
			continue;
		}
		if (strncmp(argv[i], "scatterbin=", 11) == 0) {
			if (sscanf(argv[i] + 11, "%lf", &scatter_bin_size) != 1) {
				fprintf(stderr, "%s: bad value for scatter_bin ..%s..\n", 
						                progname, argv[i]);
				exit(1);
			}
			continue;
		}
		if (strncmp(argv[i], "magmatch=", 9) == 0) {
			if (sscanf(argv[i] + 9, "%lf", &mag_match) != 1) {
				fprintf(stderr, "%s: bad value for mag_match ..%s..\n", 
						                progname, argv[i]);
				exit(1);
			}
			continue;
		}
		if (strncmp(argv[i], "minscatter=", 11) == 0) {
			if (sscanf(argv[i] + 11, "%lf", &min_scatter) != 1) {
				fprintf(stderr, "%s: bad value for min_scatter ..%s..\n", 
						                progname, argv[i]);
				exit(1);
			}
			continue;
		}
		fprintf(stderr, "%s\n", usage);
		exit(1);
	}
	if ((irange_flag > 0) && (mrange_flag > 0)) {
		fprintf(stderr, "%s: specify only one of irange or mrange, not both\n",
					progname);
		exit(1);
	}

	if (read_input(infile, min, max, varmin, varmax, vardet) != 0) {
		fprintf(stderr, "%s: read_input returns with error \n", progname);
		exit(1);
	}

	if (markoff(nbadim, badim, nbadstar, badstar, nvarstar, varstar) != 0) {
		fprintf(stderr, "%s: markoff returns with error \n", progname);
		exit(1);
	}

	/* 
	 * do I need to run "good_stars()" and "good_images()" again, now that
	 *   all the "bad" images and stars have been marked as such?  
	 *   Think about it.
	 *   MWR 12/3/2003.
	 */


	if (debug_flag) {
		print_images(stdout);
		print_stars(stdout, LONG);
	}

	for (iter = 0; iter <= num_iter; iter++) {

		solve_matrix(mconst, mag_match);

		/* figure out which stars have larger-than-expected scatter in solution */
		if (define_scatter(scatter_nsigma, min_scatter, scatter_bin_size) != 0) {
			fprintf(stderr, "%s: define_scatter returns with error\n", progname);
			exit(1);
		}
	
	}

	/* 
	 * if the user requested it, shift the output magnitude values so 
	 * that they are roughly the same as the input values 
	 */
	if (mrange_flag == 1) {
		if (calc_mag_shift(mrange_low, mrange_high, &delta_mag) != 0) {
			fprintf(stderr, "%s: shift_mags returns with error \n", progname);
			exit(1);
		}
	}
	else if (irange_flag == 1) {
		if (calc_imag_shift(irange_low, irange_high, &delta_mag) != 0) {
			fprintf(stderr, "%s: calc_imag_shift returns with error \n", progname);
			exit(1);
		}
	}
	else {
		delta_mag = 0;
	}

	write_images(imfile);
	write_sigma(sigfile, delta_mag);
	write_stars(outfile, delta_mag);

	exit(0);
}



	/***********************************************************************
	 * PROCEDURE: markoff
	 *
	 * DESCRIPTION: mark all the given stars and/or images as BAD, 
	 *              which means they're not used in the solution at all.
	 *              Go through all measurements associated with these
	 *              images and stars, and set the measurements'
	 *              "include" field to zero as well.
	 *
	 *              Mark as VARIABLE the stars which the user flagged
	 *              as "varstar" on the command line. VARIABLE stars are
	 *              are not used to solve for the em's, but later
	 *              do have their magnitudes calculated. 
	 *
	 * RETURNS:
	 *     0        if all goes well
	 *     1        if an error occurs
	 */

static int 
markoff
	(
	int nbadim, 
	int *badim, 
	int nbadstar, 
	int *badstar, 
	int nvarstar, 
	int *varstar
	)
{
	int i, j;
	int max_id;
	int num_removed;
	int num_good_removed;
	struct s_image *image;
	struct s_star *star;
	struct s_measurep *mp;
	struct s_measure *measure;


	/* first, we set to 0 the "include" field of all bad images */
	for (i = 0; i < nbadim; i++) {
		if ((badim[i] < 0) || (badim[i] >= num_images)) {
			fprintf(stderr, "%s: markoff: nbadim = %d, badim[%d] = %d? \n", 
					progname, nbadim, i, badim[i]);
			return(1);
		}
		image_arr[badim[i]].include = 0;
	}


	/* 
	 * next, we set to 0 the "include" field of all the measurements
	 *   which occurred in those images.
	 */
	num_removed = 0;
	num_good_removed = 0;

	for (i = 0; i < num_images; i++) {
		image = &(image_arr[i]);
		if (image->include == 0) {
			mp = image->mp;
			while (mp != NULL) {
				shAssert((measure = mp->measure) != NULL);
				if (measure->include != 0) {
					num_good_removed++;
				}
				num_removed++;
				measure->include = 0;
				mp = mp->next;
			}
		}
	}
#ifdef DEBUG
	printf("markoff: img  %6d removed  %6d good removed \n", 
					num_removed, num_good_removed);
#endif


	/* 
	 * now, it is _possible_ that there may be some stars
	 * with ID numbers larger than the total number of stars;
	 * the user may have edited the list of images and stars
	 * provided by "multipht".  So, we make a special search
	 * here for the maximum star ID number.
	 */
	max_id = -1;
	for (i = 0; i < num_stars; i++) {
		if (star_arr[i].id > max_id) {
			max_id = star_arr[i].id;
		}
	}
#ifdef DEBUG
	printf("markoff: num_stars %d  max star ID number %d \n", 
						num_stars, max_id);
#endif


	/* 
	 * now, we go through the list of all "bad" star ID values
	 * and find those ID values in the array of stars.
	 * We'll mark those entries not to be included in calculations.
	 */
	for (i = 0; i < nbadstar; i++) {
		if ((badstar[i] < 0) || (badstar[i] > max_id)) {
			fprintf(stderr, "%s: markoff: nbadstar = %d, badstar[%d] = %d? \n", 
					progname, nbadstar, i, badstar[i]);
			return(1);
		}
		for (j = 0; j < num_stars; j++) {
			if (star_arr[j].id == badstar[i]) {
				star_arr[j].include = 0;
				break;
			}
		}
	}
	/* 
	 * and we also need to set to 0 the "include" field of all 
	 * the _measurements_ of these stars
	 */
	num_removed = 0;
	num_good_removed = 0;
	for (i = 0; i < num_stars; i++) {
		star = &(star_arr[i]);
		if (star->include == 0) {
			mp = star->mp;
			while (mp != NULL) {
				shAssert((measure = mp->measure) != NULL);
				if (measure->include != 0) {
					num_good_removed++;
				}
				num_removed++;
				measure->include = 0;
				mp = mp->next;
			}
		}
	}
#ifdef DEBUG
	printf("markoff: star %6d removed  %6d good removed \n", 
					num_removed, num_good_removed);
#endif
		

	/* 
	 * finally, we set the "variable" field to 1 for all stars
	 *   which the user flagged as "varstar" in the command line 
	 */

	for (i = 0; i < nvarstar; i++) {
		if ((varstar[i] < 0) || (varstar[i] > max_id)) {
			fprintf(stderr, "%s: markoff: nvarstar = %d, varstar[%d] = %d? \n", 
					progname, nvarstar, i, varstar[i]);
			return(1);
		}
		for (j = 0; j < num_stars; j++) {
			if (star_arr[j].id == varstar[i]) {
				star_arr[j].variable = 1;
				break;
			}
		}
	}

	return(0);
}




	/***********************************************************************
	 * PROCEDURE: define_scatter
	 *
	 * DESCRIPTION: Once we have performed an ensemble solution, examine
	 *              the scatter-vs-mag relationship.  We want to fit
	 *              a low-order polynomial to the scatter-vs-mag for
	 *              this solution, robustly ignoring outliers.
	 *              We then use the fit to identify stars to be
	 *              marked as variables.
	 *
	 *              We divide the stars into bins by their solution
	 *              mean magnitudes, then calculate statistics on the
	 *              stars in each bin separately.
	 *
	 * RETURNS:
	 *     0        if all goes well
	 *     1        if an error occurs
	 */

static int 
define_scatter
	(
	double nsigma,       /* I: outliers must be this many sigma above the */
	                     /*    model in sigma-vs-mag to be marked "variable" */
	double min_scatter,  /* I: place this floor on empirical scatter */
	                     /*    in measurements of real stars */
	double scatter_bin_size   /* I: bin stars by magnitude to estimate */
                             /*      scatter as func of mag empirically */
	)
{
	int i;
	int num_in_bin, num_bins;
	int *bin_num_stars;
	double min_mag, max_mag;
	double bin_bottom, bin_top, bin_size;
	double min_bin_mag, max_bin_mag;
	double *bin_mag_value;
	double median, range;
	double range_width;
	double *median_array, *range_array;
	double median_a, median_b, median_c;
	struct s_star *star;
	S_SCATTER *scatter_array;


	/* determine the limiting magnitudes: bright (min_mag) and faint (max_mag) */
	min_mag = 1e6;
	max_mag = -1e6;
	for (i = 0; i < num_stars; i++) {
		star = &(star_arr[i]);
		if (star->include == 1) {
			if ((star->mag0 != NODATA) && (star->sigma != NODATA)) {
				if (star->mag0 < min_mag) {
					min_mag = star->mag0;
				}
				if (star->mag0 > max_mag) {
					max_mag = star->mag0;
				}
			}
		}
	}
#ifdef DEBUG
	printf("  min_mag %10.5f   max_mag %10.5f \n", min_mag, max_mag);
#endif

	/* figure out how many bins we will need */
	min_bin_mag = floor(min_mag);
	max_bin_mag = ceil(max_mag);
	bin_size = scatter_bin_size;
	/* we'll reset this value later, but this is an upper limit */
	num_bins = ceil((max_bin_mag - min_bin_mag)/bin_size) + 1;

	/* allocate space */
	if ((scatter_array = (struct s_scatter *) 
					calloc(num_stars, sizeof(struct s_scatter))) == NULL) {
		fprintf(stderr, "%s: define_scatter: calloc fails for scatter_array \n", 
									progname);
		return(1);
	}
	if ((bin_mag_value = (double *) calloc(num_bins, sizeof(double))) == NULL) {
		fprintf(stderr, "%s: define_scatter: calloc fails for bin_mag_value\n", 
									progname);
		return(1);
	}
	if ((bin_num_stars = (int *) calloc(num_bins, sizeof(int))) == NULL) {
		fprintf(stderr, "%s: define_scatter: calloc fails for bin_num_stars\n", 
									progname);
		return(1);
	}
	if ((median_array = (double *) calloc(num_bins, sizeof(double))) == NULL) {
		fprintf(stderr, "%s: define_scatter: calloc fails for median_array\n", 
									progname);
		return(1);
	}
	if ((range_array = (double *) calloc(num_bins, sizeof(double))) == NULL) {
		fprintf(stderr, "%s: define_scatter: calloc fails for range_array\n", 
									progname);
		return(1);
	}

	/* walk through all bins, calculating a "median" and "scatter" for each */
	num_bins = 0;
	for (bin_bottom = min_bin_mag; (bin_bottom + scatter_bin_size) < max_bin_mag;
					bin_bottom += bin_size) {
		bin_top = bin_bottom + scatter_bin_size;

#ifdef DEBUG
		printf("  bin_bottom %7.2f  bin_top %7.2f \n", bin_bottom, bin_top);
#endif


		/* fill the scatter_array[] with stars in this magnitude bin */
		num_in_bin = 0;
		for (i = 0; i < num_stars; i++) {
			star = &(star_arr[i]);
			if (star->include == 1) {
				if ((star->mag0 != NODATA) && (star->sigma != NODATA)) {
					if ((star->mag0 >= bin_bottom) && (star->mag0 < bin_top)) {
						scatter_array[num_in_bin].mag = star->mag0;
						scatter_array[num_in_bin].sigma = star->sigma;
						num_in_bin++;
						/* sanity check */
						if (num_in_bin > num_stars) {
							fprintf(stderr, "%s: define_scatter: %d stars in bin?!?\n",
										progname, num_in_bin);
							return(1);
						}
					}
				}
			}
		}
#ifdef DEBUG
		for (i = 0; i < num_in_bin; i++) {
			printf("  star %6d    mag0 %10.5f  sigma %10.5f \n",
								i, scatter_array[i].mag, scatter_array[i].sigma);
		}
#endif

		/* 
		 * calculate the median of sigma values in this bin
		 *   and the range between first and third quartiles.
		 */
		if (calc_stats(num_in_bin, scatter_array, &median, &range) != 0) {
			fprintf(stderr, "%s: define_scatter: calc_stats fails ?! \n", progname);
			return(1);
		}

		/* save these values for later use */
#ifdef DEBUG
		printf("  num_bins %6d  num_in_bin %6d \n", num_bins, num_in_bin);
#endif
		bin_mag_value[num_bins] = 0.5*(bin_bottom + bin_top);
		bin_num_stars[num_bins] = num_in_bin;
		median_array[num_bins] = median;
		range_array[num_bins] = range;
		num_bins++;

	}


	/* fit model to the "median" values as function of magnitude */
	if (fit_median(num_bins, bin_num_stars, bin_mag_value, median_array, 
							&median_a, &median_b, &median_c) != 0) {
		fprintf(stderr, "%s: define_scatter: fit_median returns with error \n",
								progname);
		return(1);
	}

#ifdef DEBUG
	{
		int i;
		for (i = 0; i < num_bins; i++) {
			printf("B:   bin %6d has %6d stars \n", i, bin_num_stars[i]);
		}
	}
#endif

	/* 
	 * fit model to the "median+scatter" values as function of magnitude.
	 * Actually, it appears that a very simple "model" will suffice:
	 * treat the width of the scatter values as constant, across all
	 * magnitudes.  It's hard to do any better than this given the
	 * very few stars available at the bright end.
	 */
	if (fit_range(num_bins, bin_num_stars, bin_mag_value, range_array,
							&range_width) != 0) {
		fprintf(stderr, "%s: define_scatter: fit_range returns with error \n",
								progname);
		return(1);
	}

	/* mark stars as variable */
	if (mark_variable(nsigma, min_scatter, num_bins, bin_mag_value, 
							median_array) != 0) {
		fprintf(stderr, "%s: define_scatter: mark_variable returns with error \n",
								progname);
		return(1);
	}

	/* all done.  de-allocate space */
	free(scatter_array);
	free(bin_num_stars);
	free(bin_mag_value);
	free(median_array);
	free(range_array);

	return(0);
}




	/***********************************************************************
	 * PROCEDURE: calc_stats
	 *
	 * DESCRIPTION: Given an array of magnitudes and sigmas for stars
	 *              within a single bin of magnitudes, calculate two
	 *              quantities:
	 * 
	 *                   a) the median of the sigma values
	 *                   b) a "width" of the sigma values
	 *
	 *              The median is straighforward.  For the "width",
	 *              we use the interquartile range (between quartiles
	 *              1 and 3), if possible.
	 * 
	 *              However, if there are fewer than a critical number
	 *              of stars in this bin, then we revert to "a reasonable
	 *              value" instead.
	 *  
	 * RETURNS:
	 *     0        if all goes well
	 *     1        if an error occurs
	 */

static int 
calc_stats
	(
	int num_in_bin,       /* I: number of stars with data in the array */
	S_SCATTER *scat_arr,  /* I: array of s_scatter structures */
	double *median,       /* O: median value of sigma for stars in array */
	double *width         /* O: "width" range of sigma for stars in array */
	)
{
	int median_index, first_index, third_index;

	/* 
	 * first, we make check: if 0 stars, we return values of -1
	 *   which will signal that this bin should be ignored.
	 */
	if (num_in_bin == 0) {
		*median = -1.0;
		*width = -1.0;
		return(0);
	}

	/* 
	 * next, we make check: if fewer than 3 stars, we do simple
	 * calcs manually here.
	 */
	if (num_in_bin < 4) {
		*median = 0.5*(scat_arr[0].sigma + scat_arr[1].sigma + 
								 scat_arr[2].sigma);
		*width = scat_arr[0].sigma;
		if (scat_arr[1].sigma > *width) {
			*width = scat_arr[1].sigma;
		}
		if (scat_arr[2].sigma > *width) {
			*width = scat_arr[2].sigma;
		}
		return(0);
	}

	/* 
	 * there are at least 4 stars.  Fine.  We use the same algorithm
	 * for all situations from here on.
	 */
	qsort((char *) scat_arr, num_in_bin, sizeof(struct s_scatter),
						(PFI) comp_scatter);
	
	/*
	 * the median value will be entry (N+1)/2 in the sorted array 
	 */
	median_index = floor(2.0*(num_in_bin/4.0));
#ifdef DEBUG
	printf("  calc_stats: median_index %6d \n", median_index);
#endif
	*median = scat_arr[median_index].sigma;

	/*
	 * now, we calculate the indices of the 25'th quarter and the 
	 * 75'th quartile, and use the range in between them for the
	 * "width" of the scatter values.
	 */
	first_index = floor(1.0*(num_in_bin/4.0));
	third_index = floor(3.0*(num_in_bin/4.0));
#ifdef DEBUG
	printf("  calc_stats: first_index %6d  third_index %6d \n", 
				first_index, third_index);
#endif
	*width = scat_arr[third_index].sigma - scat_arr[first_index].sigma;


	return(0);
}


   /**************************************************************
    * PROCEDURE: comp_scatter
    *
    * DESCRIPTION: Compare the "sigma" values for the two S_SCATTER structs.
    *
    * RETURNS:
    *       -1 if the first struct has a sigma smaller than the second
    *        0                                 equal to 
    *       +1                                 larger than
    */

static int
comp_scatter
   (
   struct s_scatter *s1,
   struct s_scatter *s2
	)
{
   if (s1->sigma < s2->sigma) {
      return(-1);
   }
   if (s1->sigma > s2->sigma) {
      return(1);
   }
   return(0);
}



	/***********************************************************************
	 * PROCEDURE: fit_median
	 *
	 * DESCRIPTION: Given an array of "sigma" values for the stars
	 *              in a set of 'num_bin' magnitude bins, try to fit
	 *              a low-order polynomial (parabola) to the data.
	 *              Ignore any bins which have a value of -1; that signals
	 *              that there were no stars in the bin.
	 * 
	 *              We fit a function of the form
	 * 
	 *                   sigma(mag)  =  a  +  b*mag  +  c*mag^2
	 *
	 *              to the data.  We return the values of coefficients
	 *              a, b, c.  
	 * 
	 * RETURNS:
	 *     0        if all goes well
	 *     1        if an error occurs
	 */


	/* 
	 * TINY_WEIGHT should be a number small compared to 1.  
	 * We use it as the weight given to the curve-fitting routines
	 * for a magnitude bin which has no stars.
	 */
#define TINY_WEIGHT  0.01

static int 
fit_median
	(
	int num_bins,          /* I: number of elements of array with data to fit */
	int *num_stars_array,  /* I: array of number of stars in each bin */
	double *mag_array,     /* I: array of mag values in middle of each bin */
	double *median_array,  /* I: array of median values, one per bin of mag */
	double *a_coeff,       /* O: fitted parabola coeff: const term */
	double *b_coeff,       /* O: fitted parabola coeff: linear term */
	double *c_coeff        /* O: fitted parabola coeff: quadratic term */
	)
{
	int i;
	double weight;
	double chisq;
	gsl_matrix *X, *cov;
	gsl_vector *y, *w, *c;


	/* 
	 * first, we make check: if fewer than 3 bins, we can't fit a
	 *   parabola.  In that case, we calculate the scatter for all
	 *   values in the few bins we have, and return parameters
	 *   which imply a constant scatter:
	 *
	 *          a_coeff   =     avg of all median values we have
	 *
	 *          b_coeff   =     0
	 *          c_coeff   =     0
	 *
	 * If there are ZERO values, then we place a value of 1.0 into
	 * the *a_coeff output argument.  That shouldn't happen ....
	 */
	if (num_bins < 3) {
		double num, sum, avg_median;

		num = 0.0;
		sum = 0.0;
		for (i = 0; i < num_bins; i++) {
			if (num_stars_array[i] > 0) {
				sum += median_array[i];
				num += 1.0;
			}
		}
		if (num < 1) {
			avg_median = 1.0;
			fprintf(stderr, "fit_median: zero median values?  Using 1.0, watch out \n");
		}
		else {
			avg_median = sum/num;
		}

		*a_coeff = avg_median;
		*b_coeff = 0.0;
		*c_coeff = 0.0;

#ifdef DEBUG
		printf(" fit_median: a %12.7f b %12.7f c %12.7f  too few bins \n", 
						*a_coeff, *b_coeff, *c_coeff);
#endif
		return(0);
	}

	/*
	 * allocate space for the variables we'll need to run the GSL
	 * curve-fitting routines.
	 */
	X = gsl_matrix_alloc(num_bins, 3);
	y = gsl_vector_alloc(num_bins);
	w = gsl_vector_alloc(num_bins);
	c = gsl_vector_alloc(3);
	cov = gsl_matrix_alloc(3, 3);

	/* 
	 * fill the vectors and matrices with information from our data 
	 */
	for (i = 0; i < num_bins; i++) {

		gsl_matrix_set(X, i, 0, 1.0);
		gsl_matrix_set(X, i, 1, mag_array[i]);
		gsl_matrix_set(X, i, 2, mag_array[i]*mag_array[i]);

		gsl_vector_set(y, i, median_array[i]);

		/* 
		 * now, the weight we give to each value could be important.
		 * Here's the plan:
		 *     if the median_array[i] value is -1, it means that there 
		 *           were no stars in that particular bin.  So, we 
		 *           use a very very small weight.
		 *     otherwise, we use a weight which grows slowly with the
		 *           number of stars in the bin.
		 * 
		 * Note that the GSL routines expect a weight which is LARGE for
		 * values with small uncertainty, and SMALL for values with 
		 * large uncertainty.  An example in the GSL manual shows
		 * 
		 *          weight   =   1.0 /  (uncert*uncert)
		 *
		 */
		if ((median_array[i] < 0.0) || (num_stars_array[i] < 1)) {
			weight = TINY_WEIGHT;
		} 
		else {
			weight = 1.0 + log((double) num_stars_array[i]);
		}
		gsl_vector_set(w, i, weight);
		
	}

	/*
	 * now allocate some workspace memory for the fitting routines,
	 *   call the fitting routines,
	 *   and then clean up the workspace
	 */
	{
		gsl_multifit_linear_workspace *work = 
				gsl_multifit_linear_alloc(num_bins, 3);

		gsl_multifit_wlinear(X, w, y, c, cov, &chisq, work);

		gsl_multifit_linear_free(work);
	}
	
	/* 
	 * pick out the best-fit parameters so that we can return them
	 */
	*a_coeff = gsl_vector_get(c, 0);
	*b_coeff = gsl_vector_get(c, 1);
	*c_coeff = gsl_vector_get(c, 2);

#ifdef DEBUG
	printf(" fit_median: a %12.7f b %12.7f c %12.7f  chisq %12.7f \n", 
						*a_coeff, *b_coeff, *c_coeff, chisq);
#endif

	/*
 	 * for very verbose case, print out the full table of sigma per bin,
 	 *    as well as the model fit
 	 */
#ifdef DEBUG5
	{ 
		int i;
		double model;

		for (i = 0; i < num_bins; i++) {
			model = *a_coeff + (*b_coeff*mag_array[i]) +
							(*c_coeff*mag_array[i]*mag_array[i]);
			printf(" bin %3d N %5d mag %6.3f  med %6.3f  model %6.3f \n",
						i, num_stars_array[i], mag_array[i], median_array[i], model);
		}
	}
#endif

	return(0);
}


	/***********************************************************************
	 * PROCEDURE: fit_range
	 *
	 * DESCRIPTION: We are given an array of "range" values for the stars
	 *              in a set of 'num_bin' magnitude bins.
	 *              The "range" is a rough measure of the width of the
	 *              stellar locus in a plot of sigma-vs-magnitude.
	 *              This routine fits a model to the range as a function
	 *              of magnitude.
	 * 	
	 *              Actually, at the moment, our model is very simple:
	 *              a constant width, for all magnitudes.  Hence,
	 *              we just take the average of all the values.
	 *              But this routine is set up to do a more complicated
	 *              fit in the future if it should be necessary.
	 * 
	 *              Ignore any bins which have a value of -1; that signals
	 *              that there were no stars in the bin.
	 * 
	 *              We place the constant width into the output argument.
	 * 
	 * RETURNS:
	 *     0        if all goes well
	 *     1        if an error occurs
	 */


static int 
fit_range
	(
	int num_bins,          /* I: number of elements of array with data to fit */
	int *num_stars_array,  /* I: array of number of stars in each bin */
	double *mag_array,     /* I: array of mag values in middle of each bin */
	double *range_array,   /* I: array of range values, one per bin of mag */
	double *a_coeff        /* O: fitted model width */
	)
{
	int i;
	double sum, num, mean;

#ifdef DEBUG
	{
		int i;

	 	printf(" entering fit_range: num_bins %6d \n", num_bins);
		for (i = 0; i < num_bins; i++) {
			printf(" bin %6d has %6d stars mag %7.2f \n", i, 
							num_stars_array[i], mag_array[i]);
		}
	}
#endif

	sum = 0.0;
	num = 0.0;
	for (i = 0; i < num_bins; i++) {
		if (num_stars_array[i] > 0) {
			sum += range_array[i]*num_stars_array[i];
			num += num_stars_array[i];
		}
	}

	/* calculate the mean value, weighted by number of stars in each bin */
	if (num < 1) {
		fprintf(stderr, "%s: fit_range: given zero stars? \n", progname);
		return(1);
	}

	mean = sum/num;
	*a_coeff = mean;

#ifdef DEBUG
	printf(" fit_range:  mean %12.7f \n", *a_coeff);
#endif

	return(0);
}




	/***********************************************************************
	 * PROCEDURE: mark_variable
	 *
	 * DESCRIPTION: We have already run an ensemble solution,
	 *              so we have the scatter-from-mean value for each star.
	 *              That information is passed in a pair of arrays
	 *              as input arguments to this function.
	 *
	 *              This function looks at each star's "sigma" value:
	 *              if it is more than "nsigma" times the typical value
	 *              for scatter at its brightness,
	 *              then 
	 *                  - we set the "variable" field of star struct to 1
	 *                  - we set the "nsigma" field of star struct to the
	 *                        ratio (this star's scatter / typical scatter)
	 * 
	 *              The idea is that we can quickly sort through the output
	 *              by the "variable" and/or "nsigma" values to find
	 *              candidates for variability.
	 *
	 *              The "min_scatter" argument serves to avoid awkward
	 *              situations (divide by zero) if the empirical stdev from 
	 *              mean for a star just happens to be zero, or implausibly tiny.
	 *              We replace such tiny values with this value.
	 * 
	 * RETURNS:
	 *     0        if all goes well
	 *     1        if an error occurs
	 */


static int 
mark_variable
	(
	double nsigma,            /* I: threshold for marking as variable */
	double min_scatter,       /* I: minimum possible scatter for a star */
	int num_bin,              /* I: number of bins of mag, median values */
	double *bin_mag_value,    /* I: array of mag in middle of each bin */
	double *median_array      /* I: array of typical scatter in each bin */
	)
{
	int i, i_bin, left_bin, right_bin;
	double mag, model_sigma, threshold;
	struct s_star *star;

	for (i = 0; i < num_stars; i++) {
		star = &(star_arr[i]);
		if ((star->mag0 != NODATA) && (star->sigma != NODATA)) {

			/* 
			 * calculate the expected scatter for a star of the same
			 * mean magnitude 
		 	 */
			mag = star->mag0;

#ifdef DEBUG2
			printf("   want to make model for mag = %6.3lf \n", mag);
#endif
			/* find the bin centers to left and right of this magnitude */
			left_bin = -1;
			right_bin = -1;
			for (i_bin = 0; i_bin < num_bin; i_bin++) {
				if (mag < bin_mag_value[i_bin]) {
					left_bin = i_bin - 1;
					right_bin = i_bin;
					break;
				}
			}
#ifdef DEBUG2
			printf("   left_bin %2d %6.3f  mag %6.3f  right_bin %2d %6.3f \n", 
						left_bin, bin_mag_value[left_bin], mag,
						right_bin, bin_mag_value[right_bin]);
#endif
			/* special case for very bright mag */
			if ((left_bin == -1) && (right_bin == 0)) {
				model_sigma = median_array[0];
#ifdef DEBUG2
				printf("   special case A: model_sigma = %lf \n", model_sigma);
#endif
			} 
			/* special case for very faint mag */
			else if ((left_bin == -1) && (right_bin == -1)) {
				model_sigma = median_array[num_bin-1];
#ifdef DEBUG2
				printf("   special case B: model_sigma = %lf \n", model_sigma);
#endif
			}
			else {
				/* interpolate linearly between these two bin centers */
				double x, frac;
				
				frac = (mag - bin_mag_value[left_bin]) / 
						(bin_mag_value[right_bin] - bin_mag_value[left_bin]);
				x = median_array[left_bin] + 
						frac*(median_array[right_bin] - median_array[left_bin]);
				model_sigma = x;
#ifdef DEBUG2
				printf("   interpolate: frac %5.3f model_sigma = %lf \n",
							frac, model_sigma);
#endif
			}
			/* 
			 * prevent tiny values (or values of zero) by enforcing a
			 *    floor on the value of the scatter
			 */
			if (model_sigma < min_scatter) {
				model_sigma = min_scatter;
			}

			/* 
			 * calculate the threshold above which this star's "sigma"
			 * value must be to qualify as "variable" 
			 */
			threshold = nsigma*model_sigma;

			/* 
			 * Does this star qualify?
			 */
			if (star->sigma > threshold) {
				star->variable = 1;
			}

			/* 
			 * Set the star's "nsigma" field to its multiple of 
			 * the typical scatter for a star of its magnitude
			 */
			/*	sanity check */
			star->nsigma = star->sigma / model_sigma;

#ifdef DEBUG
			printf("  star %6d  mag0 %10.5f sigma %10.5f  model %6.3f nsigma %6.2f  var %d \n",
						i, star->mag0, star->sigma, model_sigma, 
						star->nsigma, star->variable);
#endif

		}
	}


	return(0);
}


	/***********************************************************************
	 * PROCEDURE: calc_mag_shift
	 *
	 * DESCRIPTION: By default, the output magnitudes of stars in the
	 *              ensemble solution are set so that the brightest
	 *              member has mean mag zero.  If the user wishes, we
	 *              can shift the magnitudes so that they are close to 
	 *              the input magnitudes.  Here's how:
	 * 
	 *                 a) look at the subset of stars which have 
	 *                    ensemble magnitudes between
	 *                           (0 + mrange_low)  and  (0 + mrange_high)
	 *                        
	 *                 b) ignore any stars marked as "bad" or "variable"
	 *
	 *                 c) calculate the unweighted mean difference
	 *                    for each star
	 *                           delta  =  input_mag - ensemble_mean
	 * 
	 *                 d) find the average of all deltas
	 *
	 *              We return this average shift as an output argument;
	 *              it will be applied later, when we print out the 
	 *              magnitude information.
	 * 
	 * RETURNS:
	 *     0        if all goes well
	 *     1        if an error occurs
	 */


static int 
calc_mag_shift
	(
	double mrange_low,        /* I: bright mag for stars used to calc delta */
	double mrange_high,       /* I: faint mag for stars used to calc delta */
	double *delta_mag         /* O: size of the shift to be added to all mags */
	)
{
	int i, delta_num;
	double delta_sum, delta;
	struct s_star *star;
	struct s_measurep *mp;
	struct s_measure *measure;

	/* sanity check */
	if (mrange_low > mrange_high) {
		fprintf(stderr, "%s: shift_mags: low %lf  is greater than high %lf ?! \n",
					progname, mrange_low, mrange_high);
		return(1);
	}

	delta_sum = 0;
	delta_num = 0;
	for (i = 0; i < num_stars; i++) {
		star = &(star_arr[i]);
		if ((star->variable != 0) || (star->include == 0)) {
			continue;
		}
		if ((star->mag0 == NODATA) || (star->num_measures == 0)) {
			continue;
		}

		if ((star->mag0 < mrange_low) || (star->mag0 > mrange_high)) {
			continue;
		}

		mp = star->mp;
		while (mp != NULL) {
			shAssert((measure = mp->measure) != NULL);
			if (measure->include == 0) {
				mp = mp->next;
				continue;
			}
			delta = measure->obs_mag - star->mag0;
			delta_sum += delta;
			delta_num++;

			mp = mp->next;
		}
	}

	/* 
	 * now calculate the overall shift, if we can.
	 * If there isn't enough information, then we do nothing,
	 * and return withOUT error code.
	 */
	if (delta_num < 1) {
		return(0);
	}	
	*delta_mag = delta_sum/(double) delta_num;
#ifdef DEBUG
	printf("  delta_num %6d  delta_sum %12.2lf  delta_mag %7.3f \n",
					delta_num, delta_sum, *delta_mag);
#endif

	return(0);
}



	/***********************************************************************
	 * PROCEDURE: calc_imag_shift
	 *
	 * DESCRIPTION: By default, the output magnitudes of stars in the
	 *              ensemble solution are set so that the brightest
	 *              member has mean mag zero.  If the user wishes, we
	 *              can shift the magnitudes so that they are close to 
	 *              the input magnitudes.  
	 *
	 *              So, this function is very much like "calc_mag_shift()",
	 *              above.  But the difference lies in how we pick stars.
	 *              The "calc_mag_shift()" func picks stars based on their
	 *              ensemble output differential magnitudes -- which 
	 *              means, basically, "1 mag below brightest in ensemble
	 *              to 2 mag below brightest in ensemble".
	 *
	 *              This func, on the other hand, picks stars based on
	 *              the _input_ magnitudes.  So, if we know some properties
	 *              of the input dataset (for example, that stars with
	 *              mags between 8 and 10 are good, non-saturated), then
	 *              we can say "pick stars with input mags between 8 and 10
	 *              to define the shift from output differential mags
	 *              back to something like input mags."
	 * 
	 *              Here's how we do it.
	 *
	 *                 a) look at the subset of stars which have 
	 *                    input magnitudes between
	 *                           (irange_low)  and  (irange_high)
	 *                        
	 *                 b) ignore any stars marked as "bad" or "variable"
	 *
	 *                 c) calculate the unweighted mean difference
	 *                    for each star
	 *                           delta  =  input_mag - ensemble_mean
	 * 
	 *                 d) find the average of all deltas
	 *
	 *              We return this average shift as an output argument;
	 *              it will be applied later, when we print out the 
	 *              magnitude information.
	 * 
	 * RETURNS:
	 *     0        if all goes well
	 *     1        if an error occurs
	 */


static int 
calc_imag_shift
	(
	double irange_low,        /* I: bright mag for stars used to calc delta */
	double irange_high,       /* I: faint mag for stars used to calc delta */
	double *delta_mag         /* O: size of the shift to be added to all mags */
	)
{
	int i, delta_num;
	double delta_sum, delta;
	double sum, num;
	double avg_input_mag;
	struct s_star *star;
	struct s_measurep *mp;
	struct s_measure *measure;

	/* sanity check */
	if (irange_low > irange_high) {
		fprintf(stderr, "%s: calc_imag_shift: low %lf  is greater than high %lf ?! \n",
					progname, irange_low, irange_high);
		return(1);
	}

	delta_sum = 0;
	delta_num = 0;
	for (i = 0; i < num_stars; i++) {
		star = &(star_arr[i]);
		if ((star->variable != 0) || (star->include == 0)) {
			continue;
		}
		if ((star->mag0 == NODATA) || (star->num_measures == 0)) {
			continue;
		}

		/* 
		 * okay, this star might qualify.  Let us calc its average
		 * input magnitude value.
		 */
		mp = star->mp;
		sum = 0.0;
		num = 0;
		while (mp != NULL) {
			shAssert((measure = mp->measure) != NULL);
			if (measure->include == 0) {
				mp = mp->next;
				continue;
			}
			sum += measure->obs_mag;
			num++;

			mp = mp->next;
		}

		if (num < 1) {
			/* we can't use this star --- no valid input mags */
			continue;
		}

		/* this is the average input magnitude of the star */
		avg_input_mag = sum/num;

		/* check to see if it falls within the requested range */
		if ((avg_input_mag < irange_low) || (avg_input_mag > irange_high)) {
			continue;
		}
			
		/* 
		 * okay, this star qualifies.  Calc the difference between 
		 * its average input magnitude, and its differential magnitude
		 * in the output solution.
		 */
		delta = avg_input_mag - star->mag0;
		delta_sum += delta;
		delta_num++;

	}

	/* 
	 * now calculate the overall shift, if we can.
	 * If there isn't enough information, then we do nothing,
	 * and return withOUT error code.
	 */
	if (delta_num < 1) {
		*delta_mag = 0.0;
	}	
	else {
		*delta_mag = delta_sum/(double) delta_num;
	}
#ifdef DEBUG
	printf("  delta_num %6d  delta_sum %12.2lf  delta_mag %7.3f \n",
					delta_num, delta_sum, *delta_mag);
#endif

	return(0);
}















