


	/*******************************************************************
	 * Definitions for constants and structures used in the "solvepht"
	 * program, which reads in the re-arranged measurements
	 * and figures out a big least-squares photometric solution.
	 *
	 * This is the second step in a analysis procedure.  The first step
	 * (multipht) re-arranges the snapshot data to put all measurements
	 * of the same star in one place.
	 *
	 * MWR 9/8/2004
	 */

   /*
    * This is a preprocessor macro, which acts like a function.  The user
    * calls it with one argument, a condition which should evaluate to
    * 0 or 1.  If it evaluates to 1, then nothing happens; but if it
    * evaluates to 0, then the program prints an error message, giving
    * location of the error, and halts execution with an error code.
    * 
    * Thus, one typically uses it to make a 'sanity check', such as
    * making sure that a pointer (which really, really should have
    * a valid value) isn't NULL:
    *
    *     fp = fopen("file", "r");
    *        ...
    *     shAssert(fp != NULL);
    *     fgets(line, LINELEN, fp);
    *        ...
    */
#define shAssert(x) if((x)!=1){fprintf(stderr,"assertion fails in file %s, line %d\n",__FILE__,__LINE__);exit(1);}
 

#define MAX_STAR     10000	    /* max number stars total we can have */
#define MAX_IMAGE     3000     /* max number of images we can have */
#define STRLEN         256     /* max length of star names and filters */
#define LINELEN        256
#define NODATA          99.0   /* value assigned to a variable w/ no data */
#define MAG_CONST        0.0   /* set brightest star to this mag in output */

	/*
	 * We perform a number of calculations based on the scatter 
	 * exhibited by stars of a certain brightness.  It is possible that
	 * the scatter may occasionally be zero, due to a small number
	 * of measurements.  To prevent dividing by zero, we use this
	 * as a lower bound on any estimate of scatter.
	 */
#define MIN_POSSIBLE_SCATTER  0.010

	/* 
	 * if the user wishes us to guess which stars in the field might
	 * be variable, we make a model for the trend of sigma-vs-mag,
	 * and an estimate of the width of the scatter around the model;
	 * then we calculate this many times that scatter, and mark as
	 * variable stars which have a larger sigma value.
	 *
	 * In other words, we look for outliers which are more than
	 * SCATTER_NSIGMA times the typical width above the mean value.
	 *
	 *         if   (sigma)  >  (model_sigma) + SCATTER_NSIGMA*(width)
	 *                  mark as variable
	 *      
	 * This is the default value -- the user can modify it via the
	 * "nsigma=" option on the command line.
	 */
#define SCATTER_NSIGMA    3.0

	/*
	 * How many times should we iterate over the solution by default?
	 * Each time we run solution, we may mark some stars as "variable",
	 * which are then not used in determining the next iteration(s).
	 */
#define NUM_ITERATIONS    0

	/*
	 * The user may wish to shift the output magnitudes:
	 * by default, they are normalized so that the brightest star	
	 * in the ensemble has magnitude zero.  If the user supplies
	 * the "mrange" command-line option, then we will pick a set
	 * of stars with _output_ ensemble magnitudes between
	 * two limits.  The values below are the limits by default,
	 * but the user can override them by specifying values
	 * in the "mrange" argument.
	 */
#define MRANGE_LOW   1.0
#define MRANGE_HIGH  2.5

   /*
    * Similar to the "MRANGE" options above, these variables
    * define the range of stars which are used to re-set the
    * solution's differential magnitudes to some output magnitudes.
    * In this case, however, the values refer to the INPUT magnitudes
    * of stars, rather than to the output differential magnitudes.
    */
#define IRANGE_LOW   8.5
#define IRANGE_HIGH  10.0



	/*
	 * In order to flag some stars as possible variable candidates,
	 * we need to know the scatter around mean for "ordinary" stars.
	 * Since this scatter will be large for faint stars and small
	 * for bright stars, we divide all stars up into bins by
	 * magnitude, and then calc stats within each bin.  This sets
	 * the width of those bins, in magnitude.  Is used in 
	 * define_scatter() function, in solvepht.c
	 */
#define SCATTER_BIN_SIZE  0.50

	/*
	 * If a star is variable, we can't rely on the scatter of
	 * its measurements from mean to determine the uncertainty
	 * in each individual measurement.  So, we look for stars
	 * of similar mean magnitude, and use them to estimate the
	 * uncertainty per measurement.  We accept any star within
	 * this many magnitudes of the variable star's mean mag.
	 * Used in "find_sigm()" in solvematrix.c.
	 */
#define MAG_MATCH        0.60


	/* some flags which tell the program its function */
#define PRINT     1			
#define AUTO      2			/* don't ask for input from user */

#define LONG      1        /* print lots of output */
#define SHORT     2        /* print less output */

typedef int (*PFI)();


struct s_image {
	int index;				/* identifying index used in output routines */
	char name[STRLEN];
	char object[STRLEN];
	double exptime;
	double jd;
	int num_measures;			/* number of stars in the image */
	struct s_measurep *mp;	/* linked list of measurements in this image */
	int num_match;			/* # stars that match template image */
	double xshift;       /* x-shift in mm to match fiducial image */
	double yshift;       /* y-shift in mm to match fiducial image */
	int include;			/* if 1 include stars from this image; if 0 don't */
};


	/* a star --- one structure per physical object in space */
struct s_star {
	int id;						/* ID among all stars */
	double ra;              /* position along one axis.  Not necessarily RA, */
	                        /*    but this is a convenient label */
	double dec;             /* position along the other axis.  Not */
	                        /*    necessarily Dec, but a reasonable label */
	double mag0;            /* mean magnitude in solution output */
	double sigma;           /* scatter from mean in solution output */
	double nsigma;          /* significance of scatter above typical value */
	int num_measures;       /* number of times it was detected */
	struct s_measurep *mp;  /* linked list of measurements of this star */
	int include;            /* include in solution? 1=yes, 0=no */
	int variable;				/* yes/no treat this star as a variable */
};


	/* one measurement from an image */
struct s_measure {
	int measure_id;         /* ID among all measures */
	int image_id;				/* ID number on this measure's image */
	int star_id;				/* ID number of the star corresponding to this */
	struct s_image *image;
	struct s_star *star;
	double ra;              /* position along one axis (may not be RA) */
	double dec;             /* position along the other axis (may not be Dec) */
	double obs_mag;         /* measured mag of this star */
	double obs_sigma;       /* uncertainty in measured mag */
	int w3;						/* yes/no weight for this star, this image */
	double w4;					/* weight calculated from signal/noise */
	int saturated;				/* yes/no is this measure saturated? */
	int include;            /* include this measure in solution? 1=yes, 0=no */
};



	/*
	 * we use this structure as a container and build linked lists 
	 * out of it.  Each structure points to one measurement of a star.
	 */

struct s_measurep {
	struct s_measure *measure;
	struct s_measurep *next;
};



	/* these are the heads of some important arrays/lists */

struct s_image *image_arr;
struct s_measurep *measure_list;
struct s_star *star_arr;

	/* some global variables */

int num_images;               /* total number of images */
int num_stars;                /* total number of unique stars */
int num_measures;             /* total number of measurements */
int debug_flag;

	/* some global functions */
/* solveread.c */
int read_input(char *file, double min, double max, 
               double varmin, double varmax, int vardet);
/* solveprint.c */
void print_stars(FILE *fp, int len);
void print_images(FILE *fp);
/* solvematrix.c */
void solve_matrix(double zero_point, double mag_match);
int write_images(char *file);
int write_sigma(char *file, double delta_mag);
int write_stars(char *file, double delta_mag);



