
	/************************************************************************
	 * routines to print out lists of stars and images, either before
	 * or after we have solved the big linear equations.
	 */

#include <stdio.h>
#include <math.h>
#include "solvepht.h"

#define PAGE_LEN  20	/* # lines to print between page prompts */
#define STOP      1		/* don't print any more */
#define CONTINUE  2		/* keep printing */


static int print_one_star(FILE *fp, struct s_star *p, int len);
static int print_one_image(FILE *fp, struct s_image *im);
static int page_scroll(void);



	/*************************************************************************
	 * PROCEDURE: print_stars
	 *
	 * DESCRIPTION: print information on each star: they should be in order 
	 *              by master list x-coordinate 
	 *
	 * RETURNS:
	 *    nothing
	 */

void
print_stars
	(
	FILE *fp, 
	int len
	)
{
	int i;

	for (i = 0; i < num_stars; i++) {
		if (print_one_star(fp, &(star_arr[i]), len) == STOP) {
			break;
		}
	}	
}



	/************************************************************************
	 * PROCEDURE: print_one_star
	 *
	 * DESCRIPTION: Print the pertinent information from a single star entry
	 *              (which records all instances of the same star) 
	 *              to the given FILE pointer.  Each 'master star' 
	 *              gets N+1 lines of output, where N is the number of 
	 *              actual instances of the star.  
	 *              The first line includes the number N at its end.  
	 *
	 * RETURNS:
	 *    return the result of page_scroll(). 
	 *
	 */

static int
print_one_star
	(
	FILE *fp, 
	struct s_star *star, 
	int len
	)
{
	int i;
	struct s_measurep *mp;
	struct s_measure *measure;

	fprintf(fp, "%6d   %9.5f %9.5f %6d %d %d\n", 
						star->id, star->ra, 
						star->dec, 
						star->num_measures, star->include, star->variable);

	if (len == SHORT) {
		return(page_scroll());
	}
	else {
		if (page_scroll() == STOP) {
			return(STOP);
		}
	}
	
	for (mp = star->mp, i = 0; i < star->num_measures; i++) {
		if (mp == NULL)  {
			break;
		}
		measure = mp->measure;
		if (measure == NULL) {
			break;
		}
		fprintf(fp, 
			"    %6d %6d %4d %9.1f %9.1f %6.3f %7.1f %1d \n",
						measure->star_id, measure->measure_id, measure->image_id, 
						measure->ra, measure->dec,
						measure->obs_mag, measure->w4,
						measure->include);
		mp = mp->next;
		if (page_scroll() == STOP) {
			return(STOP);
		}
	}

	return(CONTINUE);
}


	/***********************************************************************
	 * PROCEDURE: print_images
	 *
	 * DESCRIPTION: Print information on each image, which should be 
	 *              in order by time of observation.
	 *
	 * RETURNS:
	 *     nothing
	 */

void
print_images
	(
	FILE *fp
	)
{
	int i;

	for (i = 0; i < num_images; i++) {
		if (print_one_image(fp, &(image_arr[i])) == STOP) {
			break;
		}
	}	

}


	/********************************************************************
	 * PROCEDURE: print_one_image
	 *
	 * DESCRIPTION: Print the pertinent information in a single image 
	 *              structure to the given FILE pointer
	 *
	 * RETURNS:
	 *  the result of page_scroll() 
	 */

static int
print_one_image
	(
	FILE *fp,
   struct s_image *im
	)
{

	fprintf(fp, "%5d %s %7.2f %14.7f %6d %6d %d\n",
			im->index, im->name, im->exptime, 
			im->jd, im->num_measures, im->num_match, im->include);
	return(page_scroll());
}




	/************************************************************************
	 * PROCEDURE: page_scroll
	 *
	 * DESCRIPTION: Print a message on the screen and wait for 
	 *              the user to type 'return'; depending on the user's 
	 *              response, return STOP or CONTINUE
	 *
	 * RETURNS:
	 *    STOP      if the user wants to stop
	 *    CONTINUE  if the user wants to keep going
	 */

static int
page_scroll
	(
	void
	)
{
	static int num = 0;
	int c;

	if (++num >= PAGE_LEN) {
		num = 0;
		printf("             --  Type ENTER for more, or 'q' to quit --\n");
		if (((c = getchar()) == 'q') || (c == 'Q')) {
			return(STOP);
		}
		else {
			return(CONTINUE);
		}
	}
	else {
		return(CONTINUE);
	}
}
