
	/*************************************************************************
	 * Read the contents of the nicely arranged "multisnap" output file,
	 * and build up in memory a linked list of images, and a linked list
	 * of stars.
	 *
	 * MWR 10/18/2003
	 *
	 * Created a new function "new_measurep" to allocate space for
	 * _and_ initialize a struct s_measurep.  The key is to set
	 * the 'next' field to NULL when the new structure is created.
	 * Old version did not set 'next', and apparently that can under
	 * rare (?) circumstances cause segfaults.
	 *
	 * MWR 11/15/2016
	 */

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include "solvepht.h"

#undef DEBUG

static void eliminate(void);
static int read_it(FILE *fp);
static int good_images(void);
static int good_stars(void);
static int read_stars(FILE *fp);
static int read_one_star(FILE *fp, struct s_star *p);
static int read_images(FILE *fp);
static int read_one_image(FILE *fp, struct s_image *im);
static void check_mags(double min, double max);
static void check_varmin(double varmin, double varmax);
static void check_vardet(int vardet);
static void add_mp_to_biglist(struct s_measurep *mp);
static void add_mp_to_star(struct s_star *star, struct s_measurep *mp);
static void add_mp_to_image(struct s_image *image, struct s_measurep *mp);
static int check_measurements(void);
static struct s_measurep *new_measurep(void);


	/*
	 * Read information from the multipht output file.  Create the following
	 * global variables:
	 *
	 *       image_arr                        array of s_image
	 *       star_arr                         array of s_star
	 *       measure_list                     linked list of s_measurep
	 *
	 * Each of these items has an associated "num_" variable which will,
	 * after we have finished, count the number of elements; either
	 * in the array, or in the linked list.
	 *
	 * Each element in the
	 *           image_arr, star_arr
	 * has its own private linked list of s_measurep structures, which
	 * in turn point to s_measures in the master linked list.
	 * This way, we don't have to make many copies of the big sets
	 * of information about each measure; we only need to make many copies
	 * of the tiny "s_measurep" structure.
	 */


	/************************************************************************
	 * PROCEDURE: read_input
	 *
	 * DESCRIPTION: read information on all 'acceptable' images and stars from
	 *              the given file (or stdin, if passed an empty string "").
	 *              A star with ANY instrumental mag (m < min) or (m > max) is 
	 *              marked as 'unacceptable'. 
	 *
	 * RETURNS:
	 *     0        if all goes well
	 *     1        if an error occurs
	 */

int
read_input
	(
	char *file, 
	double min, 
	double max,
	double varmin,
	double varmax,
	int vardet
	)
{
	FILE *fp;
	int num_good_images;
	int num_good_stars;

	if (strcmp(file, "") == 0) {
		fp = stdin;
	}
	else if ((fp = fopen(file, "r")) == NULL) {
		fprintf(stderr, "read_input: can't open file %s for reading\n", file);
		return(1);
	}

	if (read_it(fp) != 0) {
		fprintf(stderr, "read_input: read_it returns with error \n");
		return(1);
	}

	fclose(fp);

	/* discard stars with average instrumental mag outside bounds */
	check_mags(min, max);	

	/* 
	 * mark as "variable" stars which have any measures < varmin 
	 *    or > varmax 
	 */
	check_varmin(varmin, varmax);

	/*
	 * mark as "variable" stars which have fewer than "vardet" 
	 *    measurements
	 */
	check_vardet(vardet);

	/* eliminate stars, images identified as bad on command line */
	eliminate();

	/* 
	 * make sure each image has at least one good measurement, 
	 *   and count the good images which remain after we've checked.
	 */
	num_good_images = good_images();

	/* 
	 * make sure each star has at least one good measurement, 
	 *   and count the good stars which remain after we've checked.
	 */
	num_good_stars = good_stars();

#ifdef DEBUG
	printf("there are %d good images \n", num_good_images);
	printf("there are %d good stars \n", num_good_stars);
#endif

	return(0);
}


	/**********************************************************************
	 * PROCEDURE: eliminate
	 *
	 * DESCRIPTION: set to zero the "include" field of all s_measures 
	 *              which are either in an 'unacceptable' image, 
	 *              or an instance of an 'unacceptable' star. 
	 *
	 * RETURNS:
	 *    nothing
	 */

static void 
eliminate
	(
	void
	)
{
	int i;
	int num_removed;
	int num_good_removed;
	struct s_image *image;
	struct s_star *star;
	struct s_measurep *mp;
	struct s_measure *measure;

	num_removed = 0;
	num_good_removed = 0;

	for (i = 0; i < num_images; i++) {
		image = &(image_arr[i]);
		if (image->include == 0) {
			mp = image->mp;
			while (mp != NULL) {
				shAssert((measure = mp->measure) != NULL);
				if (measure->include != 0) {
					num_good_removed++;
				}
				num_removed++;
				measure->include = 0;
				mp = mp->next;
			}
		}
	}
#ifdef DEBUG
	printf("eliminate: img  %6d removed  %6d good removed \n", 
					num_removed, num_good_removed);
#endif

	num_removed = 0;
	num_good_removed = 0;

	for (i = 0; i < num_stars; i++) {
		star = &(star_arr[i]);
		if (star->include == 0) {
			mp = star->mp;
			while (mp != NULL) {
				shAssert((measure = mp->measure) != NULL);
				if (measure->include != 0) {
					num_good_removed++;
				}
				num_removed++;
				measure->include = 0;
				mp = mp->next;
			}
		}
	}
#ifdef DEBUG
	printf("eliminate: star %6d removed  %6d good removed \n", 
					num_removed, num_good_removed);
#endif
		
		
}



	/*************************************************************************
	 * PROCEDURE: read_it
	 *
	 * DESCRIPTION: read the data in the multisnap output file.  It contains
	 *              measurements from all stars, arranged by star (not
	 *              by image).
	 *
	 * RETURNS:
	 *     0        if all goes well
	 *     1        if an error occurs
	 */


static int 
read_it
	(
	FILE *fp
	)
{

	/* 
	 * There are two kinds of information to read: information 
	 * about images (read into the 'image_arr[]' array), and information
	 * about stars (read into the 'star_arr[]' and 'measure_arr[]' arrays). 
	 * the first thing to do is to read how many images and how many
	 * stars will appear in the input. 
	 */

	if (fscanf(fp, "%d images  %d stars \n", &num_images, &num_stars) != 2) {
		fprintf(stderr, "read_it: invalid first line of input file \n");
		return(1);
	}

	/* now read information on each image: get one line per image */
	if (read_images(fp) != 0) {
		fprintf(stderr, "read_it: read_images returns with error \n");
		return(1);
	}

	/* and now read information on each star: get N+1 lines per star,
	   where N is the number of images of each star */
	if (read_stars(fp) != 0) {
		fprintf(stderr, "read_it: read_images returns with error \n");
		return(1);
	}

	/* 
	 * at this point, we can make a sanity check to verify that
	 *   each image has a linked list of all its measurements,
	 *   and each star has a linked list of all its measurements.
	 */
	if (check_measurements() != 0) {
		fprintf(stderr, "read_it: check_measurements returns with error \n");
		return(1);
	}



	return(0);
}


	/************************************************************************
	 * PROCEDURE: good_images
	 *
	 * DESCRIPTION: calculate the number of 'good' images.
	 *
	 * RETURNS:
	 *    the number of good images
	 */

static int 
good_images
	(
	void
	)
{
	int i, sum;

	sum = 0;
	for (i = 0; i < num_images; i++) {
		if (image_arr[i].include != 0) {
			sum++;
		}
	}
	return(sum);
}


	/**********************************************************************
	 * PROCEDURE: good_stars
	 *
	 * DESCRIPTION: calculate the number of 'good' stars.    
	 *              Make sure that there is at least one 'good' 
	 *              instance of each star. 
	 *
	 *              If a star has fewer than 1 good measurement, then
	 *              set the star's "include" field to zero.
	 *
	 * RETURNS:
	 *    number of good stars
	 */

static int
good_stars
	(
	void
	)
{
	int i, sum, good_instances;
	struct s_measurep *mp;
	struct s_measure *measure;

	sum = 0;
	for (i = 0; i < num_stars; i++) { 
		if (star_arr[i].include != 0) {
			good_instances = 0;
			for (mp = star_arr[i].mp; mp != NULL; mp = mp->next) {
				measure = mp->measure;
				if (measure != NULL) {
					if (measure->include != 0) {
						good_instances++;
					}
				}
			}
			/* 
			 * if there are no good instances, un-include the star;
			 *   otherwise, this is a good star, so increment the counter.
			 */
			if (good_instances == 0) {
				star_arr[i].include = 0;
			}
			else {
				sum++;
			}
		}
	}
	
	return(sum);
}


	/**************************************************************************
	 * PROCEDURE: read_stars
	 *
	 * DESCRIPTION: Read information on each star.  They should be in order 
	 *              by master list RA coordinate 
	 *
	 * RETURNS:
	 *    0         if all goes well
	 *    1         if an error occurs
	 */

static int
read_stars
	(
	FILE *fp
	)
{
	int i;

	if ((star_arr = (struct s_star *) 
				calloc(num_stars, sizeof(struct s_star))) == NULL) {
		fprintf(stderr, "read_stars: can't calloc for star_arr array \n");
		return(1);
	}
	for (i = 0; i < num_stars; i++) {
		if (read_one_star(fp, &(star_arr[i])) != 0) {
			fprintf(stderr, "read_stars: read_one_star fails for star %d \n",
					i);
		 	return(1);
		}
	}	

	return(0);
}



	/*************************************************************************
	 * PROCEDURE: read_one_star
	 *
	 * DESCRIPTION: Read the pertinent information from a single star entry
	 *              (which records all measurements of the same star) 
	 *              from the given FILE pointer.  Each star
	 *              gets N+1 lines of output, where N is the number 
	 *              of actual instances of the star.  
	 *              The first line includes the number N at its end.  
	 *
	 *              We create 
	 *                 - one s_measure structure for each measurement
	 *                 - one s_measurep structure for each measurement
	 *
	 *              The s_measurep is a "container" for the s_measure
	 *              data.  We place the s_measurep structures into a 
	 *              linked list which contains ALL the measurements from
	 *              ALL the stars in the dataset.  
	 *
	 *              For this particular star, we are given an s_star
	 *              structure, which will contain a linked list
	 *              of s_measurep structures; these will be separate
	 *              from the main linked list, but will point to the
	 *              same s_measure structures with the data for this
	 *              star.
	 *              
	 *              The s_star structure passed as an arg is one 
	 *              member of an array of s_star structures, containing
	 *              all the stars in the dataset.
	 *
	 *              
	 * RETURNS:
	 *     0        if all goes well
	 *     1        if an error occurs
	 */


static int
read_one_star
	(
	FILE *fp, 
	struct s_star *star
	)
{
	char line[LINELEN];
	int i, index;
	struct s_image *image;
	struct s_measure *measure;
	struct s_measurep *mp;


	shAssert(star != NULL);

	if (fgets(line, LINELEN, fp) == NULL) {
		fprintf(stderr, "read_one_star: fgets returns NULL on first line ?! \n");
		return(1);
	}
	if (sscanf(line, "%d %lf %lf %d\n", &(star->id), &(star->ra), 
						&(star->dec), &(star->num_measures)) != 4) {
		fprintf(stderr, "read_one_star: can't read first line\n");
		return(1);
	}
	star->mag0 = NODATA;
	star->sigma = NODATA;
	star->nsigma = NODATA;
	star->include = 1;
	star->variable = 0;

	for (i = 0; i < star->num_measures; i++) {
		if ((measure = (struct s_measure *) malloc(sizeof(struct s_measure))) == NULL) {
			fprintf(stderr, 
					"read_one_star: can't malloc for instance %d of star %d\n",
							 i, star->id);
			return(1);
		}

		if (fgets(line, LINELEN, fp) == NULL) {
			fprintf(stderr, "read_one_star: fgets returns NULL on line %d ?! \n", 
					i);
			return(1);
		}
      if (sscanf(line, "  %d %d %lf %lf %d %lf %lf \n",
            &(measure->star_id), &(measure->image_id),
				&(measure->ra), &(measure->dec),
            &(measure->saturated), 
            &(measure->obs_mag), 
            &(measure->w4)) != 7) {
			fprintf(stderr, 
				"read_one_star: can't read instance %d of star %d\n",
						i, star->id);
			return(1);
		}

		measure->include = 1;

		index = measure->image_id;
		if ((index < 0) || (index >= num_images)) {
			fprintf(stderr, "star with ID %d has invalid image index %d\n",
					star->id, index);
			return(1);
		}
		image = &(image_arr[index]);
		measure->image = image;

		/* 
		 * create a s_measurep structure to hold this in the big linked list
		 *   of all measurements, and add this s_measurep to the end of the big
		 *   linked list.
		 */
#if OLD
		if ((mp = (struct s_measurep *) malloc(sizeof(struct s_measurep))) == NULL) {
#else
		if ((mp = new_measurep()) == NULL) {
#endif
			fprintf(stderr, "malloc fails for measurep for star with ID %d \n",
					star->id);
			return(1);
		}
		mp->measure = measure;
		add_mp_to_biglist(mp);


		/* 
		 * walk down the list of linked s_measureps belonging to 'star',
		 * create a new s_measurep, and insert the new one at the end 
		 */
#if OLD
		if ((mp = (struct s_measurep *) malloc(sizeof(struct s_measurep))) == NULL) {
#else
		if ((mp = new_measurep()) == NULL) {
#endif
			fprintf(stderr, "malloc fails for star measurep for star with ID %d \n",
					star->id);
			return(1);
		}
		mp->measure = measure;
		add_mp_to_star(star, mp);


		/*
		 * Walk down the list of all s_measurep's belonging to the
		 *   the image for this measurement, create a new s_measurep
		 *   to point to this measurement, and add it to the tail of the list
		 */
#if OLD
		if ((mp = (struct s_measurep *) malloc(sizeof(struct s_measurep))) == NULL) {
#else
		if ((mp = new_measurep()) == NULL) {
#endif
			fprintf(stderr, "malloc fails for image measurep for star with ID %d \n",
					star->id);
			return(1);
		}
		mp->measure = measure;
		add_mp_to_image(image, mp);


	}

	return(0);
}



	/************************************************************************
	 * PROCEDURE: read_images
	 *
	 * DESCRIPTION: Read in information on each image.  They should be
	 *              in order by time of observation.
	 *
	 * RETURNS:
	 *    0         if all goes well
	 *    1         if an error occurs
	 *
	 */

static int
read_images
	(
	FILE *fp
	)
{
	int i;

	if ((image_arr = (struct s_image *)
				calloc(num_images, sizeof(struct s_image))) == NULL) {
		fprintf(stderr, "read_images: can't calloc for %d images\n",
				num_images);
		return(1);
	}

	for (i = 0; i < num_images; i++) {
		if (read_one_image(fp, &(image_arr[i])) != 0) {
			fprintf(stderr, "read_images: read_one_image fails for image %d \n",
					i);
			return(1);
		}
	}	

	return(0);
}


	/********************************************************************
	 * PROCEDURE: read_one_image
	 *
	 * DESCRIPTION: Read the pertinent information in a single image 
	 *              structure from the given FILE pointer.
	 *
	 * RETURNS:
	 *    0         if all goes well
	 *    1         if an error occurs
	 */

static int
read_one_image
	(
	FILE *fp, 
	struct s_image *im
	)
{
	char line[LINELEN+1];

	if (fgets(line, LINELEN, fp) == NULL) {
		fprintf(stderr, "read_one_image: can't read a line for image \n");
		return(1);
	}

	if (sscanf(line, "%d %s %lf %lf %d %d",
			&(im->index), im->name, &(im->exptime), 
			&(im->jd), &(im->num_measures), &(im->num_match)) != 6) {
		fprintf(stderr, "read_one_image: can't read data from line \n");
		return(1);
	}
	im->include = 1;
	im->mp = (struct s_measurep *) NULL;

	return(0);
}



	/**********************************************************************
	 * PROCEDURE: check_mags
	 *
	 * DESCRIPTION: Set to zero the 'include' field of any measurement 
	 *              with an instrumental magnitude 
	 *              which is outside the given input instrumental magnitude 
	 *              limits.
	 *
	 * RETURNS:
	 *     nothing
	 */

static void 
check_mags
	(
	double min, 
	double max
	)
{
	int num_kept, num_removed;
	int num_good_kept, num_good_removed;
	struct s_measurep *mp;
	struct s_measure *measure;

	num_kept = 0;
	num_removed = 0;
	num_good_kept = 0;
	num_good_removed = 0;

	mp = measure_list;
	while (mp != NULL) {
		measure = mp->measure;
		shAssert(measure != NULL);
		if ((measure->obs_mag < min) || (measure->obs_mag > max)) {
			if (measure->include != 0) {
				num_good_removed++;
			}
			measure->include = 0;
			num_removed++;
		}
		else {
			if (measure->include != 0) {
				num_good_kept++;
			}
			num_kept++;
		}

		mp = mp->next;
	}

#ifdef DEBUG
	printf("check_mags:  all   %6d kept, %6d removed \n", num_kept, num_removed);
	printf("check_mags:  good  %6d kept, %6d removed \n", 
				num_good_kept, num_good_removed);
#endif
}




	/**********************************************************************
	 * PROCEDURE: add_mp_to_biglist
	 *
	 * DESCRIPTION: Given an s_measurep struct, add it to the tail
	 *              of the giant list of all measurep's.  The head 
	 *              of this giant list is "measure_list".
	 *
	 *              Increment the count of items on the list, which
	 *              is stored in "num_measures".
	 *              We also set the "measure_id" field inside the
	 *              s_measure struct to which this mp points
	 *              at this time.
	 *
	 * RETURNS:
	 *     nothing
	 */

static void 
add_mp_to_biglist
	(
	struct s_measurep *mp        /* I: add this to tail of big list */
	)
{
	static struct s_measurep *tail;

	if (measure_list == NULL) {
		measure_list = mp;
		tail = mp;
	} else {
		tail->next = mp;
		tail = tail->next;
	}

	/* set the "measure_id" field of the s_measure to which this mp points */
	shAssert(mp->measure != NULL);
	mp->measure->measure_id = num_measures;
	num_measures++;

}


	/**********************************************************************
	 * PROCEDURE: add_mp_to_star
	 *
	 * DESCRIPTION: Given a star and an s_measurep struct, add the measurep
	 *              to the tail of the star's linked list of measurep's.
	 *
	 * RETURNS:
	 *     nothing
	 */

static void 
add_mp_to_star
	(
	struct s_star *star,         /* I/O: add to linked list of this star ... */
	struct s_measurep *mp        /* I: this measurep structure */
	)
{
	struct s_measurep *tail;

	if (star->mp == NULL) {
		star->mp = mp;
	} else {
		tail = star->mp;
		while (tail->next != NULL) {
			tail = tail->next;
		}
		tail->next = mp;
	}

}


	/**********************************************************************
	 * PROCEDURE: add_mp_to_image
	 *
	 * DESCRIPTION: Given an image and an s_measurep struct, add the measurep
	 *              to the tail of the image's linked list of measurep's.
	 *
	 * RETURNS:
	 *     nothing
	 */

static void 
add_mp_to_image
	(
	struct s_image *image,       /* I/O: add to linked list of this image ... */
	struct s_measurep *mp        /* I: this measurep structure */
	)
{
	struct s_measurep *tail;

	if (image->mp == NULL) {
		image->mp = mp;
	} else {
		tail = image->mp;
		while (tail->next != NULL) {
			tail = tail->next;
		}
		tail->next = mp;
	}

}



	/**********************************************************************
	 * PROCEDURE: check_measurements
	 *
	 * DESCRIPTION: Sanity checks.
	 *
	 *              First, we look at each image.  Check that the 
	 *              "num_measures" field equals the length of the
	 *              linked list of s_measurep structs.
	 *
	 *
	 * RETURNS:
	 *     0        if all checks are okay
	 *     1        if any check fails
	 */

static int 
check_measurements
	(
	void
	)
{
	int i, num;
	struct s_image *image;
	struct s_measurep *mp;

	for (i = 0; i < num_images; i++) {
		image = &(image_arr[i]);

		/* special case: if zero stars in an image, nothing to check */
		if (image->num_measures == 0) {
			if (image->mp != NULL) {
				fprintf(stderr, "check_measurements: image %d has 0 stars by non-NULL mp? \n",
					image->index);
				return(1);
			}
			continue;
		}

		num = 0;
		mp = image->mp;
		while (mp->next != NULL) {
			num++;
			mp = mp->next;
		}
		num++;
		if (num != image->num_measures) {
			fprintf(stderr, "check_measurements: image %d has num_measures %d but %d in list\n",
					image->index, image->num_measures, num);
			return(1);
		}

	}

	return(0);
}




	/**********************************************************************
 	 * PROCEDURE: check_varmin
 	 *
 	 * DESCRIPTION: Check each star to see if it has any input magnitude
 	 *              values less than "varmin", or greater than "varmax".  
	 *              If so, we set the 
 	 *              star's "variable" flag to 1, meaning that we will
 	 *              include it in the ensemble, but not use it in 
 	 *              determining the solution.
 	 *
 	 * RETURNS:
 	 *     nothing
 	 */
 
static void 
check_varmin
 	(
 	double varmin,
 	double varmax
 	)
{
 	struct s_measurep *mp;
 	struct s_measure *measure;
 	struct s_star *star;
 
 	mp = measure_list;
 	while (mp != NULL) {
 		measure = mp->measure;
 		shAssert(measure != NULL);
 		if (measure->obs_mag < varmin) {
 			star = &(star_arr[measure->star_id]);
 			star->variable = 1;
#ifdef DEBUG
 			printf("check_varmin: star %5d has obs %6.3f < %6.3f, so mark as variable \n",
 							measure->star_id, measure->obs_mag, varmin);
 #endif
 		}
 		if (measure->obs_mag > varmax) {
 			star = &(star_arr[measure->star_id]);
 			star->variable = 1;
#ifdef DEBUG
 			printf("check_varmin: star %5d has obs %6.3f > %6.3f, so mark as variable \n",
 							measure->star_id, measure->obs_mag, varmax);
 #endif
 		}
 		mp = mp->next;
 	}
 
}
 
 

	/**********************************************************************
 	 * PROCEDURE: check_vardet
 	 *
 	 * DESCRIPTION: Check each star to count the number of detections
 	 *              it has.  If that number is <= vardet,
	 *              then we set the 
 	 *              star's "variable" flag to 1, meaning that we will
 	 *              include it in the ensemble, but not use it in 
 	 *              determining the solution.
 	 *
 	 * RETURNS:
 	 *     nothing
 	 */
 
static void 
check_vardet
 	(
 	int vardet
 	)
{
	int i, num_measure;
	int num_marked;
 	struct s_measurep *mp;
 	struct s_measure *measure;
 	struct s_star *star;

	num_marked = 0;

	for (i = 0; i < num_stars; i++) {
		star = &(star_arr[i]);
		if (star->include == 1) {

			/* first, count the number of good measurements of this star */
			num_measure = 0;
			mp = star->mp;
			while (mp != NULL) {
				shAssert((measure = mp->measure) != NULL);
				if (measure->include != 0) {
					num_measure++;
				}
				mp = mp->next;
			}

			/* 
			 * now, if the star has too few good measurements,
			 *    we will mark the star as "variable",
			 *    and mark the measures as not to be included in solution
			 */
			if (num_measure <= vardet) {
#ifdef DEBUG
				printf("check_vardet: star %5d has %d < %d measures, mark var \n",
						i, num_measure, vardet);
#endif
				star->variable = 1;
				num_marked++;
			}
		}
	}
#ifdef DEBUG
	printf("check_vardet: marked as variable %d stars \n", 
					num_marked);
#endif
		
}
 
 
 
 
 

	/**********************************************************************
 	 * PROCEDURE: new_measurep
 	 *
 	 * DESCRIPTION: Try to allocate space for a new s_measurep 
	 *              structure. 
  	 *             
	 *              Set the 'next' element of the structure to NULL.
 	 *
 	 * RETURNS:
 	 *              pointer to new struct       if all goes well
 	 *              NULL                        if an error occurs
 	 */
 
static struct s_measurep *
new_measurep
 	(
 	void
 	)
{
 	struct s_measurep *new;

	if ((new = (struct s_measurep *) malloc(sizeof(struct s_measurep))) == NULL) {
		fprintf(stderr, "new_measurep: malloc fails for measurep \n");
		return(NULL);
	}
	new->next = (struct s_measurep *) NULL;

	return(new);
}
 
 
 
 
 
